#include  <linux/config.h>

#include <linux/input.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/proc_fs.h>
#include <linux/kbd_kern.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include <linux/random.h>
#include <linux/spinlock.h>
#include <linux/smp_lock.h>
#include <linux/poll.h>
#include <linux/pm.h>
#include <linux/delay.h>
#include <asm/arch/pxa-regs.h>
#include <asm/delay.h>
#include <asm/io.h>

#ifdef FIELD_IRQ_DIRECTLY
#include <asm/irq.h>
#include <linux/interrupt.h>
#endif

#undef KBD_DEBUG_FLOW

// For size of keymap
#include "fiona_keymap.h"

// For actual keymap
#include "fiona_keymap.c"

// Make up some constants
#define FOXCONN_VENDOR_ID           0xDEAD
#define FIONA_KEYBOARD_DEVICE_NAME    "Fiona IOC Keyboard Device"

int     g_fiona_kbd_used = 0;
struct  input_dev g_fiona_kbd_dev;


/*****************************************************************************
 * Queue Support
 *****************************************************************************/
#define AUX_BUF_SIZE       2048    

struct kbd_queue {
    unsigned long head;
    unsigned long tail;
    wait_queue_head_t proc_list;
    struct fasync_struct *fasync;
    unsigned char buf[AUX_BUF_SIZE];
};

static struct kbd_queue *queue;    /* Keyboard data buffer. */
spinlock_t g_kbd_lock = SPIN_LOCK_UNLOCKED;
static int g_kdb_stream_enabled=0;


static int
fasync_kbd(int fd, struct file *filp, int on)
{
    int retval;

    retval = fasync_helper(fd, filp, on, &queue->fasync);
    if( retval < 0)
        return retval;
    return 0;
}

static void kbd_clear_init_queue(void)
{
    memset(queue, 0, sizeof(*queue));
    queue->head = queue->tail = 0;
    init_waitqueue_head(&queue->proc_list);
}


static void
kdb_queue_packet(uint8_t *buf, int n )
{
    unsigned long flags;

    spin_lock_irqsave(&g_kbd_lock, flags);
    if( g_fiona_kbd_used && g_kdb_stream_enabled )
    {
        int head = queue->head;

        while(n--)
        {
           queue->buf[head] = *buf++;
           head = (head + 1) & (AUX_BUF_SIZE-1);
        }
        if (head != queue->tail)
        {
            queue->head = head;
            kill_fasync(&queue->fasync, SIGIO, POLL_IN);
            wake_up_interruptible(&queue->proc_list);
        }
    }
    spin_unlock_irqrestore(&g_kbd_lock, flags);
}

/*
 * Add byte to output queue
 */
static void
kdb_queue_byte(int val)
{
    unsigned long flags;

    spin_lock_irqsave(&g_kbd_lock, flags);
    if (g_fiona_kbd_used) {
        int head = queue->head;

        queue->buf[head] = val;
        head = (head + 1) & (AUX_BUF_SIZE-1);
        if (head != queue->tail) {
            queue->head = head;
            kill_fasync(&queue->fasync, SIGIO, POLL_IN);
            wake_up_interruptible(&queue->proc_list);
        }
    }
    spin_unlock_irqrestore(&g_kbd_lock, flags);
}

/*
 *  get byte from queue
 */
static unsigned char
get_from_queue(void)
{
    unsigned char result;
    unsigned long flags;

    spin_lock_irqsave(&g_kbd_lock, flags);
    result = queue->buf[queue->tail];
    queue->tail = (queue->tail + 1) & (AUX_BUF_SIZE-1);
    spin_unlock_irqrestore(&g_kbd_lock, flags);
    return result;
}
/*
 *  Check if queue is empty
 */
static inline int
queue_empty(void)
{
    return queue->head == queue->tail;
}



/*****************************************************************************
 * Core Driver 
 *****************************************************************************/

static int fiona_kbd_open(struct input_dev *dev)
{
#ifdef KBD_DEBUG_FLOW
    printk("fiona_kbd_open()\n");
#endif

    if (g_fiona_kbd_used++)
        return 0;

#ifdef FIELD_IRQ_DIRECTLY
	if (request_irq(IRQ_BTUART, fiona_kbd_interrupt, 0, "keyboard", NULL)) {
        printk(KERN_ERR "fiona_kbd_open(): Can't allocate irq %d\n", fiona_kbd_irq);
        g_fiona_kbd_used--;
        return -EBUSY;
    }
#endif

    return 0;
}

static void fiona_kbd_close(struct input_dev *dev)
{
#ifdef KBD_DEBUG_FLOW
    printk("fiona_kbd_close()\n");
#endif

    // One less client
    g_fiona_kbd_used--;

    // If there are no more clients, clear the queue out....
    if (0 == g_fiona_kbd_used) {
        kbd_clear_init_queue();
#ifdef FIELD_IRQ_DIRECTLY
        free_irq(IRQ_BTUART, fiona_kbd_interrupt);
#endif
    }
}

#ifdef FIELD_IRQ_DIRECTLY
static void fiona_kbd_interrupt(int irq, void *dummy, struct pt_regs *fp)
{
	input_report_key(&g_fiona_kbd_dev, kbd_ioc_xlate[scancode], pressed);
	input_sync(&g_fiona_kbd_dev);
}
#endif

void fiona_kbd_report_key(int scancode, int pressed)
{
    if (!g_fiona_kbd_used) {
#ifdef KBD_DEBUG_FLOW
        printk("Throwing away %s '%c' (key code 0x%02x)\n",(pressed==0)?"Released":"Pressed",kbd_ioc_xlate_char[scancode],scancode);
#endif
        return;
    }

#ifdef KBD_DEBUG_FLOW
    printk("%s '%c' (key code 0x%02x)\n",(pressed==0)?"Released":"Pressed",kbd_ioc_xlate_char[scancode],scancode);
#endif

#if 0
    // This will bug on scroll up/down - shouldn't need this anyway
	input_report_key(&g_fiona_kbd_dev, kbd_ioc_xlate[scancode], pressed);
	input_sync(&g_fiona_kbd_dev);
#endif

    // Queue up the character in our queue if it's up
    if (pressed) {
#ifdef KBD_DEBUG_FLOW
        printk("Queued up '%c'\n",kbd_ioc_xlate_char[scancode]);
#endif
        kdb_queue_byte(kbd_ioc_xlate_char[scancode]);
    }
}

/*
 * Put bytes from input queue to user buffer.
 */
static ssize_t
fiona_kbd_read(struct file * file, char * buffer, size_t count, loff_t *ppos)
{
    DECLARE_WAITQUEUE(wait, current);
    ssize_t i = count;
    unsigned char c;

#ifdef KBD_DEBUG_FLOW
    printk("fiona_kbd_read()\n");
#endif

    if (queue_empty()) {
        if (file->f_flags & O_NONBLOCK)
            return -EAGAIN;
        add_wait_queue(&queue->proc_list, &wait);
repeat:
        set_current_state(TASK_INTERRUPTIBLE);
        if (queue_empty() && !signal_pending(current)) {
            schedule();
            goto repeat;
        }
        current->state = TASK_RUNNING;
        remove_wait_queue(&queue->proc_list, &wait);
    }
    while (i > 0 && !queue_empty()) {
        c = get_from_queue();
        put_user(c, buffer++);
#ifdef KBD_DEBUG_FLOW
        printk(KERN_INFO "fiona_kbd_read: read_byte 0x%x\n", (int) c );
#endif

        i--;
    }
    if (count-i) {
        file->f_dentry->d_inode->i_atime = CURRENT_TIME;
        return count-i;
    }
    if (signal_pending(current))
        return -ERESTARTSYS;
    return 0;
}

/*
 * Poll to the aux device.
 */
static unsigned int
fiona_kbd_poll(struct file *file, poll_table * wait)
{
    poll_wait(file, &queue->proc_list, wait);
    if (!queue_empty())
    {
#ifdef KBD_DEBUG_FLOW
        printk(KERN_INFO "fiona_kbd_poll: kbd_poll != 0\n");
#endif

        return POLLIN | POLLRDNORM;
    }

#ifdef KBD_DEBUG_FLOW
    printk(KERN_INFO "fiona_kbd_poll: kbd_poll = 0\n");
#endif

    return 0;
}

struct file_operations keyboard_fops = {
    read:       fiona_kbd_read,
    write:      NULL,
    poll:       fiona_kbd_poll,
    open:       fiona_kbd_open,
    release:    fiona_kbd_close,
    fasync:     fasync_kbd,
};

/*
 * Initialize driver.
 */
#define KEYBOARD_MINOR 134
static struct miscdevice keyboard_device = {
    KEYBOARD_MINOR, "keyboard", &keyboard_fops 
};

static int __init fiona_kbd_init(void)
{
    int x = 0;
    int retVal = 0;

    // Setup our driver
	g_fiona_kbd_dev.open = fiona_kbd_open;
	g_fiona_kbd_dev.close = fiona_kbd_close;
    g_fiona_kbd_dev.name = kmalloc(strlen(FIONA_KEYBOARD_DEVICE_NAME)+1, GFP_KERNEL);
    if (g_fiona_kbd_dev.name == NULL)  {
        printk(KERN_ERR "fiona_keyboard.c: Can't allocate space for device name!\n");
        return -ENOMEM;
    }
    strcpy(g_fiona_kbd_dev.name,FIONA_KEYBOARD_DEVICE_NAME);

    // Describe IOC ??
	g_fiona_kbd_dev.id.bustype = BUS_RS232;
	g_fiona_kbd_dev.id.vendor = FOXCONN_VENDOR_ID;
	g_fiona_kbd_dev.id.product = 0;
	g_fiona_kbd_dev.id.version = 5;

	g_fiona_kbd_dev.keycode = (void*)kbd_ioc_xlate;
	g_fiona_kbd_dev.keycodemax = FIONA_KEYMAP_SIZE;
	g_fiona_kbd_dev.keycodesize = 1;

#ifdef FIELD_IRQ_DIRECTLY
    // Reserve our serial port IRQ
	if (request_irq(IRQ_BTUART, fiona_kbd_interrupt, 0, "keyboard", NULL)) {
        printk(KERN_ERR "fiona_keyboard.c: Can't allocate irq %d\n", IRQ_BTUART);
        return -EBUSY;
    }
#else
    if (register_ioc_keyboard_handler(fiona_kbd_report_key)) {
        printk("Could not register Fiona keyboard handler with IOC !\n");
        return -EBUSY;
    }
    else printk("Registered %s.\n",FIONA_KEYBOARD_DEVICE_NAME);
#endif
	
    // Just generic key and button events, but we do want key repeat behavior
	g_fiona_kbd_dev.evbit[0] = (BIT(EV_KEY) | BIT(EV_REP));

    // Load up our keybit field
    for (x=0; x<FIONA_KEYMAP_SIZE; x++)
    	g_fiona_kbd_dev.keybit[LONG(kbd_ioc_xlate[x])] = BIT(kbd_ioc_xlate[x]);
	
#ifdef USB_REGISTER
    // Register as an input device
	input_register_device(&g_fiona_kbd_dev);
#else
    retVal = misc_register(&keyboard_device);
    if( retVal < 0) {
#ifdef FIELD_IRQ_DIRECTLY
        free_irq ( IRQ_BTUART, NULL);
#else
        unregister_ioc_keyboard_handler(fiona_kbd_report_key);
#endif
    }
    else  {
#ifdef KBD_DEBUG_FLOW
    printk("Keyboard open, queue initialized\n");
#endif

        queue = (struct kbd_queue *) kmalloc(sizeof(*queue), GFP_KERNEL);
        kbd_clear_init_queue();
    }
#endif

    return(0);
}

static void __exit fiona_kbd_exit(void)
{
#ifdef KBD_DEBUG_FLOW
    printk("fiona_kbd_exit()\n");
#endif
    unregister_ioc_keyboard_handler(fiona_kbd_report_key);
    input_unregister_device(&g_fiona_kbd_dev);
    kfree(g_fiona_kbd_dev.name);
    kfree(queue);
#ifdef FIELD_IRQ_DIRECTLY
    free_irq(BUTTON_IRQ, fiona_kbd_interrupt);
#endif
}

module_init(fiona_kbd_init);
module_exit(fiona_kbd_exit);
