/*
 *  Gumstix SMC91C111 chip intialization driver
 *
 *  Author:     Craig Hughes
 *  Created:    December 9, 2004
 *  Copyright:  (C) 2004 Craig Hughes
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/device.h>

#include <asm/hardware.h>
#include <asm/arch/pxa-regs.h>
#include <asm/delay.h>

#include <asm/arch/gumstix.h>

static struct resource gumstix_smc91x_resources[] = {
	[0] = {
		.start  = 0x04000300,
		.end    = 0x040fffff,
		.flags  = IORESOURCE_MEM,
		},
	[1] = {
		.start  = GUMSTIX_ETH_IRQ,
		.end    = GUMSTIX_ETH_IRQ,
		.flags  = IORESOURCE_IRQ,
		},
};

static struct platform_device gumstix_smc91x_device = {
	.name           = "smc91x",
	.id             = -1,
	.num_resources  = ARRAY_SIZE(gumstix_smc91x_resources),
	.resource       = gumstix_smc91x_resources,
};


int __init gumstix_smc91x_init(void)
{
	/* Set up nPWE */
	pxa_gpio_mode(GPIO49_nPWE_MD);

	/* Reset the SMC91c111 */
	pxa_gpio_mode(GPIO_GUMSTIX_ETH_RST_MD);
	GPSR(GPIO_GUMSTIX_ETH_RST) = GPIO_bit(GPIO_GUMSTIX_ETH_RST);
	udelay(1); // Hold RESET pin high for at least 100ns
	GPCR(GPIO_GUMSTIX_ETH_RST) = GPIO_bit(GPIO_GUMSTIX_ETH_RST);
	udelay(50); // PHY will be read-ready after 50ms

	return platform_device_register(&gumstix_smc91x_device);
}

void __exit gumstix_smc91x_exit(void)
{
	platform_device_unregister(&gumstix_smc91x_device);
}

void gumstix_smc91x_load(void) {}
EXPORT_SYMBOL(gumstix_smc91x_load);

module_init(gumstix_smc91x_init);
module_exit(gumstix_smc91x_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Craig Hughes <craig@gumstix.com>");
MODULE_DESCRIPTION("Gumstix board SMC91C111 chip initialization driver");
MODULE_VERSION("1:0.1");
