/**********************************************************************
 * Philips ISP1761 device controller driver
 *
 * (c) 2002 Koninklijke Philips Electronics N.V., All rights reserved
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * File Name:	pdc.h	(Based on usb_pdc.h)
 *
 * Refering linux kernel version 2.6.9
 *
 * History:	
 *
 * Version	Date		Author			Comments
 * ----------------------------------------------------------------------
 * 1.0		29/11/05	Prabhakar Kalasani	Initial Creation
 *
 ************************************************************************/

#ifndef __PDC_H__
#define __PDC_H__

#define CONTROL_EP      			0x00
#define ISOC_EP         			0x01
#define BULK_EP         			0x02
#define INT_EP          			0x03

#define DIR_IN          			0x01
#define DIR_OUT         			0x00

#define AVAILABLE       			0x00
#define IN_USE          			0x01
#define STALL           			0x02

/* Registers & bits */

#define DC_ADDRESS_REG				0x0200
#define ADDRESS_DEVEN				0x80
	
#define DC_MAX_PKT_SIZE_REG			0x0204
#define MAXP_SIZE_NTRANS_1			0x0000
#define MAXP_SIZE_NTRANS_2			0x0800
#define MAXP_SIZE_NTRANS_3			0x1800

#define DC_EP_TYPE_REG				0x0208
#define EP_TYP_ENDPTYP_ISO			0x0001
#define EP_TYP_ENDPTYP_BULK			0x0002
#define EP_TYP_ENDPTYP_INTR			0x0003
#define EP_TYP_DBLBUF				0x0004
#define EP_TYP_ENABLE				0x0008
#define EP_TYP_NOEMPKT				0x0010

#define EPCONFIG_EPDIR				0x0000

#define DC_MODE_REG				0x020C
#define MODE_SOFTCT				0x0001
#define MODE_PWRON				0x0002
#define MODE_WKUPCS				0x0004
#define MODE_GLINTENA				0x0008
#define MODE_SFRESET				0x0010
#define MODE_GOSUSP				0x0020
#define MODE_SNDRSU				0x0040
#define MODE_CLKAON				0x0080
#define MODE_VBUSSTAT				0x0100
#define MODE_DMACLKON				0x0200


#define DC_INTR_CONFIG_REG			0x0210
#define CONFIG_INTPOL_L 			0x00
#define CONFIG_INTPOL_H        			0x01
#define CONFIG_INTLVL_L   			0x00
#define CONFIG_INTLVL_P        			0x02
#define CONFIG_INT_DBG_OUTACK_NAK_NYET		0x00
#define CONFIG_INT_DBG_OUT_ACK_NYET 		0x04
#define CONFIG_INT_DBG_OUT_ACK_NYET_FNAK	0x08
#define CONFIG_INT_DBG_IN_ACK_NAK		0x00
#define CONFIG_INT_DBG_IN_ACK			0x10
#define CONFIG_INT_DBG_IN_ACK_FNAK		0x20
#define CONFIG_INT_DBG_ACK_NAK			0x00
#define CONFIG_INT_DBG_ACK			0x40
#define CONFIG_INT_ACK_FNAK			0x80
#define DEFAULT_INT_CONFIG			(CONFIG_INT_ACK_FNAK | \
						 CONFIG_INT_DBG_ACK | \
						 CONFIG_INT_DBG_IN_ACK| \
						 CONFIG_INT_DBG_OUT_ACK_NYET )	 //0xD4

#define DC_DEBUG_REG				0x0212

#define DC_INTR_ENABLE_REG			0x0214
#define IE_BRST					0x00000001
#define IE_SOF       				0x00000002
#define IE_PSOF					0x00000004
#define IE_SUSP 				0x00000008
#define IE_RESM					0x00000010
#define IE_HS_STA				0x00000020
#define IE_DMA					0x00000040
#define IE_VBUS					0x00000080
#define IE_EP0SETUP				0x00000100
#define IE_EP0RX				0x00000400
#define IE_EP0TX				0x00000800
#define IE_EP1RX				0x00001000
#define IE_EP1TX				0x00002000
#define IE_EP2RX				0x00004000
#define IE_EP2TX				0x00008000
#define IE_EP3RX				0x00010000
#define IE_EP3TX				0x00020000
#define IE_EP4RX				0x00040000
#define IE_EP4TX				0x00080000
#define IE_EP5RX				0x00100000
#define IE_EP5TX				0x00200000
#define IE_EP6RX				0x00400000
#define IE_EP6TX				0x00800000
#define IE_EP7RX				0x01000000
#define IE_EP7TX				0x02000000

#define IE_NON_CTRL_EP_MASK			(IE_EP1RX | IE_EP1TX | IE_EP2RX | IE_EP2TX | \
						 IE_EP3RX | IE_EP3TX | IE_EP4RX | IE_EP4TX | \
						 IE_EP5RX | IE_EP5TX | IE_EP6RX | IE_EP6TX | \
						 IE_EP7RX | IE_EP7TX)
#define IE_BASIC_CTRL_EP_MASK  			(IE_EP0RX | IE_EP0TX | IE_EP0SETUP | IE_VBUS | \
						 IE_HS_STA | IE_SUSP | IE_RESM | IE_BRST) 

#define DC_INTERRUPT_REG			0x0218
#define INT_REG_BRESET				0x00000001
#define INT_REG_SOF				0x00000002
#define INT_REG_PSOF				0x00000004
#define INT_REG_SUSP				0x00000008
#define INT_REG_RESUME				0x00000010
#define INT_REG_HS_STAT				0x00000020
#define INT_REG_DMA				0x00000040
#define INT_REG_EP0SETUP			0x00000100
#define INT_REG_EP0_RX				0x00000400
#define INT_REG_EP0_TX				0x00000800
#define INT_REG_EP1_RX         			0x00001000
#define INT_REG_EP1_TX         			0x00002000
#define INT_REG_EP2_RX         			0x00004000
#define INT_REG_EP2_TX         			0x00008000
#define INT_REG_EP3_RX         			0x00010000
#define INT_REG_EP3_TX         			0x00020000
#define INT_REG_EP4_RX        			0x00040000
#define INT_REG_EP4_TX         			0x00080000
#define INT_REG_EP5_RX         			0x00100000
#define INT_REG_EP5_TX         			0x00200000
#define INT_REG_EP6_RX         			0x00400000
#define INT_REG_EP6_TX         			0x00800000
#define INT_REG_EP7_RX         			0x01000000
#define INT_REG_EP7_TX         			0x02000000

#define DC_BUFFER_LENGTH_REG			0x021C

#define DC_BUFFER_STATUS_REG			0x021E
#define BUF0_FILLED				0x01
#define BUF1_FILLED				0x02
#define BUF1_AND_2_FILLED			0x03

#define DC_DATAPORT_REG				0x0220
 
#define DC_CTRLFN_REG				0x0228
#define CNTRLFN_STALL				0x01
#define CNTRLFN_STATUS_NAK			0x00
#define CNTRLFN_STATUS_EM_ACK			0x02
#define CNTRLFN_VENDP				0x08
#define CNTRLFN_CLBUF				0x10
#define CNTRLFN_DSEN				0x04

#define DC_EP_INDEX_REG				0x022c
#define EP0SETUP	   			0x20
#define EP0RX               			0x00
#define EP0TX               			0x01
#define EP1RX               			0x02
#define EP1TX               			0x03
#define EP2RX               			0x04
#define EP2TX               			0x05
#define EP3RX               			0x06
#define EP3TX               			0x07
#define EP4RX               			0x08
#define EP4TX               			0x09
#define EP5RX               			0x0a
#define EP5TX               			0x0b
#define EP6RX               			0x0c
#define EP6TX               			0x0d
#define EP7RX               			0x0e
#define EP7TX               			0x0f

#define EP_DIRECTION_IS_IN(ep)	(ep->ep_index & DIR_IN)

#define DC_DMA_COMMAND_REG			0x0230
#define DMA_CMD_CLEAR_BUFFER			0x0F

#define DC_DMA_TRANSFER_CTR_REG			0x0234

#define DC_DMA_CONFIG_REG			0x0238

#define DC_DMA_HARDWARE_REG			0x023C

#define DC_DMA_INTR_REASON_REG			0x0250

#define DC_DMA_INTR_ENABLE_REG			0x0254

#define DC_DMA_EP_INDEX_REG			0x0258

#define DC_DMA_BURST_CTR_REG			0x0264

#define DC_CHIP_ID_REG				0x0270

#define DC_FRAME_NUMBER_REG			0x0274

#define DC_SCRATCH_REG       			0x0278

#define DC_UNLOCK_DEVICE_REG			0x027C

#define DC_INTR_PULSE_SIZE_REG			0x0280
#define DEFAULT_INT_PULSE_SIZE			0x1E

#define DC_TESTMODE_REG				0x0284
#define TESTMODE_SE0_NAK			0x01
#define TESTMODE_JSTATE				0x02
#define TESTMODE_KSTATE				0x04
#define TESTMODE_PRBS				0x08
#define TESTMODE_FORCEFS	   		0x10
#define TESTMODE_FORCEHS	   		0x80

#define HC_HW_MODE_REG				0x0300

#define OTG_CONTROL_REG				0x0374

#define	isp1761_vendor_id			0x04CC
#define	isp1761_product_id			0x1b41 //0x1A64
#define	isp1761_product_string			{ 68,3,'1',0, '7',0, '6',0, '2',0, '-',0, 'O',0, 'T',0, 'G',0, ' ',0,'M',0, 'A',0, 'S',0, 'S',0, ' ',0,'S',0, 't',0, 'o',0, 'r',0, 'a',0, 'g',0, 'e',0, ' ',0,'P',0, 'X',0, 'A',0, '-',0, '2',0, '5',0, '0',0,' ',0, 'K',0, 'i',0, 't',0 }

#define	isp1761_mfg_string			{ 46,3,'P',0, 'h',0, 'i',0, 'l',0, 'i',0, 'p',0, 's',0,' ',0, 'S',0, 'e',0, 'm',0, 'i',0,'c',0, 'o',0, 'n',0, 'd',0, 'u',0, 'c',0, 't',0, 'o',0, 'r',0, 's',0 }

#define HIGH_SPEED 				1
#define FULL_SPEED 				0


/* Device states */

#define STATE_DETACHED			0x00	// D+ pullup disconnected
#define STATE_ATTACHED			0x01	// D+ pullup connected
#define STATE_POWERED			0x02	// VBUS detected
#define STATE_POWERED_SUSPENDED		0x03
#define STATE_DEFAULT			0x04	// Address == 0, config == 0
#define STATE_DEFAULT_SUSPENDED		0x05
#define STATE_ADDRESS			0x06	// Address != 0, config == 0
#define STATE_ADDRESS_SUSPENDED		0x07
#define STATE_CONFIGURED		0x08	// Address != 0, config != 0
#define STATE_CONFIGURED_SUSPENDED	0x09

#define STATE_SUSPENDED			0x01	// bit for suspended state


/* Values for event_flag in the connection notification code. */

#define USB_CONNECT_EVENT	1
#define USB_DISCONNECT_EVENT	2


#ifdef CONFIG_ARCH_FIONA

#define PDC_PHYS_BASE		0x0C100000	// Physical address of PDC
#define PDC_PHYS_LENGTH		0x00080000	// Addr[17:1]

#define PDC_IRQ			IRQ_GPIO(3)	// Connected to GPIO3

#endif

#define EP0_FIFO_SIZE		((unsigned)  64)
#define BULK_FIFO_SIZE_FS	((unsigned)  64)
#define BULK_FIFO_SIZE_HS	((unsigned) 512)

#define PDC_NUM_ENDPOINTS	3

#define VBUS_TIMER_INTERVAL	(HZ)		// time between VBUS checks


struct pdc_ep {
	struct usb_ep		ep;
	struct pdc_globals	*dev;
	
	const struct usb_endpoint_descriptor	*desc;
	struct list_head	queue;
	unsigned long		irqs;
	
	u8			ep_number;	// 0,1,2
	u8			ep_index;	// EP0TX, EP2RX, etc.
	u8			bEndptAddress;	// 0x81, 0x02, etc.
	u8			bmAttributes;	// Bulk, etc.
	u8			fifo_busy;
	u8			stopped;

	u8			needs_kick;
};


struct pdc_request {
	struct usb_request	req;
	struct list_head	queue;
	u32			timestamp;
};


struct pdc_globals {
	struct usb_gadget		gadget;
	struct usb_gadget_driver	*driver;
	struct device			*dev;
	
	struct timer_list		timer;
	struct timer_list		vbus_timer;

	spinlock_t			lock;
	
	void				*base_address;
	
	u32				intr_enables;
	u32				intr_enables_bulk;
	u32				configured_vbus_draw;
	u32				current_vbus_draw;

	u8				device_state;
	u8				device_address;
	u8				device_configuration;
	u8				got_irq;

	u8				stats_counter;
	u8				remote_wakeup;
	u8				suspend_enabled;
	u8				controller_suspended;
	u8				cbw_error;
//	u8				got_disc;
//	u8				req_pending;
//	u8				req_std;
//	u8				req_config;

	struct pdc_ep			ep [PDC_NUM_ENDPOINTS];
};


/* -----  Function prototypes  ----- */

static int pdc_ep_enable(struct usb_ep *_ep,
		const struct usb_endpoint_descriptor *desc);
static int pdc_ep_disable(struct usb_ep *_ep);
static struct usb_request *pdc_ep_alloc_request(struct usb_ep *_ep,
		int gfp_flags);
static void pdc_ep_free_request(struct usb_ep *_ep, struct usb_request *_req);
static void *pdc_ep_alloc_buffer(struct usb_ep *_ep, unsigned bytes,
		dma_addr_t *dma, int gfp_flags);
static void pdc_ep_free_buffer(struct usb_ep *_ep, void *buf, dma_addr_t dma,
		unsigned bytes);
static int pdc_ep_queue(struct usb_ep *_ep, struct usb_request *_req,
		int gfp_flags);
static int pdc_ep_dequeue(struct usb_ep *_ep, struct usb_request *_req);
static int pdc_ep_set_halt(struct usb_ep *_ep, int value);
static int pdc_ep_fifo_status(struct usb_ep *_ep);
static void pdc_ep_fifo_flush(struct usb_ep *_ep);


static int pdc_get_frame(struct usb_gadget *_gadget);
static int pdc_wakeup(struct usb_gadget *_gadget);
static int pdc_vbus_draw(struct usb_gadget *_gadget, unsigned mA);


static int __init pdc_probe(struct device *device);
static int __exit pdc_remove(struct device *device);

static u16  pdc_read16(u16 reg);
static u32  pdc_read32(u16 reg);
static void pdc_write16(u16 reg, u16 data);
static void pdc_write32(u16 reg, u32 data);
static u8   pdc_vbus_state(void);
static void pdc_set_current_draw(struct pdc_globals *dev, u32 mA);
static void pdc_suspend_controller(int force_off);

static void pdc_locked_go_to_state(u8 state, u8 address, u8 configuration);
static void pdc_locked_do_ep_fifo_flush(struct usb_ep *_ep);
static void pdc_locked_do_ep_clear_halt(struct usb_ep *_ep);
static void pdc_locked_do_ep_set_halt(struct usb_ep *_ep);

static void nop_release (struct device *dev);

static void pdc_disable(struct pdc_globals *dev);
static void pdc_enable(struct pdc_globals *dev);

static void dump_state(struct pdc_globals *dev);
static void pdc_nuke(struct pdc_ep *ep, int status);
static void pdc_reinit(struct pdc_globals *dev);


/* -----  Debugging support  ----- */

#define DBG_LOG_LEVEL		KERN_ALERT

#define DBG_PRINTK(stuff...)	printk(DBG_LOG_LEVEL "pdc: " stuff)

#ifdef DEBUG

#define DBG_MSG(stuff...)	printk(DBG_LOG_LEVEL "pdc: " stuff)

#ifdef DEBUG_VERBOSE
#define DBG_VMSG(stuff...)	DBG_MSG(stuff)
#else
#define DBG_VMSG(stuff...)	do {} while (0)
#endif

#ifdef DEBUG_ENTRY
#define DBG_ENTRY(stuff...)	DBG_MSG(">>> %s\n", __FUNCTION__)
#define DBG_EXIT(stuff...)	DBG_MSG("<<< %s\n", __FUNCTION__)
#else
#define DBG_ENTRY(stuff...)	do {} while (0)
#define DBG_EXIT(stuff...)	do {} while (0)
#endif

#ifdef DEBUG_TRACE
#define DBG_TRACE(stuff...)	DBG_MSG(stuff)
#else
#define DBG_TRACE(stuff...)	do {} while (0)
#endif

#else

#define DBG_MSG(stuff...)	do {} while (0)
#define DBG_VMSG(stuff...)	do {} while (0)
#define DBG_ENTRY(stuff...)	do {} while (0)
#define DBG_EXIT(stuff...)	do {} while (0)
#define DBG_TRACE(stuff...)	do {} while (0)

#undef DEBUG_VERBOSE
#undef DEBUG_ENTRY
#undef DEBUG_TRACE

#endif /* DEBUG */

#endif /* __USB_PDC_H__ */


