/******************************************************************************
 * Philips ISP176x Host Controller Interface code file
 *
 * (c) 2002 Koninklijke Philips Electronics N.V. All rights reserved. <usb.linux@philips.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * File Name: pehci.c
 * 
 * Refering linux kernel version 2.6.9
 *
 * History:     
 *
 * Date                  Author                  Comments
 * ------------------------------------------------------------------------------
 * June 21, 2005         krishan                  Initial version
 * July 4, 2005          krishan                  Handling unprotected urbs
 * Feb 6, 2006           Grant H.                 Added ISO support
 * April 21, 2006        Grant H.                 Added bug fixes
 * 
 ********************************************************************************
 */


#define EHCI_TUNE_CERR 3
#define URB_NO_INTERRUPT 0x0080
#define EHCI_TUNE_RL_TT 0
#define EHCI_TUNE_MULT_TT 1
#define EHCI_TUNE_RL_HS 0
#define EHCI_TUNE_MULT_HS 1

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/pci.h>

#include <linux/usb.h>

#include <linux/version.h>
#include <stdarg.h>
//#include <stdio.h>


#include "pehci.h"
#include "pehcireg.h"

#include <asm/byteorder.h>
#include <asm/io.h>
#include <asm/dma.h>

#include <asm/arch-pxa/pxa-regs.h>

#include <asm/irq.h>
#include <asm/system.h>
#include <asm/unaligned.h>

#if defined(CONFIG_ARCH_FIONA)
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#endif

#include <asm/arch/fpow.h>

#ifdef CONFIG_USB_DEBUG
	#define DEBUG
#else
	#undef DEBUG
#endif

#define DEBUG  /* FIXME - force this on for now */

#ifdef DEBUG
#define PEHCI_BUG(hcd)	pehci_bug(__FUNCTION__, __LINE__, hcd )
#else
#define PEHCI_BUG(hcd)
#endif


/*Enable all other interrupt.*/
#ifdef MSEC_INT_BASED
#define INTR_ENABLE_MASK (HC_MSEC_INT /* | HC_EOT_INT | HC_CLKREADY_INT | HC_INTL_INT | HC_ATL_INT | HC_ISO_INT*/)
#else
#define INTR_ENABLE_MASK (/*HC_MSEC_INT |*/ HC_INTL_INT | HC_ATL_INT | HC_EOT_INT /*| HC_ISO_INT*/)
#endif

/*---------------------------------------------------
*    Globals for EHCI 
-----------------------------------------------------*/

/* used when updating hcd data */
static spinlock_t hcd_data_lock = SPIN_LOCK_UNLOCKED;
#define DRIVER_AUTHOR "krishan"
#define DRIVER_DESC "ISP1762 'Enhanced' Host Controller (EHCI) Driver"


static const char	hcd_name [] = "Philips ISP1762";

/* td-ptd map buffer for all 1362 buffers */
static td_ptd_map_buff_t td_ptd_map_buff[TD_PTD_TOTAL_BUFF_TYPES];

static __u8		td_ptd_pipe_x_buff_type[TD_PTD_TOTAL_BUFF_TYPES] = {
							TD_PTD_BUFF_TYPE_ATL,
							TD_PTD_BUFF_TYPE_INTL,
							TD_PTD_BUFF_TYPE_ISTL
 							};

/*global memory blocks*/
isp1762_mem_addr_t memalloc[BLK_TOTAL];

#include "mem.c"

#include "qtdptd.c"

#ifdef OTG
#include "otg.c"
#endif

/* from hal/hal_x86.c */

static u32 isp1762_reg_read32(phci_hcd *hcd, u32 offset ,int value){
                return readl(hcd->iobase + offset);
}

static void isp1762_reg_write32(phci_hcd *hcd, u32 offset ,int value)
{
                writel(value,(hcd->iobase + offset));
}
/*--------------------------------------------------------------*
 *  
 * Module dtatils: isp1762_mem_read
 * hcd                  -               host controller struct
 * start_add    -               start address, write from
 * buffer               -               write to
 * length               -               length to write
 *--------------------------------------------------------------*/


static u32
isp1762_mem_read(phci_hcd *hcd,
                __u32 start_add, __u32 * buffer, __u32 length)
{
        u8 *temp_base_mem = 0;
        int a = (int)length;
        __u32 w;
        temp_base_mem= (hcd->iobase + start_add);
        /*initialize the Register 0x33C-used to manage Multiple threads */
        writel(start_add,(hcd->iobase+0x33c));
        while(a>0){
                w = readl(temp_base_mem);
                *buffer++ = w;
                temp_base_mem = temp_base_mem+4;
                start_add +=4;
                a -=4;
            }

        /* get the tail end of a read that is not a multiple of 4 bytes */

        if(a > 0) {
          w = cpu_to_le32(readl(temp_base_mem));
          memcpy(buffer, &w, a);
        }

        return 0;

}

/*--------------------------------------------------------------*
 *  
 * Module dtatils: isp1762_mem_write
 * hcd                  -               host controller struct
 * start_add    -               start address, write to
 * buffer               -               write from
 * length               -               length to write
 *--------------------------------------------------------------*/

/* Memory read function IO */
static u32
isp1762_mem_write(phci_hcd  *hcd,
                __u32 start_add, __u32 * buffer, __u32 length)

{
        u8 *temp_base_mem = 0;
                int a = length;
                temp_base_mem = (hcd->iobase + start_add);
        while(a>0){
                writel(*buffer, temp_base_mem);
                temp_base_mem = temp_base_mem+4;
                buffer++;
                start_add +=4;
                a -=4;
        }
  return ((a < 0) || (a == 0))?0:(-1);
}

static u32
pehci_hcd_mem_read(phci_hcd *hcd,u32 local,u32 system,u32 length)
{
                return isp1762_mem_read(hcd, local,(u32*)system,length);
}


static u32
pehci_hcd_mem_write(phci_hcd *hcd, u32 local, u32 system, u32 length)
{
                return isp1762_mem_write(hcd, local,(u32*)system, length);
}


#if defined(CONFIG_ARCH_FIONA)

static struct proc_dir_entry *proc_usb_parent;
static struct proc_dir_entry *proc_usb_hostmode_ctrl;
static struct proc_dir_entry *proc_usb_errors_ctrl;

#define OTG_HOST_ENABLE_STRING		"ENABLE"

#define OTG_MODE_PERIPHERAL		0
#define OTG_MODE_HOST			1

static int otg_mode = OTG_MODE_PERIPHERAL;

static void set_otg_mode(int mode);

static int	num_interrupts = 0;
static int	num_bulk_pkts = 0;
static int	num_ctrl_pkts = 0;
static int	num_intr_pkts = 0;

#ifdef CONFIG_FIONA_PM_USB_HOST

// Static global holding fpow device entry returned from fpow_register()
extern struct fpow_component *g_usb_host_fpow_component_ptr;
extern int usb_host_register_with_fpow(void *private, struct fpow_component **fpow_dev);
extern void usb_host_unregister_with_fpow(struct fpow_component *fpow_dev);
#endif // CONFIG_FIONA_PM_USB_HOST

#endif // CONFIG_ARCH_FIONA

static void dump_stats(void)
{
#if 0
	printk("<1>interrupts = %d, ctrl pkts = %d, bulk pkts = %d, "
			"intr pkts = %d\n", num_interrupts, num_ctrl_pkts,
			num_bulk_pkts, num_intr_pkts);
#endif
}


static int
pehci_rh_control (struct usb_hcd *usb_hcd,u16  typeReq,
				 u16 wValue,u16 wIndex,char *buf,
				  u16		wLength);




/*----------------------------------------------------*/

#define PEHCI_REG_DUMP(register)  printk("<1>  0x%03X: %35s = 0x%08X\n", register, \
					 #register, readl(hcd->iobase + register));


static void
pehci_dump(phci_hcd *hcd)
{
	if (!hcd) {
		printk("<1>Null hcd pointer!\n");
		return;
	}
	
	printk("<1>PEHCI registers:\n");

	PEHCI_REG_DUMP(HC_SP_PORTROUTE1_REG);		/* 0x0C  Companion Port Route Register 1 */
	PEHCI_REG_DUMP(HC_SP_PORTROUTE2_REG);		/* 0x10	 Companion Port Route Register 1 */
	PEHCI_REG_DUMP(HC_USBCMD_REG);			/* 0x20	 USB Command Register */
	PEHCI_REG_DUMP(HC_USBSTS_REG);			/* 0x24	 USB Status Register */
	PEHCI_REG_DUMP(HC_INTERRUPT_REG_EHCI);		/* 0x28	 INterrupt Enable Register */
	PEHCI_REG_DUMP(HC_FRINDEX_REG);			/* 0x2C	 Frame Index Register */

	PEHCI_REG_DUMP(HC_CTRLDSSEGMENT_REG);		/* 0x30	 4G Segment Selector Register (Only for 64bit)*/
	PEHCI_REG_DUMP(HC_PERIODICBASE_REG);		/* 0x34	 Periodic List Base Address Register */
	PEHCI_REG_DUMP(HC_ASYNCLISTADDR_REG);		/* 0x38	 Next Asynchronous List register */
	PEHCI_REG_DUMP(HC_CONFIGFLAG_REG);		/* 0x60	 Conigured Flag  Register */

	PEHCI_REG_DUMP(HC_PORTSC1_REG);			/* 0x64	 Port Status Control for Port1 */
	PEHCI_REG_DUMP(HC_PORTSC2_REG);			/* 0x68	 Port Status Control for Port2 */
	PEHCI_REG_DUMP(HC_PORTSC3_REG);			/* 0x6C	 Port Status Control for Port3 */
	PEHCI_REG_DUMP(HC_PORTSC4_REG);			/* 0x70	 Port Status Control for Port4 */

	//ISO Transfer Registers
	PEHCI_REG_DUMP(HC_ISO_PTD_DONEMAP_REG);		/* 0x130 ISO PTD Done Map Register */
	PEHCI_REG_DUMP(HC_ISO_PTD_SKIPMAP_REG);		/* 0x134 ISO PTD Skip Map Register */
	PEHCI_REG_DUMP(HC_ISO_PTD_LASTPTD_REG);		/* 0x138 ISO PTD Last PTD Register */

	//INT Transfer Registers
	PEHCI_REG_DUMP(HC_INT_PTD_DONEMAP_REG);		/* 0x140 INT PTD Done Map Register */
	PEHCI_REG_DUMP(HC_INT_PTD_SKIPMAP_REG);		/* 0x144 INT PTD Skip Map Register */
	PEHCI_REG_DUMP(HC_INT_PTD_LASTPTD_REG);		/* 0x148 INT PTD Last PTD Register  */

	//ATL Transfer Registers
	PEHCI_REG_DUMP(HC_ATL_PTD_DONEMAP_REG);		/* 0x150 ATL PTD Last PTD Register  */
	PEHCI_REG_DUMP(HC_ATL_PTD_SKIPMAP_REG);		/* 0x154 ATL PTD Last PTD Register  */
	PEHCI_REG_DUMP(HC_ATL_PTD_LASTPTD_REG);		/* 0x158 ATL PTD Last PTD Register  */

	PEHCI_REG_DUMP(HC_ATL_PTDDONE_THRESHOLDCNT_REG);	/* 0x350 Revision Register (read only) */
	PEHCI_REG_DUMP(HC_ATL_PTDDONE_THRESHOLDTIMEOUT_REG);	/* 0x351 Control Register */

	//General Purpose Registers
	PEHCI_REG_DUMP(HC_HW_MODE_REG);			/* 0x300 H/W Mode Register  */
	PEHCI_REG_DUMP(HC_CHIP_ID_REG);			/* 0x304 Chip ID Register */
	PEHCI_REG_DUMP(HC_SCRATCH_REG);			/* 0x308 Scratch Register */
	PEHCI_REG_DUMP(HC_RESET_REG);			/* 0x30C HC Reset Register */

	//Interrupt Registers
	PEHCI_REG_DUMP(HC_INTERRUPT_REG);		/* 0x310 Interrupt Register */
	PEHCI_REG_DUMP(HC_INTENABLE_REG);		/* 0x314 Interrupt enable Register */
	PEHCI_REG_DUMP(HC_ISO_IRQ_MASK_OR_REG);		/* 0x318 ISO Mask OR Register */
	PEHCI_REG_DUMP(HC_INT_IRQ_MASK_OR_REG);		/* 0x31C INT Mask OR Register*/
	PEHCI_REG_DUMP(HC_ATL_IRQ_MASK_OR_REG);		/* 0x320 ATL Mask OR Register */
	PEHCI_REG_DUMP(HC_ISO_IRQ_MASK_AND_REG);	/* 0x324 ISO Mask AND Register */
	PEHCI_REG_DUMP(HC_INT_IRQ_MASK_AND_REG);	/* 0x328 INT Mask AND Register */
	PEHCI_REG_DUMP(HC_ATL_IRQ_MASK_AND_REG);	/* 0x32C ATL Mask AND Register */

	//RAM Registers
	PEHCI_REG_DUMP(HC_DMACONFIG_REG);		/* 0x330 DMA Config Register */
	PEHCI_REG_DUMP(HC_TRANS_COUNT_REG);		/* 0x334 Transfer Counter Register */
	PEHCI_REG_DUMP(HC_BUFFER_STAT_REG);		/* 0x338 Buffer status Register */
	PEHCI_REG_DUMP(HC_MEM_READ_REG);		/* 0x33C Memory Register */

	/*interrupt count and buffer status register*/

	PEHCI_REG_DUMP(HC_BUFFER_STATUS_REG);		/* 0x334 */
	PEHCI_REG_DUMP(HC_INT_THRESHOLD_REG);		/* 0x340 */

#if defined(CONFIG_ARCH_FIONA)
	PEHCI_REG_DUMP(HC_PORT1_CONTROL_REG);		/* 0x374 ISP176x Port 1 Control Register */
#endif
}


static void
pehci_bug(const char *fn_name, int line, phci_hcd *hcd)
{
	printk("<1>PEHCI BUG at %s:%d\n", fn_name, line);
	pehci_dump(hcd);
}




static void 
pehci_complete_device_removal(phci_hcd *hcd,struct ehci_qh *qh)
{

	td_ptd_map_t		*td_ptd_map;
	td_ptd_map_buff_t	*td_ptd_buff;

	if(qh->type == TD_PTD_BUFF_TYPE_ISTL){
		return;
	}

	td_ptd_buff = &td_ptd_map_buff[qh->type];
	td_ptd_map = &td_ptd_buff->map_list[qh->qtd_ptd_index];

	/*this flag should only be set when device is going*/
	td_ptd_map->state = TD_PTD_REMOVE;
	/*if nothing there*/
	if(list_empty(&qh->qtd_list)){
		if(td_ptd_map->state != TD_PTD_NEW){
			phci_hcd_release_td_ptd_index(qh);
		}
		qha_free(qha_cache, qh);
		qh = 0;
		return;
	}
	
	/*MUST not come down below this*/
	printk("<1>List must not have been empty!\n");

	return;
}




/*functions looks for the values in register
  specified in ptr, if register values masked
  with the mask and result is equal to done,
  operation is successful else fails with timeout*/
static int
pehci_hcd_handshake (
		phci_hcd *hcd,
		u32 ptr,
		u32 mask,
		u32 done,
		int usec)

{
	u32	result = 0;
	do {
		result = isp1762_reg_read32(hcd,ptr,result);
		if (result == ~(u32)0)		/* card removed */
			return -ENODEV;
		result &= mask;
		if (result == done)
			return 0;
		udelay (1);
		usec--;
	} while (usec > 0);
	return -ETIMEDOUT;
}


#ifndef MSEC_INT_BASED
/*schedule atl and interrupt tds,
only when we are not running on sof interrupt
*/
static void
pehci_hcd_td_ptd_submit_urb(
		phci_hcd	 *hcd,
		struct ehci_qh *qh,
		u8 bufftype)
{
	unsigned long flags;
	struct ehci_qtd *qtd = 0;
	struct urb *urb = 0;
	struct _isp1762_qha *qha = 0;
	u32 location  = 0;
	u32 skipmap = 0;
	u32 buffstatus = 0;
	u32 ormask = 0;
	u32 intormask = 0;
	u32 length = 0;
	struct list_head *head;

	td_ptd_map_t	*td_ptd_map;
	td_ptd_map_buff_t *ptd_map_buff;
	struct isp1762_mem_addr *mem_addr = 0;


	phci_print("%s: enter, buff type %d\n", __FUNCTION__, bufftype);

	spin_lock_irqsave(&hcd->lock, flags);
	
	ptd_map_buff = &td_ptd_map_buff[bufftype];
	qha = hcd->qha;
	switch(bufftype){
		case TD_PTD_BUFF_TYPE_ATL:
			skipmap = isp1762_reg_read32(hcd, hcd->regs.atltdskipmap, skipmap);
			ormask = isp1762_reg_read32(hcd, hcd->regs.atl_irq_mask_or, ormask);
 			//andmask = isp1762_reg_read32(hcd, hcd->regs.atl_irq_mask_and, andmask);
			break;
		case TD_PTD_BUFF_TYPE_INTL:
			skipmap = isp1762_reg_read32(hcd, hcd->regs.inttdskipmap, skipmap);
			intormask = isp1762_reg_read32(hcd, hcd->regs.int_irq_mask_or,intormask);
			break;
		default:
			skipmap = isp1762_reg_read32(hcd, hcd->regs.isotdskipmap, skipmap);
			break;

		}

	buffstatus = isp1762_reg_read32(hcd,hcd->regs.buffer_status,buffstatus);

	/*header, qtd, and urb of current transfer*/
      location = qh->qtd_ptd_index;
      td_ptd_map = &ptd_map_buff->map_list[location];

	
	if(!(qh->qh_state & QH_STATE_TAKE_NEXT)){
		pehci_check("qh will schdule from interrupt routine,map %x\n",td_ptd_map->ptd_bitmap);
        spin_unlock_irqrestore(&hcd->lock, flags);
        return;
	}
    head = &qh->qtd_list;
    qtd = list_entry(head->next, struct ehci_qtd, qtd_list);

	/*already scheduled??????, may be from interrupt*/
	if(!(qtd->state & QTD_STATE_NEW)){
		pehci_check("qtd already in, state %x\n",qtd->state);
        	spin_unlock_irqrestore(&hcd->lock, flags);
        	return;
	}

	qtd->state &= ~QTD_STATE_NEW;
	qtd->state |= QTD_STATE_SCHEDULED;

	qh->qh_state &= ~QH_STATE_TAKE_NEXT;
	/*take the first td*/
	td_ptd_map->qtd = qtd;
	/*take the urb*/
	urb = qtd->urb;


	ptd_map_buff->active_ptds++;

	/*trust the atl worker, at this location there wont be any td*/
	/*if this td is the last one*/
	if(qtd->state & QTD_STATE_LAST){
			qh->hw_current = cpu_to_le32(0);
	/*else update the hw_next of qh to the next td*/
	}else{
		qh->hw_current = qtd->hw_next;
	}
	memset(qha, 0, sizeof( isp1762_qha));

	pehci_check("td being scheduled is %p, device %d, map %x\n", qtd,urb->dev->devnum, td_ptd_map->ptd_bitmap);
	/*NEW, now need to get the memory for this transfer*/
	length = qtd->length;
	mem_addr = &qtd->mem_addr;
	phci_hcd_mem_alloc(length,mem_addr,0);
	if(length &&
		((mem_addr->phy_addr == 0 ) || (mem_addr->virt_addr == 0))){

		err("Cannot allocate %d bytes for the current td\n", length);

		/*should not happen*/
		/*can happen only when we exceed the limit of devices we support
		  MAX 4 mass storage at a time*/
        	spin_unlock_irqrestore(&hcd->lock, flags);
        	return;
	}

	phci_hcd_qha_from_qtd(hcd, qtd, qtd->urb, (void *) qha,
			td_ptd_map->ptd_ram_data_addr, qh);
	if(qh->type == TD_PTD_BUFF_TYPE_INTL){
		phci_hcd_qhint_schedule(hcd, qh,qtd,(isp1762_qhint *)qha,qtd->urb);
	}
	
	phci_print_qha(qha);

	 /*write qha into the header of the host controller*/
	isp1762_mem_write(hcd, td_ptd_map->ptd_header_addr,
			(u32 *)(qha),PHCI_QHA_LENGTH);


	/*if this is SETUP/OUT token , then need to write into the buffer*/
	/*length should be valid and supported by the ptd*/
	if(qtd->length && (qtd->length <= HC_ATL_PL_SIZE))
	switch(PTD_PID(qha->td_info2)){
		case OUT_PID:
		case SETUP_PID:
		pehci_hcd_mem_write(hcd,(u32)mem_addr->phy_addr,
				  (le32_to_cpu(qtd->hw_buf[0])),
				  length);
				 break;
		}

	  /*unskip the tds at this location*/
	  switch(bufftype){
	    case TD_PTD_BUFF_TYPE_ATL:
	  		skipmap &= ~td_ptd_map->ptd_bitmap;
			/*enable atl interrupts on donemap*/
	  		ormask |=  td_ptd_map->ptd_bitmap;
	  		isp1762_reg_write32(hcd, hcd->regs.atl_irq_mask_or, ormask);
	  		break;

	  	case TD_PTD_BUFF_TYPE_INTL:
	  		skipmap &= ~td_ptd_map->ptd_bitmap;
			intormask |= td_ptd_map->ptd_bitmap;
			isp1762_reg_write32(hcd, hcd->regs.int_irq_mask_or, intormask);
	  	   	break;

	  	case TD_PTD_BUFF_TYPE_ISTL:
	  		skipmap &= ~td_ptd_map->ptd_bitmap;
	  		isp1762_reg_write32(hcd, hcd->regs.isotdskipmap, skipmap);
	  		break;
	   	}

  /*if any new schedule, enable the atl buffer*/
 switch(bufftype){
	case TD_PTD_BUFF_TYPE_ATL:
		isp1762_reg_write32(hcd, hcd->regs.buffer_status, buffstatus | ATL_BUFFER);
       	isp1762_reg_write32(hcd, hcd->regs.atltdskipmap, skipmap);
		buffstatus |= ATL_BUFFER;
	break;
	case TD_PTD_BUFF_TYPE_INTL:
		isp1762_reg_write32(hcd, hcd->regs.buffer_status, buffstatus | INT_BUFFER);
		isp1762_reg_write32(hcd, hcd->regs.inttdskipmap, skipmap);
	break;
	case TD_PTD_BUFF_TYPE_ISTL:
		/*not supposed to be seen here*/
		isp1762_reg_write32(hcd, hcd->regs.buffer_status, buffstatus | ISO_BUFFER);
		break;
	}

exit:
	//spin_unlock(&hcd_data_lock);
	spin_unlock_irqrestore(&hcd->lock, flags);
	
	return;
}
#endif


/*schedule next (atl/int)tds and any pending tds*/
static void
pehci_hcd_schedule_pending_ptds(phci_hcd *hcd, u32 donemap, u8 bufftype, u16 only)

{

	struct ehci_qtd *qtd = 0;
	struct ehci_qh  *qh = 0;
	struct list_head *qtd_list = 0;
	struct _isp1762_qha allqha;
	struct _isp1762_qha *qha = 0;
	u32 mask = 0x1, index = 0;
	u32 location  = 0;
	u32 skipmap = 0;
	u32    newschedule = 0;
	u32 buffstatus = 0;
	u32 schedulemap = 0;
	u32 lasttd = 1, lastmap = 0;
	struct urb *urb = 0;
	urb_priv_t *urbpriv = 0;    
	int length = 0;
	u32 ormask = 0,andmask = 0;
	u32 intormask = 0;
	td_ptd_map_t	*td_ptd_map = 0;
	td_ptd_map_buff_t *ptd_map_buff;
	struct isp1762_mem_addr *mem_addr = 0;

	phci_print("%s: enter, bufftype %d\n", __FUNCTION__, bufftype);

	/*need to hold this lock if another interrupt is comming
	 for previously scheduled transfer, while scheduling new tds
	 */
	/*do we need this ...???*/	
	spin_lock(&hcd_data_lock);
	
	ptd_map_buff = &td_ptd_map_buff[bufftype];
	qha  = &allqha;
	switch(bufftype){
	case TD_PTD_BUFF_TYPE_ATL:
		skipmap = isp1762_reg_read32(hcd, hcd->regs.atltdskipmap, skipmap);
		rmb();
		ormask = isp1762_reg_read32(hcd, hcd->regs.atl_irq_mask_or, ormask);
 		andmask = isp1762_reg_read32(hcd, hcd->regs.atl_irq_mask_and, andmask);
  		break;
	case TD_PTD_BUFF_TYPE_INTL:
		skipmap = isp1762_reg_read32(hcd, hcd->regs.inttdskipmap, skipmap);
		/*read the interrupt mask registers*/
		intormask = isp1762_reg_read32(hcd, hcd->regs.int_irq_mask_or,intormask);
		break;
	default:
		err("bogus type of buffer\n");
		goto exit;
	}

 		
	buffstatus = isp1762_reg_read32(hcd,hcd->regs.buffer_status,buffstatus);
	/*td headers need attention*/
	schedulemap = donemap;
	while(schedulemap){
		index = schedulemap & mask;
		schedulemap &= ~mask;
		mask <<= 1;

  		if(!index){
  			location++;
  			continue;
  		}

   		td_ptd_map = &ptd_map_buff->map_list[location];
		/*	can happen if donemap comes after
			removal of the urb and associated tds
		*/
        if((td_ptd_map->state ==  TD_PTD_NEW) || 
                (td_ptd_map->state == TD_PTD_REMOVE)){
            qh = td_ptd_map->qh;
            pehci_check("should not come here, map %x,pending map %x\n", td_ptd_map->ptd_bitmap,
                    ptd_map_buff->pending_ptd_bitmap);

            pehci_check("buffer type %s\n", (bufftype == 0)?"ATL":"INTL");      
            donemap &= ~td_ptd_map->ptd_bitmap;
            /*clear the pending map*/
            ptd_map_buff->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;
            location++;
            continue;
  		}

  		/*no endpoint at this location*/
	   	if(!(td_ptd_map->qh)){
	   		 err("queue head can not be null here\n");
			/*move to the next location*/
			ptd_map_buff->pending_ptd_bitmap &=
					~td_ptd_map->ptd_bitmap;
        		location++;
			continue;
 		}

	   	/*current endpoint*/
		qh = td_ptd_map->qh;


		/*dont want to see this ??????*/
		if(!(skipmap & td_ptd_map->ptd_bitmap)){
			/*should not happen, if happening, then*/
			pehci_check("buffertype %d,td_ptd_map %x,skipnap %x\n",
				bufftype, td_ptd_map->ptd_bitmap,skipmap);
			lastmap = td_ptd_map->ptd_bitmap;
			donemap &= ~td_ptd_map->ptd_bitmap;
			ptd_map_buff->pending_ptd_bitmap &=
					~td_ptd_map->ptd_bitmap;
  			location++;
  			continue;
  		}

  	   	

	   
		/*if we processed all the tds in ths transfer*/
		if(td_ptd_map->lasttd){
      			err("should not show  map %x,qtd %p\n",
					td_ptd_map->ptd_bitmap,td_ptd_map->qtd);
			/*this can happen in case the transfer is not being
			 * procesed by the host , tho the transfer is there
			 * */
			qh->hw_current = cpu_to_le32(td_ptd_map->qtd);
			ptd_map_buff->pending_ptd_bitmap &=
					~td_ptd_map->ptd_bitmap;
			location++;
			continue;
		}


		/*if we have ptd that is going for reload*/
       		if((td_ptd_map->qtd) && (td_ptd_map->state & TD_PTD_RELOAD)){
        		warn("%s: reload td\n",__FUNCTION__);
        		td_ptd_map->state &= ~TD_PTD_RELOAD;
        		qtd = td_ptd_map->qtd;
        		goto loadtd;
      		}

	 	/* qh is there but no qtd so it means fresh transfer*/
		if((td_ptd_map->qh) && !(td_ptd_map->qtd)){
		  if(list_empty(&qh->qtd_list)){
			/*should not happen again, as it comes here
			  when it has td in its map
			*/	
			pehci_check("must not come here any more, td map %x\n",td_ptd_map->ptd_bitmap);
	       	  	/*this location is idle and can be free next time if
          	   	no new transfers are comming for this*/
                	donemap &= ~td_ptd_map->ptd_bitmap;
          	   	td_ptd_map->state |= TD_PTD_IDLE;
	       	  	ptd_map_buff->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;
          	  	location++;
        	  	continue;
        	}
			qtd_list = &qh->qtd_list;
			qtd = td_ptd_map->qtd =	list_entry(qtd_list->next, struct ehci_qtd, qtd_list);
			/*got the td, now goto reload*/
			goto loadtd;
	 }

		/*if there is already one qtd there in the transfer*/
		if(td_ptd_map->qtd){
		  	/*new schedule*/
			qtd = td_ptd_map->qtd;	
		}

    loadtd:

    	/*should not happen*/
		if(!qtd){
			err("this piece of code should not be executed\n");
			ptd_map_buff->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;
			location++;
			continue;
			}

		ptd_map_buff->active_ptds++;
		/*clear the pending map here*/
		ptd_map_buff->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;

		

		/*if this td is the last one*/
		 if(qtd->state & QTD_STATE_LAST){
		 	/*no qtd anymore*/
			qh->hw_current = cpu_to_le32(0);

			/*else update the hw_next of qh to the next td*/
    		}else{
            		qh->hw_current = qtd->hw_next;
		}

		if(location != qh->qtd_ptd_index){
			err("endpoint header location and scheduling "
					"information are not same\n");
		}

		/*next location*/
       		location++;
		/*found new transfer*/
       		newschedule = 1;
		/*take the urb*/
		urb = qtd->urb;
		/*sometimes we miss due to skipmap
		  so to make sure that we dont put again the 
		  same stuff
		*/
		if (!(qtd->state & QTD_STATE_NEW)) {
			err("Never Error: We should not put the same stuff\n");
			continue;
		}
		
		urbpriv = (urb_priv_t*)urb->hcpriv;
		urbpriv->timeout = 0;   

		/*no more new*/
		qtd->state &= ~QTD_STATE_NEW;
		qtd->state |= QTD_STATE_SCHEDULED;

			

		/*NEW, now need to get the memory for this transfer*/
		length = qtd->length;
		mem_addr = &qtd->mem_addr;
		phci_hcd_mem_alloc(length,mem_addr,0);
		if (length &&
			((mem_addr->phy_addr == 0 ) || (mem_addr->virt_addr == 0))) {

			err("Failed to allocate %d bytes for the current td\n",length);
			location++;
			continue;
		}

//		pehci_check("qtd being scheduled %p, device %d,map %x\n", qtd,
//				urb->dev->devnum,td_ptd_map->ptd_bitmap);

		memset(qha, 0, sizeof(isp1762_qha));
		/*convert qtd to qha*/
		phci_hcd_qha_from_qtd(hcd, qtd, qtd->urb, (void *)qha,
					td_ptd_map->ptd_ram_data_addr, qh);

		if (qh->type == TD_PTD_BUFF_TYPE_INTL) {
			phci_hcd_qhint_schedule(hcd, qh,qtd,(isp1762_qhint *)qha,qtd->urb);
		}


		length = PTD_XFERRED_LENGTH(qha->td_info1 >> 3);
		if(length > HC_ATL_PL_SIZE){
			err("Never Error: Bogus length,length %d(max %d)\n",
					qtd->length, HC_ATL_PL_SIZE);
		}

		/*write qha into the header of the host controller*/
		isp1762_mem_write(hcd, td_ptd_map->ptd_header_addr,
					(u32 *)(qha), PHCI_QHA_LENGTH);
		/*if this is SETUP/OUT token , then need to write into the buffer*/
		/*length should be valid*/
		if (qtd->length && (length <= HC_ATL_PL_SIZE))
			switch (PTD_PID(qha->td_info2)) {
			case OUT_PID:
			case SETUP_PID:
				pehci_hcd_mem_write(hcd, (u32)mem_addr->phy_addr,
						(le32_to_cpu(qtd->hw_buf[0])),length);
				break;
			}

		/*unskip the tds at this location*/
	  switch(bufftype){
	    case TD_PTD_BUFF_TYPE_ATL:
	  	skipmap &= ~td_ptd_map->ptd_bitmap;
	  	lastmap = td_ptd_map->ptd_bitmap;
		/*try to reduce the interrupts*/
	  	ormask |= td_ptd_map->ptd_bitmap;
	    	isp1762_reg_write32(hcd, hcd->regs.atl_irq_mask_or, ormask);
	  	break;

  		case TD_PTD_BUFF_TYPE_INTL:
	  	skipmap &= ~td_ptd_map->ptd_bitmap;
		lastmap = td_ptd_map->ptd_bitmap;
	  	intormask |= td_ptd_map->ptd_bitmap;
	  	isp1762_reg_write32(hcd, hcd->regs.int_irq_mask_or,intormask);
		break;

	  	case TD_PTD_BUFF_TYPE_ISTL:
	  		skipmap &= ~td_ptd_map->ptd_bitmap;
	  		isp1762_reg_write32(hcd, hcd->regs.isotdskipmap, skipmap);
	  		isp1762_reg_write32(hcd, hcd->regs.isotdlastmap,lasttd/*td_ptd_map->ptd_bitmap*/);
	  		break;
	   	}
	 

	}
  /*if any new schedule, enable the atl buffer*/

  if(newschedule){
	switch(bufftype){
	case TD_PTD_BUFF_TYPE_ATL:
       	isp1762_reg_write32(hcd, hcd->regs.buffer_status, buffstatus | ATL_BUFFER);
		/*i am comming here to only those tds that has to be scheduled*/
		/*so skip map must be in place*/
		if(skipmap & donemap){
			err("must be both ones compliment of each other\n");
			warn("problem, skipmap %x, donemap %x,\n",skipmap,donemap);
			
			}
		 skipmap &= ~donemap;
		isp1762_reg_write32(hcd, hcd->regs.atltdskipmap, skipmap);
	
  	break;
	case TD_PTD_BUFF_TYPE_INTL:
		isp1762_reg_write32(hcd, hcd->regs.buffer_status, buffstatus | INT_BUFFER);
	    skipmap &= ~donemap;
		isp1762_reg_write32(hcd, hcd->regs.inttdskipmap, skipmap);
	   	break;
	case TD_PTD_BUFF_TYPE_ISTL:
			isp1762_reg_write32(hcd, hcd->regs.buffer_status, buffstatus | ISO_BUFFER);
			break;
		}
    }

pehci_check("qtd being scheduled %p, device %d,map %x,skipmap %x\n",
		qtd, urb->dev->devnum,td_ptd_map->ptd_bitmap,
		skipmap);

exit:
	spin_unlock(&hcd_data_lock);
}


static void
pehci_hcd_qtd_schedule(phci_hcd *hcd, struct ehci_qtd *qtd, struct ehci_qh *qh,
			td_ptd_map_t	*td_ptd_map)
{
	struct urb		*urb;
	urb_priv_t		*urbpriv = 0;    
	u32			length;
	struct isp1762_mem_addr	*mem_addr = 0;
	struct _isp1762_qha	*qha, qhtemp;

	pehci_info("++ qtd_sched\n");
	/*redundant*/
	qha = &qhtemp;

	/*if this td is the last one*/
	if (qtd->state & QTD_STATE_LAST) {
		/*no qtd anymore*/
		qh->hw_current = cpu_to_le32(0);

	/* else update the hw_next of qh to the next td*/
	}else{
		qh->hw_current = qtd->hw_next;
	}

	urb = qtd->urb;
	urbpriv = (urb_priv_t*)urb->hcpriv;
	urbpriv->timeout = 0;       

	/*NEW, now need to get the memory for this transfer*/
	length = qtd->length;
	mem_addr = &qtd->mem_addr;
	phci_hcd_mem_alloc(length,mem_addr,0);
	if (length &&
		((mem_addr->phy_addr == 0 ) || (mem_addr->virt_addr == 0))) {
		err("Could not allocate %d bytes for the current td\n",length);
		return;
	}

//	pehci_info("qtd %p,dev %d,map %x\n", qtd,
//			urb->dev->devnum,td_ptd_map->ptd_bitmap);


	memset(qha, 0, sizeof(isp1762_qha));
	/*convert qtd to qha*/
	phci_hcd_qha_from_qtd(hcd, qtd, qtd->urb, (void *) qha,
			td_ptd_map->ptd_ram_data_addr, qh);

	if (qh->type == TD_PTD_BUFF_TYPE_INTL) {
		phci_hcd_qhint_schedule(hcd, qh,qtd,(isp1762_qhint *)qha,qtd->urb);
	}


	length = PTD_XFERRED_LENGTH(qha->td_info1 >> 3);
	if (length > HC_ATL_PL_SIZE) {
		err("Bogus length of %d (max %d)\n",qtd->length, HC_ATL_PL_SIZE);
	}

	/*write qha into the header of the host controller*/
	isp1762_mem_write(hcd, td_ptd_map->ptd_header_addr,
			(u32 *)(qha), PHCI_QHA_LENGTH);
	/*if this is SETUP/OUT token , then need to write into the buffer*/
	/*length should be valid*/
	if (qtd->length && ( length <= HC_ATL_PL_SIZE))
		switch (PTD_PID(qha->td_info2)) {
		case OUT_PID:
		case SETUP_PID:
//			pehci_info("qtd %p,device %d,map %x\n", qtd,
//					urb->dev->devnum,td_ptd_map->ptd_bitmap);
			pehci_hcd_mem_write(hcd, (u32)mem_addr->phy_addr,
					(le32_to_cpu(qtd->hw_buf[0])),length);
			break;
		}

	/*qtd is scheduled*/
	qtd->state &= ~QTD_STATE_NEW;
	qtd->state |= QTD_STATE_SCHEDULED;
	
//    pehci_info("-- %s: Exit\n",__FUNCTION__);
	return;
}



static void 
pehci_hcd_urb_complete(phci_hcd *hcd, struct ehci_qh *qh, struct urb *urb,
			td_ptd_map_t *td_ptd_map, struct pt_regs *regs)
{
	static u32 remove  = 0;
	urb_priv_t *urb_priv = (urb_priv_t *)urb->hcpriv;
	
//	pehci_info("++ %s: Entered\n",__FUNCTION__);
//	pehci_check("complete td %p\n", td_ptd_map->qtd);
    	urb_priv->timeout = 0;

	if((td_ptd_map->state == TD_PTD_REMOVE) ||
		(urb_priv->state == DELETE_URB) ||
		!HCD_IS_RUNNING(hcd->state))
		remove = 1;
		
	qh->qh_state = QH_STATE_COMPLETING;
	/*remove the done tds*/
	
	spin_lock(&hcd_data_lock);
	
	phci_hcd_urb_free_priv(hcd, urb_priv,qh);
	
	spin_unlock(&hcd_data_lock);
			
	urb_priv->timeout = 0;
	kfree(urb_priv);
	urb->hcpriv = 0;
	

	/*if normal completion???*/
	if(urb->status == -EINPROGRESS)
		urb->status = 0;


	spin_unlock(&hcd->lock);
	usb_hcd_giveback_urb (&hcd->usb_hcd, urb, regs);
	spin_lock(&hcd->lock);


	/* handle the remove case */
	if (remove) {
		remove = 0;
		if (list_empty(&qh->qtd_list)) {
			pehci_info("no transfers anymore, free the endpoint\n");
			phci_hcd_release_td_ptd_index(qh);
		}
	}
//    pehci_info("-- %s: Exit\n",__FUNCTION__);
}



/*update the error status of the td*/
static void
pehci_hcd_update_error_status(
				u32 ptdstatus,
				struct urb *urb)
{
	/*if ptd status is halted*/
	if(ptdstatus & PTD_STATUS_HALTED){
		if(ptdstatus & PTD_XACT_ERROR){
			/*transaction error results due to retry count goes to zero*/
			if(PTD_RETRY(ptdstatus)){
				/*halt the endpoint*/
				pehci_check("transaction error , retries %d\n", PTD_RETRY(ptdstatus));
				urb->status = -EPIPE;}
			else {
				pehci_check("transaction error , retries %d\n", PTD_RETRY(ptdstatus));
				/*protocol error*/
				urb->status = -EPROTO;}
			}
		else if(ptdstatus & PTD_BABBLE){
			pehci_check("babble error, qha %x\n", ptdstatus);
			/*babble error*/
			urb->status = -EOVERFLOW;}
	  	else if(PTD_RETRY(ptdstatus)){
			pehci_check("endpoint halted with retrie remaining %d\n",PTD_RETRY(ptdstatus));
			urb->status = -EPIPE;}
		else { /*unknown error*/
			pehci_check("protocol error, qha %x\n", ptdstatus);
			urb->status = -EPIPE;}

	    /*if halted need to recover*/
		if (urb->status == -EPIPE) {
		}
	  }
}


/*interrupt transfer handler*/
/*1. read done map
  2. read the ptd to see any errors
  3. copy the payload to and from
  4. update ehci td
  5. make new ptd if transfer there and earlier done
  6. schedule
*/
static void
pehci_hcd_intl_worker(
		phci_hcd *hcd,
		struct pt_regs *regs)
{
  int i = 0;
  u32 donemap = 0, donetoclear;
  u32 mask = 0x1,index = 0;
    u32 pendingmap = 0;
  u32 location = 0;
  u32 length = 0;
  u32 skipmap = 0;
  u32 ormask = 0;
  u32 usofstatus = 0;
  struct urb *urb;
  struct ehci_qtd *qtd = 0;
  struct ehci_qh *qh = 0;
  struct _isp1762_qhint *qhint = hcd->qhint;
  td_ptd_map_t		*td_ptd_map;
  td_ptd_map_buff_t	*ptd_map_buff;
  struct isp1762_mem_addr *mem_addr = 0;
    u32 dontschedule = 0;       

  ptd_map_buff = &(td_ptd_map_buff[TD_PTD_BUFF_TYPE_INTL]);
  pendingmap = ptd_map_buff->pending_ptd_bitmap;

   /*read the done map for interrupt transfers*/
   donetoclear = donemap = isp1762_reg_read32(hcd, hcd->regs.inttddonemap,donemap);
   if(donemap){
      phci_print("%s: donemap 0x%08x\n",__FUNCTION__,donemap);
      /*skip done tds*/
      skipmap = isp1762_reg_read32(hcd, hcd->regs.inttdskipmap, skipmap);
      skipmap |= donemap;
      isp1762_reg_write32(hcd, hcd->regs.inttdskipmap, skipmap);
      donemap |= 	pendingmap;
    }
/*if sof interrupt is enabled*/
#ifdef MSEC_INT_BASED
  	else{
    	 /*if there is something pending , put this transfer in*/
           if(ptd_map_buff->pending_ptd_bitmap){
	             pehci_hcd_schedule_pending_ptds(hcd,pendingmap,(u8)TD_PTD_BUFF_TYPE_INTL,1);
		}
            return;
        }
#else
    else{
        return;
    }

#endif


 ormask = isp1762_reg_read32(hcd, hcd->regs.int_irq_mask_or, ormask);
 /*process all the endpoints first those are done*/
 donetoclear = donemap;
 while(donetoclear){
  		/*index is the number of endpoints open currently*/
  		index = donetoclear & mask;
      	donetoclear &= ~mask;
  		mask <<= 1;
  		/*what if we are in the middle of schedule
  		  where nothing is done*/
  		if(!index ){
  			location++;
  			continue;
  			}

  		/*read our td_ptd_map*/
  		td_ptd_map = &ptd_map_buff->map_list[location];
		
		/*if this one is already in the removal*/
		if(td_ptd_map->state == TD_PTD_REMOVE ||
			td_ptd_map->state ==  TD_PTD_NEW){
			info("interrupt td is being removed\n");
			/*this will be handled by urb_remove*/
			/*if this is last urb no need to complete it again*/
			donemap &= ~td_ptd_map->ptd_bitmap;
			/*if there is something pending*/
			ptd_map_buff->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;
			continue;
		}


		/*if we found something already in*/
		if(!(skipmap & td_ptd_map->ptd_bitmap)){
			pehci_check("intr td_ptd_map %x,skipnap %x\n",
				td_ptd_map->ptd_bitmap,skipmap);
			donemap &= ~td_ptd_map->ptd_bitmap;
			/*in case pending*/
			ptd_map_buff->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap; ;
			location++;
  			continue;
  		}


		if(td_ptd_map->state ==  TD_PTD_NEW){
      		warn("interrupt not come here, map %x,location %d\n", td_ptd_map->ptd_bitmap,location);
      		donemap &= ~td_ptd_map->ptd_bitmap;
			/*in case pending*/
			ptd_map_buff->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;
			donemap &= ~td_ptd_map->ptd_bitmap;
			location++;
  			continue;
  		}

		/*move to the next schedule*/
		location ++;
		/*endpoint, td, urb and memory
		 * for current transfer*/
		qh = td_ptd_map->qh;
		qtd = td_ptd_map->qtd;
	    if(qtd->state & QTD_STATE_NEW){
	    	/*we need to schedule it*/
	    	goto schedule;
    	}




		
		urb = qtd->urb;
		mem_addr = &qtd->mem_addr;

		/*clear the irq mask for this transfer*/
		ormask &= ~td_ptd_map->ptd_bitmap;
		isp1762_reg_write32(hcd, hcd->regs.int_irq_mask_or, ormask);

		ptd_map_buff->active_ptds--;
	   	memset(qhint, 0, sizeof(struct _isp1762_qhint));

		/*read this ptd from the ram address,address is in the
		 td_ptd_map->ptd_header_addr*/
		isp1762_mem_read(hcd, td_ptd_map->ptd_header_addr,
				          (__u32 *)(qhint),PHCI_QHA_LENGTH);

		/*statuc of 8 uframes*/
		for(i=0;i<8;i++){
		/*take care of errors*/
        usofstatus = qhint->td_info5 >> (8 + i * 3);
       	switch(usofstatus & 0x7){
       		case INT_UNDERRUN:
       				phci_print("under run , %x\n", usofstatus);
       			break;
       			case INT_EXACT:
       				phci_print("transaction error, %x\n", usofstatus);
       			break;
       			case INT_BABBLE:
       				 phci_print("babble error, %x\n", usofstatus);
       			break;
       		}
       	 }

		if(urb->dev->speed != USB_SPEED_HIGH)
			/*length is 1K for full/low speed device*/
			length = PTD_XFERRED_NONHSLENGTH(qhint->td_info4);
		else
			/*length is 32K for high speed device*/
			length = PTD_XFERRED_LENGTH(qhint->td_info4);

		pehci_hcd_update_error_status(qhint->td_info4, urb);
		/*halted, need to finish all the transfer on this endpoint*/
		if(qhint->td_info4 & PTD_STATUS_HALTED){
			//td_ptd_map->lasttd = 1;
			qtd->state |= QTD_STATE_LAST;
			/*in case of halt, next transfer will start with toggle zero,
			 *USB speck, 5.8.5*/
			 qh->datatoggle = td_ptd_map->datatoggle = 0;
			donemap &= ~td_ptd_map->ptd_bitmap;
			ptd_map_buff->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;
			dontschedule = 1;
			 goto copylength;
		}


	copylength:
 	 /*preserve the current data toggle*/
	 qh->datatoggle = td_ptd_map->datatoggle = PTD_NEXTTOGGLE(qhint->td_info4);
	 /*copy data from the host*/
	 switch(PTD_PID(qhint->td_info2)){
			case IN_PID:
			 if(length && (length <= MAX_PTD_BUFFER_SIZE))
			 	/*do read only when there is somedata*/
			 	pehci_hcd_mem_read(hcd, (u32)mem_addr->phy_addr, (le32_to_cpu(qtd->hw_buf[0])),length);
			 case OUT_PID:
				 	urb->actual_length += length;
					qh->hw_current = qtd->hw_next;
					phci_hcd_mem_free(&qtd->mem_addr);
					qtd->state &= ~QTD_STATE_NEW;
					qtd->state |= QTD_STATE_DONE;
			 	break;
			}

	 
	        if(qtd->state & QTD_STATE_LAST){
	      		pehci_hcd_urb_complete(hcd,qh,urb,td_ptd_map,regs);
            if(dontschedule){ /*cleanup will start from drivers*/
                dontschedule = 0;
                continue;
            }
            /*take the next if in the queue*/
            if(!list_empty(&qh->qtd_list)){
                struct list_head *head;
                /*last td of previous urb*/
                head = &qh->qtd_list;
                qtd = list_entry(head->next,struct ehci_qtd, qtd_list);
                td_ptd_map->qtd = qtd;
                qh->hw_current = cpu_to_le32(qtd);
                qh->qh_state = QH_STATE_LINKED;

            }else{
                td_ptd_map->qtd = (struct ehci_qtd *) le32_to_cpu(0);
                qh->hw_current = cpu_to_le32(0);
                qh->qh_state = QH_STATE_TAKE_NEXT;
                donemap &= ~td_ptd_map->ptd_bitmap;
                ptd_map_buff->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;
                continue;
	      			}
	      	}



		 schedule:    
	  	{
	  		/*current td comes from qh->hw_current*/	
	  		ptd_map_buff->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;
	  		ormask |= td_ptd_map->ptd_bitmap;
			ptd_map_buff->active_ptds++;
			pehci_check("inter schedule next qtd %p, active tds %d\n", qtd,ptd_map_buff->active_ptds);
		  	pehci_hcd_qtd_schedule(hcd,qtd,qh,td_ptd_map);
	  	} 	 

	} /*end of while*/


/*clear all the tds inside this routine*/
  skipmap &= ~donemap;
  isp1762_reg_write32(hcd, hcd->regs.inttdskipmap, skipmap);
  ormask |= donemap;
  isp1762_reg_write32(hcd, hcd->regs.int_irq_mask_or,ormask);

}


/*atl(bulk/control) transfer handler*/
/*1. read done map
  2. read the ptd to see any errors
  3. copy the payload to and from
  4. update ehci td
  5. make new ptd if transfer there and earlier done
  6. schedule
*/
static void
pehci_hcd_atl_worker(
      phci_hcd *hcd,
      struct pt_regs *regs)
{

  u32 donemap = 0, donetoclear;
    u32 pendingmap = 0;
    u32 rl = 0; 
  u32 mask = 0x1,index = 0;
  u32 location = 0;
  u32 nakcount = 0;
  u32 active = 0;
  u32 length = 0;
  //u32 temp = 0;
  u32 skipmap = 0;
    u32 tempskipmap = 0;        
  u32 ormask=0;
  //char *tempbuffer = 0;
  struct urb *urb;
  struct ehci_qtd *qtd = 0;
  struct ehci_qh *qh;
  struct _isp1762_qha   atlqha;
  struct _isp1762_qha *qha;// = hcd->qha;
  td_ptd_map_t		*td_ptd_map;
  td_ptd_map_buff_t	*ptd_map_buff;
  urb_priv_t		*urbpriv = 0;	
  struct isp1762_mem_addr *mem_addr = 0;
    u32 dontschedule = 0;       

	ptd_map_buff = &(td_ptd_map_buff[TD_PTD_BUFF_TYPE_ATL]);
	pendingmap = ptd_map_buff->pending_ptd_bitmap;

#ifdef MSEC_INT_BASED
    /*running on skipmap rather donemap,
      some cases donemap may not be set 
      for complete transfer
     */                 
    skipmap = isp1762_reg_read32(hcd, hcd->regs.atltdskipmap, skipmap);
    tempskipmap = ~skipmap;
    if(tempskipmap){
        donemap = isp1762_reg_read32(hcd, hcd->regs.atltddonemap,donemap);
        skipmap |= donemap;
        isp1762_reg_write32(hcd, hcd->regs.atltdskipmap, skipmap);
        qha = &atlqha;
        donemap |= pendingmap;
        tempskipmap &= ~donemap;
    }

/*if sof interrupt enabled*/
	else{
		/*if there is something pending , put this transfer in*/
		if(pendingmap){
			pehci_info("scheduling pending map %x, "
					"active tds %d, pendingmap %x\n",
					ptd_map_buff->pending_ptd_bitmap,
					ptd_map_buff->active_ptds,
					pendingmap);
			pehci_hcd_schedule_pending_ptds(hcd, pendingmap,
					(u8) TD_PTD_BUFF_TYPE_ATL, 1);
		}
		return;
      	}
#else
    donemap = isp1762_reg_read32(hcd, hcd->regs.atltddonemap,donemap);
    if(donemap){
        skipmap = isp1762_reg_read32(hcd, hcd->regs.atltdskipmap, skipmap);
        skipmap |= donemap;
        isp1762_reg_write32(hcd, hcd->regs.atltdskipmap, skipmap);
        qha = &atlqha;
    }
	else{
		return;
	}
#endif

 /*read the interrupt mask registers*/
 ormask = isp1762_reg_read32(hcd, hcd->regs.atl_irq_mask_or, ormask);
 //andmask = isp1762_reg_read32(hcd, hcd->regs.atl_irq_mask_and, andmask);

 /*this map is used only to update and
  * scheduling for the tds who are not
  * complete. the tds those are complete
  * new schedule will happen from
  * td_ptd_submit_urb routine
  * */
 donetoclear = donemap;
    donetoclear |= tempskipmap; 
 /*process all the endpoints first those are done*/
 while(donetoclear){
  		/*index is the number of endpoint open currently*/
  		index = donetoclear & mask;
      	        donetoclear &= ~mask;
  		mask <<= 1;
  		/*what if we are in the middle of schedule
  		  where nothing is done
  		*/
  		if(!index ){
  			location++;
  			continue;
  			}

		/*read our td_ptd_map*/
  		td_ptd_map = &ptd_map_buff->map_list[location];

		/*urb is in remove*/
		if(td_ptd_map->state == TD_PTD_NEW ||
			td_ptd_map->state ==  TD_PTD_REMOVE){
            pehci_check("atl td is being removed,map %x, skipmap %x\n", td_ptd_map->ptd_bitmap, skipmap);
            pehci_check("temp skipmap %x, pendign map %x,done %x\n",tempskipmap, pendingmap,donemap);
			/*unlink urb will take care of this*/
			donemap &= ~td_ptd_map->ptd_bitmap;
			/*in case pending*/
			ptd_map_buff->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;
			location++;
			continue;
			}


        /*move to the next endpoint*/
        location ++;
        /*endpoint, td, urb and memory
         * for current endpoint*/
        qh = td_ptd_map->qh;
        qtd = td_ptd_map->qtd;
        if(!qh || !qtd)
        {
            err("Never Error:QH and QTD must not be zero\n");
            donemap &= ~td_ptd_map->ptd_bitmap;
            /*in case pending*/
            ptd_map_buff->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;            
  			continue;
  		}
#ifdef MSEC_INT_BASED           
        /*new td must be scheduled*/
        if((qtd->state & QTD_STATE_NEW) /*&&
                                          (pendingmap & td_ptd_map->ptd_bitmap)*/){
            /*this td will come here first time from
             *pending tds, so its qh->hw_current needs to
             * adjusted
             */
            qh->hw_current = QTD_NEXT (qtd->qtd_dma);
            goto schedule;
        }
#endif                  

		
		urb = qtd->urb;
		urbpriv = (urb_priv_t *)urb->hcpriv;
		mem_addr = &qtd->mem_addr;

#ifdef MSEC_INT_BASED
        /*check here for the td if its done*/
        if(donemap & td_ptd_map->ptd_bitmap){
            /*nothing to do*/
            ;
        }else{
            /*if td is not done, lets check how long 
              its been scheduled
             */
            if(tempskipmap & td_ptd_map->ptd_bitmap){
                /*i will give 20 msec to complete*/
                if(urbpriv->timeout < 20){
                    urbpriv->timeout++; 
                    continue;
                }
                urbpriv->timeout++;
                /*otherwise check its status*/
            }

        }
#endif          
        memset(qha, 0, sizeof(struct _isp1762_qha));

        /*read this ptd from the ram address,address is in the
          td_ptd_map->ptd_header_addr*/
        isp1762_mem_read(hcd, td_ptd_map->ptd_header_addr,(u32 *)(qha),PHCI_QHA_LENGTH);

#ifdef MSEC_INT_BASED   
        /*since we are running on skipmap
          tds will be checked for completion state
         */
        if((qha->td_info1 & QHA_VALID)){

//            pehci_check("pendign map %x, donemap %x, tempskipmap %x\n",
//                    pendingmap,donemap, tempskipmap);
            /*this could be one of the unprotected urbs, clear it*/
            ptd_map_buff->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;
            /*here also we need to increment the tds timeout count*/
            urbpriv->timeout++;
            continue;
        }else{
            /*this td is going to be done,
              this td could be the one un-skipped but no donemap or
              maybe it could be one of those where we get unprotected urbs,
              so checking against tempskipmap may not give us correct td
             */

            skipmap |= td_ptd_map->ptd_bitmap;
            isp1762_reg_write32(hcd, hcd->regs.atltdskipmap,
                    skipmap);

            /*of course this is going to be as good
              as td that is done and donemap is set
              also skipmap is set
             */                 
            donemap |= td_ptd_map->ptd_bitmap;
        }
#endif          


        /*clear the corrosponding mask register*/
        ormask &= ~td_ptd_map->ptd_bitmap;
        isp1762_reg_write32(hcd, hcd->regs.atl_irq_mask_or,ormask);

        ptd_map_buff->active_ptds--;

        urbpriv->timeout = 0;

        /*take care of errors*/
        pehci_hcd_update_error_status(qha->td_info4, urb);
        /*halted, need to finish all the transfer on this endpoint*/
        if(qha->td_info4 & PTD_STATUS_HALTED){
            qtd->state |= QTD_STATE_LAST;

            donemap &= ~td_ptd_map->ptd_bitmap; 
            /*in case pending*/
            ptd_map_buff->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;
            /*in case of halt, next transfer will start with toggle
              zero,USB speck, 5.8.5*/
            qh->datatoggle =  td_ptd_map->datatoggle = 0;
            /*cleanup the ping*/
            qh->ping = 0;
            /*force cleanup after this*/
            dontschedule = 1;
            goto copylength;
        }



        /*read the reload count*/
        rl = (qha->td_info3 >> 23);
        rl &= 0xf;



        /*if there is a transaction error and the status is not halted,
         * process whatever the length we got.if the length is what we
         * expected complete the transfer*/
        if((qha->td_info4 & PTD_XACT_ERROR) &&
                !(qha->td_info4 & PTD_STATUS_HALTED) &&
                (qha->td_info4 & QHA_ACTIVE)){

            if(PTD_XFERRED_LENGTH(qha->td_info4) == qtd->length){
                ;/*nothing to do its fake*/}
            else{

                pehci_info("xact error, info1 0x%08x,info4 0x%08x\n", qha->td_info1,qha->td_info4);

                /*if this is the case then we need to 
                  resubmit the td again*/
                qha->td_info1 |= QHA_VALID;
                skipmap &= ~td_ptd_map->ptd_bitmap;
                ormask |= td_ptd_map->ptd_bitmap;
                donemap &= ~td_ptd_map->ptd_bitmap;

                /*set the retry count to 3 again*/
                qha->td_info4 |= (rl << 19);
                /*set the active bit, if cleared, will be cleared if we have some length*/
                qha->td_info4 |= QHA_ACTIVE;
                /*clear the xact error*/
                qha->td_info4 &= ~PTD_XACT_ERROR;
                isp1762_reg_write32(hcd, hcd->regs.atl_irq_mask_or,ormask);

                /*copy back into the header, payload is already
                 * present no need to write again
                 */
                isp1762_mem_write(hcd, td_ptd_map->ptd_header_addr,(u32*)(qha),PHCI_QHA_LENGTH);
                /*unskip this td*/
                isp1762_reg_write32(hcd, hcd->regs.atltdskipmap, skipmap);
                continue;
		}
		goto copylength;

		}

		/*check for the nak count and active condition
		 * to reload the ptd if needed*/
		nakcount = qha->td_info4 >> 19;
		nakcount &= 0xf;
		active  = qha->td_info4 & QHA_ACTIVE;
    	        /*if nak count is zero and active bit is set , it
		 *means that device is naking and need to reload
		 *the same td*/
		if(!nakcount && active){
     	         pehci_check("%s: ptd is going for reload,length %d\n",
				__FUNCTION__,length);
     	 	/*make this td valid*/
		qha->td_info1 |= QHA_VALID;
		donemap &= ~td_ptd_map->ptd_bitmap;
		/*just like fresh td*/
        /*set the retry count to 3 again*/
        qha->td_info4 |= (rl << 19);
		ptd_map_buff->active_ptds++;
		skipmap &= ~td_ptd_map->ptd_bitmap;
      	  	ormask |= td_ptd_map->ptd_bitmap;
      	  	isp1762_reg_write32(hcd, hcd->regs.atl_irq_mask_or,ormask);
      	  	/*copy back into the header, payload is already
      	   	* present no need to write again */
      	  	isp1762_mem_write(hcd, td_ptd_map->ptd_header_addr,
  				          (u32 *)(qha),PHCI_QHA_LENGTH);
      	  	/*unskip this td*/
		  isp1762_reg_write32(hcd, hcd->regs.atltdskipmap, skipmap);
	      continue;
		}


	copylength:
 	/*read the length transferred*/
    length = PTD_XFERRED_LENGTH(qha->td_info4);

	/*short complete in case of BULK only*/
 	if((length < qtd->length) &&
		usb_pipebulk(urb->pipe)){
 		/*if current ptd is not able to fetech enough data as
 		 * been asked then device has no data, so complete this transfer
 		 * */
        if((urb->transfer_flags & URB_SHORT_NOT_OK)){
            pehci_check("short read, length %d(expected %d)\n", 
                    length, qtd->length);       
            urb->status = -EREMOTEIO;
            donemap &= ~td_ptd_map->ptd_bitmap;
            ptd_map_buff->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;
            /*force the cleanup from here*/
            dontschedule = 1;
        }
 	       	   qtd->state |= QTD_STATE_LAST;
		 /*if this is the only td, donemap will be cleared
		   at completion, otherwise take the next one
		*/
		//donemap &= ~td_ptd_map->ptd_bitmap;
 	  }
	
   /*preserve the current data toggle*/
    qh->datatoggle =td_ptd_map->datatoggle = PTD_NEXTTOGGLE(qha->td_info4);
    qh->ping = PTD_PING_STATE(qha->td_info4);   
	/*copy data from*/
	switch(PTD_PID(qha->td_info2)){
		case IN_PID:
            qh->ping = 0;
			/*do read only when there is some data*/
			if(length && (length <= HC_ATL_PL_SIZE)){
				pehci_hcd_mem_read(hcd, (u32)mem_addr->phy_addr, (le32_to_cpu(qtd->hw_buf[0])),length);
			}


		 case OUT_PID:
                   urb->actual_length += length;
                   qh->hw_current = qtd->hw_next;
                   phci_hcd_mem_free(&qtd->mem_addr);
		   qtd->state |= QTD_STATE_DONE;
			
			break;
		 case SETUP_PID:
			qh->hw_current = qtd->hw_next;
			phci_hcd_mem_free(&qtd->mem_addr);
			qtd->state |= QTD_STATE_DONE;
		 	break;
		}


	      if(qtd->state & QTD_STATE_LAST){
	      		pehci_hcd_urb_complete(hcd,qh,urb,td_ptd_map,regs);
        if(dontschedule){ /*cleanup will start from drivers*/
            dontschedule = 0;
            /*so that we can take next one*/
            qh->qh_state = QH_STATE_TAKE_NEXT;                                  
				continue;
	      			}
        /*take the next if in the queue*/
        if(!list_empty(&qh->qtd_list)){
            struct list_head *head;
            /*last td of previous urb*/
            head = &qh->qtd_list;
            qtd = list_entry(head->next,struct ehci_qtd, qtd_list);
            td_ptd_map->qtd = qtd;
            qh->hw_current = cpu_to_le32(qtd);
            qh->qh_state = QH_STATE_LINKED;

        }else{
            td_ptd_map->qtd = (struct ehci_qtd *) le32_to_cpu(0);
            qh->hw_current = cpu_to_le32(0);
            qh->qh_state = QH_STATE_TAKE_NEXT;
            donemap &= ~td_ptd_map->ptd_bitmap;
            ptd_map_buff->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;
            continue;
        }
    }

	  schedule:    
	  	{
	  		/*current td comes from qh->hw_current*/	
	  		ptd_map_buff->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;
	  		td_ptd_map->qtd = (struct ehci_qtd *)(le32_to_cpu(qh->hw_current));
			qtd = td_ptd_map->qtd;
			ormask |= td_ptd_map->ptd_bitmap;
			ptd_map_buff->active_ptds++;
			pehci_hcd_qtd_schedule(hcd,qtd,qh,td_ptd_map);
	  	} 	 

   	} /*end of while*/


  /*clear all the tds inside this routine*/
  skipmap &= ~donemap;
  isp1762_reg_write32(hcd, hcd->regs.atltdskipmap, skipmap);
  ormask |= donemap;
  isp1762_reg_write32(hcd, hcd->regs.atl_irq_mask_or,ormask);
//pehci_info("-- %s: Exit\n",__FUNCTION__);
}



/*--------------------------------------------------------*
     root hub functions
 *--------------------------------------------------------*/

/*return root hub descriptor, can not fail*/
static void
pehci_hub_descriptor (
	phci_hcd			*hcd,
	struct usb_hub_descriptor	*desc)
{
	u32		ports = 0;
	u16		temp = 0;
    pehci_info("++ %s: Entered\n",__FUNCTION__);
	ports = isp1762_reg_read32(hcd, hcd->regs.hcsparams, ports);
	ports = ports & 0xf;
	pehci_info("%s: number of ports %d\n", __FUNCTION__, ports);

	desc->bDescriptorType = 0x29;
	desc->bPwrOn2PwrGood = 10;	/* FIXME: f(system power) */

	desc->bHubContrCurrent = 0;

	desc->bNbrPorts = ports;
	temp = 1 + (ports / 8);
	desc->bDescLength = 7 + 2 * temp;
	/* two bitmaps:  ports removable, and usb 1.0 legacy PortPwrCtrlMask */

	memset (&desc->DeviceRemovable[0], 0, temp);
	memset (&desc->PortPwrCtrlMask[temp], 0xff, temp);

	temp = 0x0008;			/* per-port overcurrent reporting */
	//if (HCS_PPC (ehci->hcs_params))
	temp |= 0x0001;		/* per-port power control */
	//if (HCS_INDICATOR (ehci->hcs_params))
	temp |= 0x0080;		/* per-port indicators (LEDs) */
	desc->wHubCharacteristics = cpu_to_le16 (temp);
    pehci_info("-- %s: Exit\n",__FUNCTION__);
}


/*after reset on root hub,
 * device high speed or non-high speed
 * */
static int
phci_check_reset_complete (
	phci_hcd	*hcd,
	int		index,
	int		port_status)
{
	phci_print("check reset complete\n");
	if (!(port_status & PORT_CONNECT)) {
		hcd->reset_done [index] = 0;
		return port_status;
		}

	/* if reset finished and it's still not enabled -- handoff */
	if (!(port_status & PORT_PE)) {
		 printk ("port %d full speed --> companion\n",
					index + 1);
		port_status |= PORT_OWNER;
		isp1762_reg_write32(hcd,hcd->regs.ports[index],port_status);

	} else
		phci_print ("port %d high speed\n", index + 1);

	return port_status;

}



/*----------------------------------------------*
  host controller initialization, removal functions
 *----------------------------------------------*/


/*initialize all three buffer(iso/atl/int) type headers*/
static void
pehci_hcd_init_map_buffers(phci_hcd	*phci) {
	td_ptd_map_buff_t	*ptd_map_buff;
	u8				buff_type, ptd_index;
	u32				bitmap;

//    pehci_info("++ %s: Entered\n",__FUNCTION__);
  	phci_print("phci_init_map_buffers(phci = 0x%p)\n",phci);
	/* initialize for each buffer type */
	for(buff_type = 0; buff_type < TD_PTD_TOTAL_BUFF_TYPES; buff_type++)
	{
		ptd_map_buff = &(td_ptd_map_buff[buff_type]);
		ptd_map_buff->buffer_type = buff_type;
		ptd_map_buff->active_ptds = 0;
		ptd_map_buff->total_ptds = 0;
		/*each bufer type can have atleast 32 ptds*/
		ptd_map_buff->max_ptds = 32;
		ptd_map_buff->active_ptd_bitmap = 0;
		/*everything skipped*/
	//	ptd_map_buff->skip_ptd_bitmap = 0xFFFFFFFF;
		/*nothing is pending*/
	    ptd_map_buff->pending_ptd_bitmap = 0x00000000;

	//	bitmap = 0x00000001;
	//	for(ptd_index=0;ptd_index<ptd_map_buff->max_ptds;ptd_index++) {
	//		ptd_map_buff->active_ptd_bitmask |= bitmap;
	//		bitmap <<= 1;
	//	}
	       /* For each ptd index of this buffer, set the fiedls */
		bitmap = 0x00000001;
		for(ptd_index = 0; ptd_index < TD_PTD_MAX_BUFF_TDS; ptd_index++)		{
		  /*datatoggle zero*/
		  ptd_map_buff->map_list[ptd_index].datatoggle = 0;
		  /*td state is not used*/
		  ptd_map_buff->map_list[ptd_index].state = TD_PTD_NEW;
		  /*no endpoint, no qtd*/
		  ptd_map_buff->map_list[ptd_index].qh = NULL;
		  ptd_map_buff->map_list[ptd_index].qtd = NULL;
		  ptd_map_buff->map_list[ptd_index].ptd_header_addr = 0xFFFF;
	       }/* for( ptd_index */

	}	/* for(buff_type */
//    pehci_info("-- %s: Exit\n",__FUNCTION__);
} /* phci_init_map_buffers */


/*put the host controller into operational mode
* called phci_hcd_start routine,
* return 0, success else
* timeout, fails*/


static int
pehci_hcd_start_controller(
				phci_hcd *hcd)
{
	u32 command = 0;
	int retval = 0;
//    pehci_info("++ %s: Entered\n",__FUNCTION__);
	/*initialize the host controller*/
	command |= CMD_RUN;
	isp1762_reg_write32(hcd,hcd->regs.command,command);

	command &=0;
	command  =  isp1762_reg_read32(hcd,hcd->regs.command,command);
	/*should be in operation in 1000 usecs*/
	if((retval = pehci_hcd_handshake(hcd,hcd->regs.command, CMD_RUN, CMD_RUN,1000))){
		err("Host is not up(CMD_RUN) in 1000 usecs\n");
		return retval;
	}

     /*put the host controller to ehci mode*/
	 command &=0;
	 command |= 1;
	 isp1762_reg_write32(hcd,hcd->regs.configflag, command);
     /*check if ehci mode switching is correct or not*/
	  if((retval = pehci_hcd_handshake(hcd, hcd->regs.configflag,1,1,100))){
	  	 err("Host is not into ehci mode in 100 usecs\n");
		 return retval;
	  }
//    pehci_info("-- %s: Exit\n",__FUNCTION__);
	 return retval;
}


/*enable the interrupts
 *called phci_1762_start routine
 * return void*/
static void
pehci_hcd_enable_interrupts(
					phci_hcd *hcd)
{
  u32  temp = 0;
//    pehci_info("++ %s: Entered\n",__FUNCTION__);
  /*disable the interrupt source*/
  temp &= 0;
  /*clear all the interrupts that may be there */
  temp |= INTR_ENABLE_MASK;
  isp1762_reg_write32(hcd, hcd->regs.interrupt, temp);

  /*enable interrupts*/
  temp = 0;
  temp |= INTR_ENABLE_MASK;
  phci_print("%s: enabled mask 0x%08x\n",__FUNCTION__, temp);
  isp1762_reg_write32(hcd,hcd->regs.interruptenable, temp);

  	/*enable the global interrupt*/
  	temp &=0;
   	temp = isp1762_reg_read32(hcd, hcd->regs.hwmodecontrol, temp);
	temp |= 0x101;/*enable the global interrupt*/
	isp1762_reg_write32(hcd, hcd->regs.hwmodecontrol, temp);
	/*maximum rate is one msec*/
	//isp1762_reg_write32(hcd, hcd->regs.interruptthreshold, 0x08000200);
	/*enable the atl interrupts OR and AND mask*/
	temp = 0;
	isp1762_reg_write32(hcd, hcd->regs.atl_irq_mask_and, temp);
	temp = 0;
	isp1762_reg_write32(hcd,hcd->regs.atl_irq_mask_or, temp);
   	temp = 0;
	isp1762_reg_write32(hcd, hcd->regs.int_irq_mask_and, temp);
	temp = 0x0;
	isp1762_reg_write32(hcd,hcd->regs.int_irq_mask_or, temp);
	temp = 0;
	isp1762_reg_write32(hcd, hcd->regs.iso_irq_mask_and, temp);
	temp = 0xffffffff;
	isp1762_reg_write32(hcd,hcd->regs.iso_irq_mask_or, temp);
//    pehci_info("-- %s: Exit\n",__FUNCTION__);
}


/*initialize the host controller register map from Isp 1762 to EHCI */
static void
pehci_hcd_init_reg(phci_hcd *hcd)
{
//    pehci_info("++ %s: Entered\n",__FUNCTION__);
    /* scratch pad for the test*/
	hcd->regs.scratch = HC_SCRATCH_REG  ;

	/*make a copy of our interrupt locations*/
	hcd->regs.command = HC_USBCMD_REG	;
	hcd->regs.usbstatus = HC_USBSTS_REG;
	hcd->regs.usbinterrupt = HC_INTERRUPT_REG_EHCI;

	hcd->regs.hcsparams = HC_SPARAMS_REG;
	hcd->regs.frameindex = HC_FRINDEX_REG;

	/*transfer specific registers*/
	hcd->regs.hwmodecontrol = HC_HW_MODE_REG;
	hcd->regs.interrupt = HC_INTERRUPT_REG;
	hcd->regs.interruptenable = HC_INTENABLE_REG;
	hcd->regs.atl_irq_mask_and = HC_ATL_IRQ_MASK_AND_REG;
	hcd->regs.atl_irq_mask_or = HC_ATL_IRQ_MASK_OR_REG;

	hcd->regs.int_irq_mask_and = HC_INT_IRQ_MASK_AND_REG;
	hcd->regs.int_irq_mask_or = HC_INT_IRQ_MASK_OR_REG;
	hcd->regs.iso_irq_mask_and = HC_ISO_IRQ_MASK_AND_REG;
	hcd->regs.iso_irq_mask_or = HC_ISO_IRQ_MASK_OR_REG;
    hcd->regs.buffer_status = HC_BUFFER_STATUS_REG;
   	hcd->regs.interruptthreshold = HC_INT_THRESHOLD_REG;
	/*initialization specific*/
	hcd->regs.reset = HC_RESET_REG;
	hcd->regs.configflag = HC_CONFIGFLAG_REG;
#if defined(CONFIG_ARCH_FIONA)
	hcd->regs.port1ctl		= HC_PORT1_CONTROL_REG;
#endif
	hcd->regs.ports[0]	= HC_PORTSC1_REG;
    hcd->regs.ports[1]      = 0;        /*port1,port2,port3 status reg are removed */
    hcd->regs.ports[2]      = 0;
    hcd->regs.ports[3]      = 0;
	/*transfer registers*/
	hcd->regs.isotddonemap = HC_ISO_PTD_DONEMAP_REG;
	hcd->regs.isotdskipmap = HC_ISO_PTD_SKIPMAP_REG;
	hcd->regs.isotdlastmap = HC_ISO_PTD_LASTPTD_REG;

	hcd->regs.inttddonemap = HC_INT_PTD_DONEMAP_REG;

	hcd->regs.inttdskipmap = HC_INT_PTD_SKIPMAP_REG;
	hcd->regs.inttdlastmap = HC_INT_PTD_LASTPTD_REG;

	hcd->regs.atltddonemap = HC_ATL_PTD_DONEMAP_REG;
	hcd->regs.atltdskipmap = HC_ATL_PTD_SKIPMAP_REG;
	hcd->regs.atltdlastmap = HC_ATL_PTD_LASTPTD_REG;

	/*initialize the host controller initial values*/
	/*disable all the buffer*/
	isp1762_reg_write32(hcd, hcd->regs.buffer_status,0);
	/*skip all the transfers*/
	isp1762_reg_write32(hcd, hcd->regs.atltdskipmap, NO_TRANSFER_ACTIVE);
	isp1762_reg_write32(hcd, hcd->regs.inttdskipmap, NO_TRANSFER_ACTIVE);
	isp1762_reg_write32(hcd, hcd->regs.isotdskipmap, NO_TRANSFER_ACTIVE);
	/*clear done map*/
	isp1762_reg_write32(hcd, hcd->regs.atltddonemap, ~NO_TRANSFER_ACTIVE);
	isp1762_reg_write32(hcd, hcd->regs.inttddonemap, ~NO_TRANSFER_ACTIVE);
	isp1762_reg_write32(hcd, hcd->regs.isotddonemap, ~NO_TRANSFER_ACTIVE);
//    pehci_info("-- %s: Exit\n",__FUNCTION__);

}



static void pehci_hcd_release(struct usb_bus *bus)
{
	struct usb_hcd *hcd = bus->hcpriv;

	if (hcd)
		hcd->driver->hcd_free(hcd);
}



/*---------------------------------------------------
*  Interrupt request function
-----------------------------------------------------*/

/*tasklet function*/
static void
pehci_hcd_tasklet(unsigned long arg)
{
	phci_hcd *hcd = (phci_hcd *) arg;
	//struct isp1762_dev *dev  = hcd->dev;
	//pehci_hcd_iso_worker(hcd,NULL);
	pehci_hcd_atl_worker(hcd,NULL);
	pehci_hcd_intl_worker(hcd,NULL);
	return;

}

static void
pehci_interrupt_handler(phci_hcd *hcd, struct pt_regs *regs)
{
   spin_lock(&hcd->lock);
    pehci_hcd_atl_worker(hcd,regs);
    pehci_hcd_intl_worker(hcd,regs);
 spin_unlock(&hcd->lock);
    return;
}




irqreturn_t
pehci_hcd_irq1(struct usb_hcd *hcd, struct pt_regs *regs)
{
	info("%s: doing nothing\n",__FUNCTION__);
	return IRQ_NONE;

}


/*isr routine*/
irqreturn_t
pehci_hcd_irq(int irq, void *__irq_data, struct pt_regs *regs) 
{
    u32   intr = 0,intr_mask = 0;
    int work = 0;
    phci_hcd    *pehci_hcd;

    struct usb_hcd *usb_hcd = (struct usb_hcd *)__irq_data;
    if(!(usb_hcd->state & USB_STATE_READY)){
    	  info("interrupt handler state not ready yet\n");
            return IRQ_NONE;
    	}

    num_interrupts++;
    if ((num_interrupts & 0x0fff) == 0)
        dump_stats();
        
   /*our host*/	
   pehci_hcd = usb_hcd_to_pehci_hcd(usb_hcd);
   
   intr = isp1762_reg_read32(pehci_hcd, HC_INTERRUPT_REG,intr);
  /*clear the interrupt*/	
   isp1762_reg_write32(pehci_hcd, HC_INTERRUPT_REG,intr);	
   intr_mask = isp1762_reg_read32(pehci_hcd, HC_INTENABLE_REG,intr_mask);

  if((intr & intr_mask) == 0) /*shared interrupt*/
		return IRQ_NONE;

 if(intr & HC_OTG_INTERRUPT) /*no otg in this host*/
			return IRQ_NONE;

	/*FIXME,just extra check*/
	if(atomic_read(&pehci_hcd->nuofsofs)){
		  return IRQ_HANDLED;	
		}

	
	atomic_inc(&pehci_hcd->nuofsofs);
	

   usb_hcd->saw_irq = 1;	  	
   #ifdef MSEC_INT_BASED
		work = 1;
	  
   #else
      if(intr & (HC_MSEC_INT & INTR_ENABLE_MASK))
      	  work = 1;	//	phci_iso_worker(hcd);

       if(intr & (HC_ATL_INT & INTR_ENABLE_MASK)){
	  spin_lock(&pehci_hcd->lock);
		pehci_hcd_atl_worker(pehci_hcd, regs);
	  spin_unlock(&pehci_hcd->lock); 	
       	  work = 0;	//phci_atl_worker(hcd);
	}

       if(intr & (HC_INTL_INT & INTR_ENABLE_MASK)){
		spin_lock(&pehci_hcd->lock);
			pehci_hcd_intl_worker(pehci_hcd, regs);
		spin_unlock(&pehci_hcd->lock);
     	    work = 0;	//phci_intl_worker(hcd);
	}
   #endif
  
	if(work)
		//tasklet_schedule(&pehci_hcd->tasklet);
		pehci_interrupt_handler(pehci_hcd, regs);
		
	atomic_dec(&pehci_hcd->nuofsofs);
	return IRQ_HANDLED;
}

 

/*reset the host controller
*called phci_hcd_start routine
*return 0, success else
*timeout, fails*/
static int
pehci_hcd_reset(
		struct usb_hcd *usb_hcd)
{
	u32		command = 0;
    u32		temp = 0;
    phci_hcd *hcd = usb_hcd_to_pehci_hcd(usb_hcd);

#if 0
	/*reset the host controller and the cpu interface registers */
	temp &= 0;
	temp |= 1;
	isp1762_reg_write32(hcd,hcd->regs.reset, temp);
#endif

	/*reset the ehci controller registers*/
	temp = 0;
	temp |= (1<<1);
	isp1762_reg_write32(hcd,hcd->regs.reset, temp);

	/*read the command register */
	command = isp1762_reg_read32(hcd,hcd->regs.command,command);

	command |= CMD_RESET;
	/*write back and wait for, 250 msec */
	isp1762_reg_write32(hcd,hcd->regs.command,command);
	/*wait for maximum 250 msecs*/
	temp = pehci_hcd_handshake(hcd,HC_USBCMD_REG, CMD_RESET,0, 250*1000);

#ifdef CONFIG_ARCH_FIONA
	/* Since the HcScratch register gets cleared during this reset
	 * routine, update with the running signature since we're getting
	 * ready to run.
	 */
	
	isp1762_reg_write32(hcd, HC_SCRATCH_REG, USB_RUNNING_SIGNATURE);
#endif

	return temp;
}
                                                                                                                                     

/*host controller initialize routine,
 *called by pehci_hcd_probe
 * */
static int
pehci_hcd_start(struct usb_hcd *usb_hcd)
{
	struct usb_device	*udev;
	struct usb_bus		*bus;
	//struct ehci_qh *qh = 0;
	int			retval;
	int          count = 0;
	phci_hcd		*pehci_hcd = NULL;

    pehci_info("++ %s: Entered\n",__FUNCTION__);
	pehci_hcd = usb_hcd_to_pehci_hcd(usb_hcd);
	
	spin_lock_init(&pehci_hcd->lock);
	atomic_set(&pehci_hcd->nuofsofs,0);
	atomic_set(&pehci_hcd->missedsofs,0);
	
	/*Initialize host controller registers*/
	pehci_hcd_init_reg(pehci_hcd);


	/*reset the host controller*/
	retval = pehci_hcd_reset(usb_hcd);
	if(retval){
		err("phci_1762_start: error failing with status %x\n",retval);
		return retval;
		}

	/*enable interrupts*/
	pehci_hcd_enable_interrupts(pehci_hcd);


	/*put controller into operational mode*/
	retval = pehci_hcd_start_controller(pehci_hcd);
	if(retval){
		err("phci_1762_start: error failing with status %x\n",retval);
		return retval;
		}

    /*Init the phci qtd <-> ptd map buffers */
    pehci_hcd_init_map_buffers(pehci_hcd);


    /*set last maps, for iso its only 1, else 32 tds bitmap*/
    isp1762_reg_write32(pehci_hcd, pehci_hcd->regs.atltdlastmap,0x80000000);
    isp1762_reg_write32(pehci_hcd, pehci_hcd->regs.inttdlastmap, 0x80000000);
    isp1762_reg_write32(pehci_hcd, pehci_hcd->regs.isotdlastmap, 0x00000001);

	/*iso transfers are not active*/
	pehci_hcd->next_uframe = -1;
	pehci_hcd->periodic_sched = 0;

	/*initialize the periodic list*/
	for(count = 0;count < PTD_PERIODIC_SIZE;count++){
		pehci_hcd->periodic_list[count].framenumber = 0;
		INIT_LIST_HEAD(&pehci_hcd->periodic_list[count].itd_head);
	}

	/*initialize the tasket*/
	pehci_hcd->tasklet.func = pehci_hcd_tasklet;
	pehci_hcd->tasklet.data = (unsigned long)pehci_hcd;
	
	/*set the state of the host to ready,
	 * start processing interrupts
	 * */
	usb_hcd->state =  USB_STATE_READY;
	pehci_hcd->state = USB_STATE_READY;

	/*initialize root hub timer*/
	init_timer(&pehci_hcd->rh_timer);
    init_timer(&pehci_hcd->watchdog);

	/*take the usb_bus*/
	bus = &usb_hcd->self;
	/*create the root hub*/
	bus->root_hub = udev = usb_alloc_dev (NULL, bus,0);
	if(!udev)
		return -ENODEV;


	/*connect the root hub*/
	//usb_connect (bus->root_hub);
	/*root hub is high speed*/
	udev->speed = USB_SPEED_HIGH;

	pehci_info("enumerate the hub \n");
	/*enumerate root hub, should never fail*/
	if(usb_register_root_hub(udev, usb_hcd->self.controller) != 0){
			err("can not enumerate the hub\n");
			usb_put_dev (udev);
			bus->root_hub = 0;
			return -ENODEV;

		}
    pehci_info("-- %s: Exit\n",__FUNCTION__);
#if defined(CONFIG_ARCH_FIONA)
	// perform default initialization of OTG port
	//set_otg_mode(OTG_MODE_PERIPHERAL);
#endif

	return 0;
}


static void
pehci_hcd_stop(struct usb_hcd *usb_hcd)
{
#ifdef CONFIG_ARCH_FIONA
	phci_hcd *hcd = usb_hcd_to_pehci_hcd(usb_hcd);
#endif

	pehci_info("%s: enter\n",__FUNCTION__);

#ifdef CONFIG_ARCH_FIONA
	
	/* When the host controller driver is unloaded, we want to leave the
	 * host controller in suspend mode.  Philips app note AN10055 describes
	 * the entire process.  Since we're an internal "captive" USB bus, we
	 * don't have to suspend the internal hub ports as described in the
	 * app note.  We only have to suspend the root hub port and the host
	 * controller.
	 */

	// Leave a note that we're suspended
	
	isp1762_reg_write32(hcd, HC_SCRATCH_REG, USB_SUSPEND_SIGNATURE);

	// Clear the Run/Stop bit in the host controller
	
	isp1762_reg_write32(hcd, HC_USBCMD_REG, 0x00000000);

	// Suspend the root hub port
	
	isp1762_reg_write32(hcd, HC_PORTSC1_REG, 0x00001084);

#else
	/* no more interrupts ... */
	if (usb_hcd->state == USB_STATE_RUNNING)
		mdelay(2);
	if (in_interrupt ()) {		/* must not happen!! */
		pehci_info( "stopped in_interrupt!\n");

		return;
	}

	/*power off our root hub*/
	pehci_rh_control(usb_hcd, ClearPortFeature, USB_PORT_FEAT_POWER,
			1, NULL, 0);

	/*let the roothub power go off*/
	mdelay(20);
#endif
	pehci_info("%s: exit\n",__FUNCTION__);

	return;
}


/*allocate the host controller*/
static struct usb_hcd *
pehci_hcd_alloc(void)
{
    phci_hcd *hcd;
	hcd = (phci_hcd *)kmalloc(sizeof(phci_hcd), GFP_KERNEL);
	if(!hcd)
		return 0;
	memset(hcd, 0, sizeof(phci_hcd));
	hcd->qha = (isp1762_qha *)kmalloc(sizeof(isp1762_qha), GFP_KERNEL);
	if(!hcd->qha)
		  return 0;
	hcd->qhint = (isp1762_qhint *)kmalloc(sizeof(isp1762_qhint), GFP_KERNEL);
	if(!hcd->qhint)
		return 0;
	hcd->isotd = (isp1762_isoptd *) kmalloc( sizeof(isp1762_isoptd), GFP_KERNEL);
	if(!hcd->isotd){
		kfree(hcd->qhint);
		kfree(hcd->qha);
		return 0;
		}
	
	/*initialize our device list on this host*/
	 INIT_LIST_HEAD(&hcd->dev_list);
	 
	 return &hcd->usb_hcd;
}


/*free host controller*/
static void
pehci_hcd_free(
			struct usb_hcd *usb_hcd)
{
   phci_hcd  *pehci_hcd = usb_hcd_to_pehci_hcd(usb_hcd);
	if(pehci_hcd){
		kfree(pehci_hcd->qha);
		kfree(pehci_hcd->qhint);
		kfree(pehci_hcd->isotd);
		kfree(pehci_hcd);
		pehci_hcd = 0;
	}
}


/*submit urb , other than root hub*/
static int
pehci_hcd_urb_enqueue(struct usb_hcd *usb_hcd, struct urb *urb, int mem_flags)
{
	struct list_head	qtd_list;
	struct ehci_qh		*qh = 0;
	phci_hcd 		*pehci_hcd = usb_hcd_to_pehci_hcd(usb_hcd);
	int			status = 0;
	int			temp = 0, max = 0,num_tds = 0,mult = 0;
	urb_priv_t		*urb_priv = NULL;
//    pehci_info("++ %s: Entered\n",__FUNCTION__);
	INIT_LIST_HEAD(&qtd_list);
	urb->transfer_flags &= ~EHCI_STATE_UNLINK;

	temp = usb_pipetype(urb->pipe);
	max = usb_maxpacket (urb->dev, urb->pipe, !usb_pipein(urb->pipe));

	switch(temp){
	case PIPE_INTERRUPT:
		/*only one td*/
		num_tds = 1;
		mult = 1 + ((max >> 11) & 0x03);
		max &= 0x07ff;
		max *= mult;

		if (urb->transfer_buffer_length > max){
			err("interrupt urb length is greater then %d\n", max);
			return -EINVAL;
		}
		
		num_intr_pkts++;
		
		break;

	case PIPE_CONTROL:
		/*calculate the number of tds, follow 1 pattern*/
		num_tds  = (urb->transfer_buffer_length == 0) ? 2 :
			((urb->transfer_buffer_length -1)/HC_ATL_PL_SIZE + 3);
			
		num_ctrl_pkts++;
		
		break;

	case PIPE_BULK:
		num_tds = (urb->transfer_buffer_length -1)/HC_ATL_PL_SIZE + 1;
		if((urb->transfer_flags & URB_ZERO_PACKET) &&
			!(urb->transfer_buffer_length % max))
			num_tds++;
//		pehci_info("bulk tds %d, device %d\n", num_tds,
//				urb->dev->devnum);
		
		num_bulk_pkts++;
		
		break;
	
	default:
		return -EINVAL; /*not supported isoc transfers*/
	}

	/*make number of tds required*/
	urb_priv = kmalloc(sizeof (urb_priv_t) +
                        num_tds * sizeof(struct ehci_qtd),mem_flags);
        if(!urb_priv){
		err("memory   allocation error\n");
		return -ENOMEM;
	}

	memset(urb_priv, 0, sizeof (urb_priv_t) +
			num_tds * sizeof(struct ehci_qtd));
	urb_priv->qtd[0] = NULL;
	urb_priv->length = num_tds;

	{
		int i  = 0;
		/*allocate number of tds here.
		  better to do this in qtd_make routine*/
		for(i = 0; i < num_tds;i++){
			urb_priv->qtd[i] = phci_hcd_qtd_allocate(mem_flags);
			if(!urb_priv->qtd[i]){
				phci_hcd_urb_free_priv(pehci_hcd,urb_priv,NULL);
				return -ENOMEM;
			}
		}
	}

	/*keep a copy of this*/
	urb->hcpriv = urb_priv;

	switch(temp){
	case PIPE_INTERRUPT:
		phci_hcd_make_qtd(pehci_hcd, &qtd_list,urb, &status);
		if(status < 0)
			return status;
		qh = phci_hcd_submit_interrupt(pehci_hcd,&qtd_list,urb,&status);
		if(status < 0)
			return status;
		break;

	case PIPE_CONTROL:
	case PIPE_BULK:
		phci_hcd_make_qtd(pehci_hcd, &qtd_list,urb, &status);
		if(status < 0)
			return status;

	   	qh = phci_hcd_submit_async(pehci_hcd, &qtd_list, urb, &status);
    		if(status < 0)
			return status;
		break;

	default:
		return -ENODEV;
	}/*end of switch*/
  
#ifndef MSEC_INT_BASED
	/*submit tds but iso*/
	pehci_hcd_td_ptd_submit_urb(pehci_hcd,qh,qh->type);
#endif
//    pehci_info("-- %s: Exit\n",__FUNCTION__);
	return 0;
}





/*---------------------------------------------*
 io request handlers
 *---------------------------------------------*/

/*unlink urb*/
static int
pehci_hcd_urb_dequeue(struct usb_hcd *usb_hcd, struct urb *urb)
{
	int			status = 0;
	td_ptd_map_buff_t	*td_ptd_buf;
	td_ptd_map_t		*td_ptd_map;
	struct ehci_qh		*qh = 0;
	u32			skipmap = 0;
	u32			buffstatus = 0;
	unsigned long		flags;
	struct ehci_qtd		*qtd = 0;

	struct list_head	*qtd_list=0;
	urb_priv_t		*urb_priv = (urb_priv_t *)urb->hcpriv;
	phci_hcd		*hcd = usb_hcd_to_pehci_hcd(usb_hcd);
//    pehci_info("++ %s: Entered\n",__FUNCTION__);

//    pehci_info("device %d\n",urb->dev->devnum);

	spin_lock_irqsave(&hcd->lock, flags);
	
	switch (usb_pipetype (urb->pipe)) {
	case PIPE_CONTROL:
	case PIPE_BULK:
		qh = urb_priv->qh;
		td_ptd_buf = &td_ptd_map_buff[TD_PTD_BUFF_TYPE_ATL];
		td_ptd_map = &td_ptd_buf->map_list[qh->qtd_ptd_index];

		/*if its already been removed*/
		if (td_ptd_map->state == TD_PTD_NEW) {
			break;
		}

            /* These TDs are not pending anymore */
            td_ptd_buf->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;
		/*tell atl worker this urb is going to be removed*/
		td_ptd_map->state = TD_PTD_REMOVE;
		urb_priv->state |= DELETE_URB;
		// spin_unlock_irqrestore(&hcd->lock, flags);

		/*read the skipmap, to see if this transfer has to be rescheduled*/
		skipmap=isp1762_reg_read32(hcd,hcd->regs.atltdskipmap,skipmap);
		pehci_info("remove skip map %x, ptd map %x\n", skipmap, td_ptd_map->ptd_bitmap);

		buffstatus = isp1762_reg_read32(hcd, hcd->regs.buffer_status,
						buffstatus);

		/*avoid race between isr completion and this one*/
		isp1762_reg_write32(hcd, hcd->regs.atltdskipmap,
					skipmap | td_ptd_map->ptd_bitmap);

		while (!(skipmap & td_ptd_map->ptd_bitmap)) {
			udelay(125);
			skipmap = isp1762_reg_read32(hcd,
					hcd->regs.atltdskipmap, skipmap);
		}

		/* if all  transfers skipped,
		 * then disable the atl buffer,
		 * so that new transfer can come in
		 * need to see the side effects
		 * */
		if (skipmap == NO_TRANSFER_ACTIVE) {
			/*disable the buffer*/
			pehci_info("disable the atl buffer\n");
			buffstatus &= ~ATL_BUFFER;
			isp1762_reg_write32(hcd,hcd->regs.buffer_status,buffstatus);
		}

		qtd_list = &qh->qtd_list;
		/*this should remove all pending transfers*/
            pehci_check("num tds %d, urb length %d,device %d\n",
                    urb_priv->length, urb->transfer_buffer_length,urb->dev->devnum); 

            pehci_check("remove first qtd address %p\n", urb_priv->qtd[0]);
            pehci_check("length of the urb %d, completed %d\n", 
                    urb->transfer_buffer_length, urb->actual_length);
            qtd = urb_priv->qtd[urb_priv->length-1];
            pehci_check("qtd state is %x\n", qtd->state);       
		//pehci_hcd_complete_transfer(hcd,&td_ptd_map->ptd_bitmap,qh->type,NULL);
		pehci_hcd_urb_complete(hcd,qh,urb,td_ptd_map,NULL);

		break;
		
	case PIPE_INTERRUPT:
		pehci_check("phci_1762_urb_dequeue: INTR needs to be done\n");
		qh = urb_priv->qh;
		td_ptd_buf = &td_ptd_map_buff[TD_PTD_BUFF_TYPE_INTL];
		td_ptd_map = &td_ptd_buf->map_list[qh->qtd_ptd_index];

		/*urb is already been removed*/
		if (td_ptd_map->state == TD_PTD_NEW) {
			kfree(urb_priv);
			break;
		}

		// spin_lock_irqsave(&hcd->lock, flags);
            td_ptd_buf->pending_ptd_bitmap &= ~td_ptd_map->ptd_bitmap;
		td_ptd_map->state = TD_PTD_REMOVE;
		urb_priv->state |= DELETE_URB;
		// spin_unlock_irqrestore(&hcd->lock, flags);

		/*read the skipmap, to see if this transfer has to be rescheduled*/
		skipmap = isp1762_reg_read32(hcd, hcd->regs.inttdskipmap, skipmap);
		isp1762_reg_write32(hcd, hcd->regs.inttdskipmap,
				skipmap | td_ptd_map->ptd_bitmap);

		qtd_list = &qh->qtd_list;
		//pehci_hcd_complete_transfer(hcd,&td_ptd_map->ptd_bitmap,qh->type,NULL);
		pehci_hcd_urb_complete(hcd,qh,urb,td_ptd_map,NULL);
		if (!list_empty(qtd_list)) {
			err("Only one TD in my endpoint\n");
			PEHCI_BUG(hcd); /*I have only one td in my endpoint*/
		}
		
		break;
		
	case PIPE_ISOCHRONOUS:
		/*nothing to do here, wait till all transfers are done in iso worker*/
		break;
	}

//	pehci_info("%s: exit, status %d\n", __FUNCTION__,status);
	
	spin_unlock_irqrestore(&hcd->lock, flags);
	
	return status;
}



// bulk qh holds the data toggle

static void
pehci_hcd_endpoint_disable (struct usb_hcd *usb_hcd, struct hcd_dev *dev, int ep)
{
	phci_hcd		*ehci = usb_hcd_to_pehci_hcd (usb_hcd);
	int			epnum;
	unsigned long		flags;
	struct ehci_qh		*qh;

//    pehci_info("++ %s: Entered\n",__FUNCTION__);
	/* ASSERT:  any requests/urbs are being unlinked */
	/* ASSERT:  nobody can be submitting urbs for this any more */

	epnum = ep & USB_ENDPOINT_NUMBER_MASK;
	if (epnum != 0 && (ep & USB_DIR_IN))
		epnum |= 0x10;


	spin_lock_irqsave (&ehci->lock, flags);
	qh = (struct ehci_qh *) dev->ep [epnum];
	if (!qh){
		    goto done;
	}else {
		/*i will complete whatever left on this endpoint*/
		 pehci_complete_device_removal(ehci,qh);
		  dev->ep [epnum] = NULL;
		  goto done;
	}
	done:
	/*FIXME, thats it, no more any work for this endpoint*/
	dev->ep[epnum] = NULL;
	spin_unlock_irqrestore (&ehci->lock, flags);
//    pehci_info("-- %s: Exit\n",__FUNCTION__);
	return;
}


/*called by core, for current frame number*/
static int
pehci_hcd_get_frame_number(struct usb_hcd *usb_hcd)
{
    u32 framenumber = 0;
    phci_hcd *pehci_hcd = usb_hcd_to_pehci_hcd(usb_hcd);
	framenumber = isp1762_reg_read32(pehci_hcd, pehci_hcd->regs.frameindex, framenumber);
	return framenumber;
}


/*root hub status data, called by root hub timer
*return 0, if no change, else
*      1, incase of high speed device
*/
static int
pehci_rh_status_data (
		struct usb_hcd *usb_hcd,
		char *buf)
{

	u32		temp = 0, status = 0;
	u32		ports = 0, i, retval = 1;
	unsigned long	flags;
	phci_hcd *hcd = usb_hcd_to_pehci_hcd(usb_hcd);

	/* init status to no-changes */
	buf [0] = 0;
	/*number of ports*/
	ports = isp1762_reg_read32(hcd, hcd->regs.hcsparams, ports);
	ports &= 0xf;
	//info("number of ports%d\n",ports); 

	spin_lock_irqsave(&hcd->lock, flags);
	/*read the port status registers*/
	for(i=0;i<ports;i++)
	{
		temp = isp1762_reg_read32(hcd,hcd->regs.ports[i],temp);
		if(temp & PORT_OWNER){
			/* dont report the port status change in case of CC HCD
			* but clear the port status , if there is any*/
			if(temp & PORT_CSC){
				temp &=~PORT_CSC;
				isp1762_reg_write32(hcd, hcd->regs.ports[i],temp);
				continue;
			}
    }

	if (!(temp & PORT_CONNECT))
			hcd->reset_done [i] = 0;
	if ((temp & (PORT_CSC 

#if !defined(CONFIG_ARCH_FIONA)
			| PORT_PEC | PORT_OCC
#endif

		)) != 0) {
			printk("port status change happened 0x%08x\n",temp);
			if (i < 7)
			    buf [0] |= 1 << (i + 1);
			else
			    buf [1] |= 1 << (i - 7);
			status = STS_PCD;
		}
	 }

 	spin_unlock_irqrestore (&hcd->lock, flags);
	return status ? retval : 0;


}






/*root hub control requests*/
static int
pehci_rh_control (
	struct usb_hcd	*usb_hcd,
	u16		typeReq,
	u16		wValue,
	u16		wIndex,
	char		*buf,
	u16		wLength)
{
	u32		ports = 0;
	u32		temp = 0, status;
	unsigned long	flags;
	int		retval = 0;
	phci_hcd *hcd = usb_hcd_to_pehci_hcd(usb_hcd);
	
	ports = isp1762_reg_read32(hcd, hcd->regs.hcsparams, ports);
//	pehci_info("%s: number of ports %d\n",__FUNCTION__, ports);
    spin_lock_irqsave(&hcd->lock,flags);
	switch(typeReq){
	case ClearHubFeature:
	switch (wValue) {
    case C_HUB_LOCAL_POWER:
		case C_HUB_OVER_CURRENT:
		  /* no hub-wide feature/status flags */
			break;
		default:
			goto error;
		}
		break;
	case ClearPortFeature:
		if (!wIndex || wIndex > (ports & 0xf)){
			pehci_info("ClearPortFeature not valid port number %d, should be %d\n",
				       wIndex, (ports & 0xf));
			    goto error;}
		wIndex--;
		temp = isp1762_reg_read32(hcd, hcd->regs.ports[wIndex], temp);
		if (temp & PORT_OWNER){
      		printk("port is owned by the CC host\n");
    	break;}

		switch (wValue) {
    	 case USB_PORT_FEAT_ENABLE:
       		 	isp1762_reg_write32(hcd,hcd->regs.ports[wIndex],
			 						temp & ~PORT_PE);

			break;
		case USB_PORT_FEAT_C_ENABLE:
#if !defined(CONFIG_ARCH_FIONA)
      		printk("disable the port\n");
			isp1762_reg_write32(hcd,hcd->regs.ports[wIndex],temp | PORT_PEC);
#endif
			break;
		case USB_PORT_FEAT_SUSPEND:
		case USB_PORT_FEAT_C_SUSPEND:
    			printk("dont know what to do here \n");
    			/* ? */
			break;
		case USB_PORT_FEAT_POWER:
			if(ports & 0x10){/*port has has power control switches*/
		    		isp1762_reg_write32(hcd, hcd->regs.ports[wIndex], temp & ~PORT_POWER);
			}
			break;
		case USB_PORT_FEAT_C_CONNECTION:
        	phci_print("connect change, status is 0x%08x\n",temp);
			isp1762_reg_write32(hcd, hcd->regs.ports[wIndex], temp | PORT_CSC);
        	//isp1762_reg_write32(hcd, hcd->regs.ports[wIndex], temp | PORT_CSC);
       		break;
		case USB_PORT_FEAT_C_OVER_CURRENT:
#if !defined(CONFIG_ARCH_FIONA)
			isp1762_reg_write32(hcd, hcd->regs.ports[wIndex], temp | PORT_OCC);
#endif
			break;

		/*no need getport status will handle this
    	case USB_PORT_FEAT_C_RESET:
			phci_print("clear port reset feature\n");
        	isp1762_reg_write32(hcd, hcd->regs.ports[wIndex], temp & ~PORT_RESET);
          	do{
              temp = isp1762_reg_read32(hcd, hcd->regs.ports[wIndex], temp);
            }while(temp & PORT_RESET);
			break;*/
		default:
			goto error;

		}
		break;

	case GetHubDescriptor:
		  pehci_hub_descriptor(hcd, (struct usb_hub_descriptor *)buf);
		break;

	case GetHubStatus:
		/* no hub-wide feature/status flags */
		memset (buf, 0, 4);
		break;
	case GetPortStatus:
		if (!wIndex || wIndex > (ports & 0xf)){
			pehci_info("GetPortStatus,not valid port number %d, should be %d\n",
				       wIndex, (ports & 0xf));
			goto error;}
		wIndex--;
		status = 0;
		temp = isp1762_reg_read32(hcd, hcd->regs.ports[wIndex], temp);
		/*connect status chnage*/
		if (temp & PORT_CSC){
			status |= 1 << USB_PORT_FEAT_C_CONNECTION;
     	 phci_print("feature CSC 0x%08x and status 0x%08x  \n", temp, status);
      	}

#if !defined(CONFIG_ARCH_FIONA)
		/*port enable change*/
		if (temp & PORT_PEC){
		   status |= 1 << USB_PORT_FEAT_C_ENABLE;
        phci_print("feature PEC  0x%08x and status 0x%08x  \n", temp, status);
        }
		/*port over-current*/
		if (temp & PORT_OCC){
			status |= 1 << USB_PORT_FEAT_C_OVER_CURRENT;
      	phci_print("feature OCC 0x%08x and status 0x%08x  \n", temp, status);
      	}
#endif

		/* whoever resets must GetPortStatus to complete it!! */
		if ((temp & PORT_RESET) && jiffies > hcd->reset_done[wIndex]){
        		status |= 1 << USB_PORT_FEAT_C_RESET;
      		phci_print("feature reset 0x%08x and status 0x%08x\n", temp, status);
			/* force reset to complete */
			isp1762_reg_write32(hcd, hcd->regs.ports[wIndex], temp & ~PORT_RESET);
			do {
			    	temp = isp1762_reg_read32(hcd, hcd->regs.ports[wIndex], temp);
        			udelay(20);
        	} while (temp & PORT_RESET);
			/* see what we found out */
			temp = phci_check_reset_complete (hcd, wIndex, temp);
		}

		// don't show wPortStatus if it's owned by a companion hc

		if (!(temp & PORT_OWNER)) {
     	if (temp & PORT_CONNECT) {
				status |= 1 << USB_PORT_FEAT_CONNECTION;
				status |= 1 << USB_PORT_FEAT_HIGHSPEED;
			}
			if (temp & PORT_PE)
				status |= 1 << USB_PORT_FEAT_ENABLE;
			if (temp & PORT_SUSPEND)
				status |= 1 << USB_PORT_FEAT_SUSPEND;
#if !defined(CONFIG_ARCH_FIONA)
			if (temp & PORT_OC)
				status |= 1 << USB_PORT_FEAT_OVER_CURRENT;
#endif
			if (temp & PORT_RESET)
				status |= 1 << USB_PORT_FEAT_RESET;
			if (temp & PORT_POWER)
				status |= 1 << USB_PORT_FEAT_POWER;
		}

   		 // we "know" this alignment is good, caller used kmalloc()...
			*((u32 *) buf) = cpu_to_le32 (status);
		break;

	case SetHubFeature:
		switch (wValue) {
		case C_HUB_LOCAL_POWER:
		case C_HUB_OVER_CURRENT:
			/* no hub-wide feature/status flags */
			break;
		default:
			goto error;
		}
		break;
	case SetPortFeature:
		if (!wIndex || wIndex > (ports & 0xf)){
			pehci_info("SetPortFeature not valid port number %d, should be %d\n",
				       wIndex, (ports & 0xf));
			goto error;}
		wIndex--;
		temp = isp1762_reg_read32(hcd, hcd->regs.ports[wIndex], temp);
		if (temp & PORT_OWNER){
			break;
		}
		switch (wValue) {
		case USB_PORT_FEAT_ENABLE:
			/*enable the port*/
			isp1762_reg_write32(hcd, hcd->regs.ports[wIndex], temp | PORT_PE);
			break;
		case USB_PORT_FEAT_SUSPEND:
        	 isp1762_reg_write32(hcd, hcd->regs.ports[wIndex], temp | PORT_SUSPEND);
			 break;
		case USB_PORT_FEAT_POWER:
        		if(ports & 0x10){
		      		isp1762_reg_write32(hcd, hcd->regs.ports[wIndex], temp | PORT_POWER);
			}
       		  break;
		case USB_PORT_FEAT_RESET:
			 if ((temp & (PORT_PE|PORT_CONNECT)) == PORT_CONNECT
					&& PORT_USB11 (temp)) {
				printk("error:port %d low speed --> companion\n",
										wIndex + 1);
				temp |= PORT_OWNER;
			} else {
				temp |= PORT_RESET;
				temp &= ~PORT_PE;

				/*
				 * caller must wait, then call GetPortStatus
				 * usb 2.0 spec says 50 ms resets on root
				 */
				hcd->reset_done [wIndex] = jiffies
				    	+ ((50 /* msec */ * HZ) / 1000);
			}
         isp1762_reg_write32(hcd,hcd->regs.ports[wIndex], temp);
          break;
		default:
			goto error;
		}
			break;
	default:
  	phci_print("this request doesnt fit anywhere\n");
 error:
		/* "stall" on error */
	pehci_info("unhandled root hub request: typereq 0x%08x, wValue %d, wIndex %d\n",
	          				typeReq, wValue, wIndex);
		retval = -EPIPE;
   }
	spin_unlock_irqrestore (&hcd->lock, flags);
	return retval;
}



/*-------------------------------------------------------------------------*/

static const struct hc_driver pehci_driver = {
	.description =		hcd_name,

	/*
	 * generic hardware linkage
	 */
	.flags =		HCD_USB2 | HCD_MEMORY,

	/*
	 * basic lifecycle operations
	 */
	.reset =		pehci_hcd_reset,
	.start =		pehci_hcd_start,
	.stop =			pehci_hcd_stop,

	/*
	 * memory lifecycle (except per-request)
	 */
	.hcd_alloc =		pehci_hcd_alloc,
	.hcd_free =		pehci_hcd_free,

	/*
	 * managing i/o requests and associated device resources
	 */
	.urb_enqueue =		pehci_hcd_urb_enqueue,
	.urb_dequeue =		pehci_hcd_urb_dequeue,
	.endpoint_disable =	pehci_hcd_endpoint_disable,

	/*
	 * scheduling support
	 */
	.get_frame_number =	pehci_hcd_get_frame_number,

	/*
	 * root hub support
	 */
	.hub_status_data =	pehci_rh_status_data,
	.hub_control =		pehci_rh_control,
};




static struct usb_hcd *usb_hcd = NULL;

#define ISP1762_IO_BASE 0x0C000000
#define ISP1762_IO_LENGTH 0x80000

#define ISP1762_HC_IRQ   IRQ_GPIO(2)
#define ISP1762_DC_IRQ   IRQ_GPIO(3)



int 
pehci_hcd_probe(struct device *dev)
{
	struct hc_driver	*driver;
	__u32			resource = 0;
	int			len = 0;
	void 			*base = 0;
	phci_hcd		*pehci_hcd;
	int			retval = 0;
	char			buf [8], *bufp = buf;
	int			status =0;
#ifndef CONFIG_ARCH_FIONA
	int			retry_count = 0;
	u16			reg_data = 0;
#endif

    pehci_info("++ %s: Entered\n",__FUNCTION__);
	if (usb_disabled())
		return -ENODEV;

	if (!(driver = (struct hc_driver *) &pehci_driver))
		return -EINVAL;

	len = ISP1762_IO_LENGTH;
	resource = ISP1762_IO_BASE;
	
	if(!request_mem_region(resource, len,driver->description)){
		err("host controller already in use\n");
		return -EBUSY;
	}

	base = ioremap_nocache(resource,len);
	if(!base){

		release_mem_region(resource, len);
		return -ENOMEM;
	}
	
#ifdef CONFIG_ARCH_FIONA

	{
		int suspended = readl((u32 *)(base + HC_SCRATCH_REG));
		int i, retries = 0;

		if (suspended == USB_SUSPEND_SIGNATURE) {
			printk("<1>Waiting for CLKREADY\n");

			do {
				i = readl((u32 *)(base + HC_INTERRUPT_REG));
				retries++;
				if ((retries & 0xFFF) == 0)
					printk("<1>  retries = %d\n",retries);
			} while (((i & CLKREADY) == 0) && (retries < 0x5000));

			writel(0x00001004, (u32 *)(base + HC_PORTSC1_REG));

			writel(0x00000001, (u32 *)(base + HC_USBCMD_REG));

			writel(USB_RUNNING_SIGNATURE,
					(u32 *)(base + HC_SCRATCH_REG));

		} else {
			printk("<1>### 1761 is marked 0x%08X ###\n",suspended);
		}
	}

#else
	retry_count = 20;
	while((reg_data != 0xFACE) && retry_count) {
		/*by default host is in 16bit mode, so
		 * io operations at this stage must be 16 bit
		 * */
		 writew(0xFACE,((u8*)base + HC_SCRATCH_REG));
		 udelay(100);
		reg_data = readw((u8*)base+HC_SCRATCH_REG);
		retry_count--;
	}

	/* Host Controller presence is detected by writing to scratch register
	 * and reading back and checking the contents are same or not
	 */
	if(reg_data != 0xFACE) {
		err("%s scratch register mismatch %x",
				"isp1762_driver_name",reg_data);
		status = -ENODEV;
		goto clean;
	}
#endif

	usb_hcd = driver->hcd_alloc ();
	if (usb_hcd == NULL){
		status = -ENOMEM;
		goto clean;
	}

	/*this is our host*/
	pehci_hcd = usb_hcd_to_pehci_hcd(usb_hcd);
	pehci_hcd->iobase = (u8*)base;
	pehci_hcd->iolength = len;

	usb_hcd->driver = driver;
	usb_hcd->description = driver->description;
	usb_hcd->self.bus_name = "system bus"; 
	usb_hcd->product_desc = "PHILIPS ISP1762";
	usb_hcd->self.controller = dev;

	if ((retval = hcd_buffer_create (usb_hcd)) != 0) {
		driver->hcd_free (usb_hcd);
		 goto clean;
	}

	dev_info (usb_hcd->self.controller, "%s\n", usb_hcd->product_desc);
	
	usb_hcd->state = USB_STATE_HALT;

	set_irq_type(ISP1762_HC_IRQ, IRQT_FALLING);

	status = request_irq (ISP1762_HC_IRQ, pehci_hcd_irq, SA_SHIRQ,
			      usb_hcd->description, usb_hcd);

	if (status != 0) {
		dev_err (usb_hcd->self.controller,
				"request interrupt %s failed\n", bufp);
		driver->hcd_free(usb_hcd);
		goto clean;
	}

	usb_hcd->irq = ISP1762_HC_IRQ;
	sprintf (buf, "%d", usb_hcd->irq);

	dev_info (usb_hcd->self.controller, "irq %s, %s %p\n", bufp,
		(driver->flags & HCD_MEMORY) ? "pci mem" : "io base",
		base);

	usb_bus_init (&usb_hcd->self);
	usb_hcd->self.op = &usb_hcd_operations;
	usb_hcd->self.hcpriv = (void *) usb_hcd;
	usb_hcd->self.release = &pehci_hcd_release;
	init_timer (&usb_hcd->rh_timer);
#ifndef DMA
		usb_hcd->self.controller->dma_mask = 0;
#else
#endif

	INIT_LIST_HEAD (&usb_hcd->dev_list);

	usb_register_bus (&usb_hcd->self);
	if ((status = driver->start (usb_hcd)) < 0) {
		dev_err (usb_hcd->self.controller, "init error %d\n", retval);
		free_irq(usb_hcd->irq,usb_hcd);
		driver->hcd_free(usb_hcd);
		goto clean;
	}

	return status;

clean:
		release_mem_region(resource, len);
		iounmap(base);

	return status;

} 



/*remove the host controller*/
static int
pehci_hcd_remove(
	struct device *dev)
{
	phci_hcd *hcd = NULL;
	u32 temp;

	if (!usb_hcd)
		return 0;

	hcd = usb_hcd_to_pehci_hcd(usb_hcd);

	if (in_interrupt ()) {
		err("Removing host controller in interrupt handler!\n");
		PEHCI_BUG(hcd);
	}

	pehci_info("%s: enter\n",__FUNCTION__);

	if (HCD_IS_RUNNING (usb_hcd->state)){
		pehci_info("host is running\n");
		usb_hcd->state = USB_STATE_QUIESCING;
	}

	// usb_hcd->state = USB_STATE_HALT;


	usb_disconnect (&usb_hcd->self.root_hub);

	usb_hcd->driver->stop (usb_hcd);
	hcd_buffer_destroy (usb_hcd);

	//free_irq (usb_hcd->irq, usb_hcd);
	/*disable global interrupt*/
	temp &=0;
	temp = isp1762_reg_read32(hcd, hcd->regs.hwmodecontrol, temp);
	temp &= ~(1);
	isp1762_reg_write32(hcd, hcd->regs.hwmodecontrol, temp);

	/*disable individual interrupts*/
	temp &= ~INTR_ENABLE_MASK;
	isp1762_reg_write32(hcd,hcd->regs.interruptenable, temp);

	usb_hcd->state = USB_STATE_HALT;
	
	free_irq(usb_hcd->irq, usb_hcd);
	
	/*let the interrupt source stabilise*/
	mdelay(10);
	/*no more processing of any interrupt or request*/
	hcd->state = USB_STATE_HALT;
	
#ifndef CONFIG_ARCH_FIONA
	/*put host controller into
	 * non-functional state
	 * */
	temp &=0;
	temp |= ~CMD_RUN;
	isp1762_reg_write32(hcd, hcd->regs.command, temp);
#endif

	iounmap(hcd->iobase);

	release_mem_region(ISP1762_IO_BASE, ISP1762_IO_LENGTH);

	usb_deregister_bus (&usb_hcd->self);

//	usb_hcd->driver->hcd_free (usb_hcd);  	
	pehci_info("%s: exit\n", __FUNCTION__);
	return 0;
}


#if defined(CONFIG_ARCH_FIONA)
static void set_otg_mode(int mode)
{
    if (usb_hcd) {
        phci_hcd *hcd = usb_hcd_to_pehci_hcd(usb_hcd);

        // reset the control OTG register to default state
        isp1762_reg_write32(hcd, hcd->regs.port1ctl,
	        PORT1_SW_SEL_HC_DC | PORT1_DM_PULLDOWN | PORT1_DP_PULLDOWN | // set
                ((PORT1_OTG_DISABLE | PORT1_OTG_SE0_EN | PORT1_BDIS_ACON_EN | // clear
                PORT1_VBUS_CHRG | PORT1_VBUS_DISCHRG | PORT1_VBUS_DRV |
                PORT1_SEL_CP_EXT | PORT1_DP_PULLUP) << 16));

        printk("usb: setting OTG port operating mode to %s\n", mode == OTG_MODE_HOST ? "host" : "peripheral");

        // set the OTG operating mode (host or peripheral)
        isp1762_reg_write32(hcd, hcd->regs.port1ctl, mode == OTG_MODE_HOST
                ? (PORT1_SEL_CP_EXT | PORT1_VBUS_DRV | (PORT1_SW_SEL_HC_DC << 16)) // host mode
                : (PORT1_OTG_DISABLE | ((PORT1_DP_PULLDOWN | PORT1_DM_PULLDOWN) << 16))); // peripheral mode

        // save the operating state
        otg_mode = mode;
    }
}


static int
proc_hostmode_read(
	char *page,
	char **start,
	off_t off,
	int count,
	int *eof,
	void *data)
{
  *eof = 1;
	
  return sprintf(page, "%sabled\n", otg_mode == OTG_MODE_HOST ? "en" : "dis");
}


static int
proc_hostmode_write(
	struct file *file,
	const char __user *buf,
	unsigned long count,
	void *data)
{
  static char *enable_string = OTG_HOST_ENABLE_STRING;
  char lbuf[64];
  int prev_otg_mode, enable_string_len = strlen(enable_string);

  memset(lbuf, 0, sizeof(lbuf));

  if (copy_from_user(lbuf, buf, enable_string_len)) {
    return -EFAULT;
  }

  prev_otg_mode = otg_mode;

  otg_mode = strncmp(lbuf, enable_string, enable_string_len) ? OTG_MODE_PERIPHERAL : OTG_MODE_HOST;

  if (otg_mode != prev_otg_mode) {
  	set_otg_mode(otg_mode);
  }

  return count;
}

static int proc_errors_read(char *page, char **start, off_t off, int count,
				int *eof, void *data)
{
	*eof = 1;
	
	return sprintf(page, "%u\n",
			(unsigned int) atomic_read(&pehci_error_counter));
}

static int proc_errors_write(struct file *file, const char __user *buf,
				unsigned long count, void *data)
{
	char		lbuf[80];
	unsigned long	value;
	
	memset(lbuf, 0, sizeof(lbuf));

	if (count > 11)
		return -EINVAL;
	
	if (copy_from_user(lbuf, buf, count))
		return -EFAULT;
	
	value = simple_strtoul(lbuf, NULL, 0);
	
	atomic_set(&pehci_error_counter, value);
	
	return count;
}

#endif


static struct device_driver pehci_hcd_driver = {
	.name		= (char *) hcd_name,
	.bus		= &platform_bus_type,
	.probe		= pehci_hcd_probe,
	.remove		= __exit_p(pehci_hcd_remove),
};

static void pehci_hcd_platform_release(struct device *device)
{
	// This is called when the reference count goes to zero.
}

static struct platform_device pehci_hcd_device = {
	.name	= (char *) hcd_name,
	.id	= 0,
	.dev	= {
		.release = pehci_hcd_platform_release,
	}
};

static int  __init pehci_module_init(void)
{
  int result = 0;

#ifdef FIXME
	return -ENOMEM;	// temporarily prevent host driver from loading
#endif

  phci_hcd_mem_init();

#if defined(CONFIG_ARCH_FIONA)
  atomic_set(&pehci_error_counter, 0);	// make sure counter starts out clear

  proc_usb_parent = create_proc_entry("usb", S_IFDIR | S_IRUGO | S_IXUGO, NULL);
  if (proc_usb_parent) {
    proc_usb_hostmode_ctrl = create_proc_entry("host_mode", 0644, proc_usb_parent);
    if (proc_usb_hostmode_ctrl) {
      proc_usb_hostmode_ctrl->data = NULL;
      proc_usb_hostmode_ctrl->read_proc = proc_hostmode_read;
      proc_usb_hostmode_ctrl->write_proc = proc_hostmode_write;
    }
    proc_usb_errors_ctrl = create_proc_entry("errors", 0644, proc_usb_parent);
    if (proc_usb_errors_ctrl) {
      proc_usb_errors_ctrl->data = NULL;
      proc_usb_errors_ctrl->read_proc = proc_errors_read;
      proc_usb_errors_ctrl->write_proc = proc_errors_write;
    }
  }
#endif

  /*register driver*/
  
  result = driver_register(&pehci_hcd_driver);
  
  if(!result) {
    result = platform_device_register(&pehci_hcd_device);
    if(!result) {
      info("pehci_module_init, driver has been registered");
    }
  }  else {
    err("pehci_module_init, driver has not been registered with error %x",result);
  }
  
#ifdef CONFIG_FIONA_PM_USB_HOST
    // Register USB Host driver with Fiona Power Manager,
    // we're ignoring result here as we'll still allow driver to
    // work even if we can't do efficient power management...
    if (usb_host_register_with_fpow((void*)&pehci_hcd_device, &g_usb_host_fpow_component_ptr))
        printk("USB Host: Failed to register with fpow !!\n");
#endif

  return result;
}

module_init (pehci_module_init);

MODULE_DESCRIPTION ("ISP 1762 slave host controller driver, version 0.01");
MODULE_AUTHOR ("krishan");
MODULE_LICENSE ("GPL");

static void __exit pehci_module_exit(void)
{
	if (proc_usb_hostmode_ctrl) {
		remove_proc_entry("usb/host_mode", NULL);
		proc_usb_hostmode_ctrl = NULL;
	}
	if (proc_usb_errors_ctrl) {
		remove_proc_entry("usb/errors", NULL);
		proc_usb_errors_ctrl = NULL;
	}
	if (proc_usb_parent) {
		remove_proc_entry("usb", NULL);
		proc_usb_parent = NULL;
	}
#ifdef CONFIG_FIONA_PM_USB_HOST
    // Unregister USB Host driver with Fiona Power Manager
    usb_host_unregister_with_fpow(g_usb_host_fpow_component_ptr);
    g_usb_host_fpow_component_ptr  = NULL;
#endif
	driver_unregister(&pehci_hcd_driver);

	platform_device_unregister(&pehci_hcd_device);
}


module_exit (pehci_module_exit);
