#include "apollo.h"
#include "apollo_waveform.c"

static char *apollo_base_map = NULL;

static reg_addr_t reg_GPSR1  = NULL;
static reg_addr_t reg_GPSR2  = NULL;

static reg_addr_t reg_GPCR1  = NULL;
static reg_addr_t reg_GPCR2  = NULL;

static reg_addr_t reg_GPLR1  = NULL;
static reg_addr_t reg_GPLR2  = NULL;

static reg_addr_t reg_GPDR1  = NULL;
static reg_addr_t reg_GPDR2  = NULL;

static int apollo_init_comm  = 0;

static unsigned char apollo_apply_fx(unsigned char data, int i, int which_fx);

static u8 lighten_2bpp_nybble_table_dark_gray[16] = {
  0x00, 0x00, 0x01, 0x02, 0x00, 0x00, 0x01, 0x02, // 00 -> 00, 01 -> 00,
  0x04, 0x04, 0x05, 0x06, 0x08, 0x08, 0x09, 0x0A  // 10 -> 01, 11 -> 10
};

static u8 lighten_2bpp_nybble_table_lite_gray[16] = {
  0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x01, // 00 -> 00, 01 -> 00,
  0x00, 0x00, 0x00, 0x01, 0x04, 0x04, 0x04, 0x05  // 10 -> 00, 11 -> 01
};

static u8 darken_2bpp_nybble_table_lite_gray[16] = {
  0x05, 0x06, 0x07, 0x07, 0x09, 0x0A, 0x0B, 0x0B, // 00 -> 01, 01 -> 10,
  0x0D, 0x0E, 0x0F, 0x0F, 0x0D, 0x0E, 0x0F, 0x0F  // 10 -> 11, 11 -> 11
};

static u8 darken_2bpp_nybble_table_dark_gray[16] = {
  0x0A, 0x0B, 0x0B, 0x0B, 0x0E, 0x0F, 0x0F, 0x0F, // 00 -> 10, 01 -> 11,
  0x0E, 0x0F, 0x0F, 0x0F, 0x0E, 0x0F, 0x0F, 0x0F  // 10 -> 11, 11 -> 11
};

static u8 noblack_2bpp_nybble_table[16] = {
  0x00, 0x01, 0x02, 0x02, 0x04, 0x05, 0x06, 0x06, // 00 -> 00, 01 -> 01,
  0x08, 0x09, 0x0A, 0x0A, 0x08, 0x09, 0x0A, 0x0A, // 10 -> 10, 11 -> 10
};

#define stipple_even_row_mask        0
#define stipple_even_row_mask_type   1

#define stipple_odd_row_mask         2
#define stipple_odd_row_mask_type    3

#define stipple_table_size           4     

#define stipple_table_mask_and       1
#define stipple_table_mask_or        0

#define STIPPLE_TABLE_MASK(t, m, d)  (t ? (m & d) : (m | d))

static u8 stipple_lighten_dark_gray_2bpp_table[stipple_table_size] = {
  0x33, stipple_table_mask_and,     // row 0: WXWX... (white keep  white keep...)
  0x33, stipple_table_mask_or       // row 1: XBXB... (keep  black keep  black...)
};

static u8 stipple_lighten_lite_gray_2bpp_table[stipple_table_size] = {
  0x33, stipple_table_mask_and,     // row 0: WXWX...
  0xCC, stipple_table_mask_and      // row 1: XWXW...
};

static u8 stipple_darken_lite_gray_2bpp_table[stipple_table_size] = {
  0xCC, stipple_table_mask_or,      // row 0: BXBX...
  0xCC, stipple_table_mask_and      // row 1: XWXW...
};

static u8 stipple_darken_dark_gray_2bpp_table[stipple_table_size] = {
  0xCC, stipple_table_mask_or,      // row 0: BXBX...
  0x33, stipple_table_mask_or       // row 1: XBXB...
};

static u8 stipple_dark_gray_2bpp_table[stipple_table_size] = {
  0xBB, stipple_table_mask_and,     // row 0: GXGX...
  0xEE, stipple_table_mask_and      // row 1: XGXG...
};

static u8 stipple_lite_gray_2bpp_table[stipple_table_size] = {
  0x77, stipple_table_mask_and,     // row 0: gXgX...
  0xDD, stipple_table_mask_and      // row 1: XgXg...
};

static int apollo_display_fx        = APOLLO_FX_NONE;
static int apollo_mask_fx           = APOLLO_FX_NONE;

static int apollo_fx_exclude_rect   = 0;
static int apollo_fx_exclude_x1     = 0;
static int apollo_fx_exclude_y1     = 0;
static int apollo_fx_exclude_x2     = 0;
static int apollo_fx_exclude_y2     = 0;
static int apollo_fx_exclude_x      = 0;
static int apollo_fx_exclude_y      = 0;

static u8* apollo_shade_table       = NULL;

static u8* apollo_stipple_table     = NULL;
static int apollo_stipple_rowbytes  = 0;
static int apollo_stipple_bytes     = 0;
static int apollo_stipple_rows      = 0;

static u8* apollo_mask_framebuffer  = NULL;
static int apollo_mask_fb_rowbytes  = 0;
static u8* apollo_mask_buffer       = NULL;
static int apollo_mask_bufsize      = 0;
static int apollo_mask_rowbytes     = 0;
static int apollo_mask_x_start      = 0;
static int apollo_mask_x            = 0;
static int apollo_mask_y            = 0;

u32 apollo_get_ack(void) {
  return(apollo_read_reg(reg_GPLR2) & (1 << 19));
}

int apollo_comm_init(void) {
  if (0 == apollo_init_comm) {
    init_waitqueue_head(&apollo_ack_wq);
    
    apollo_base_map = (char *)ioremap_nocache(BASE_ADDR, MAP_SIZE);
    
    reg_GPSR1 = (reg_addr_t) (apollo_base_map + (GPSR1_ADDR & MAP_MASK));
    reg_GPSR2 = (reg_addr_t) (apollo_base_map + (GPSR2_ADDR & MAP_MASK));
    
    reg_GPCR1 = (reg_addr_t) (apollo_base_map + (GPCR1_ADDR & MAP_MASK));
    reg_GPCR2 = (reg_addr_t) (apollo_base_map + (GPCR2_ADDR & MAP_MASK));
    
    reg_GPLR1 = (reg_addr_t) (apollo_base_map + (GPLR1_ADDR & MAP_MASK));
    reg_GPLR2 = (reg_addr_t) (apollo_base_map + (GPLR2_ADDR & MAP_MASK));
    
    reg_GPDR1 = (reg_addr_t) (apollo_base_map + (GPDR1_ADDR & MAP_MASK));
    reg_GPDR2 = (reg_addr_t) (apollo_base_map + (GPDR2_ADDR & MAP_MASK));
    
    apollo_init_irq();
    
    apollo_init_comm = 1;
  }

  return 0;
}

int apollo_comm_shutdown(void) {
  if (1 == apollo_init_comm) {
    apollo_shutdown_irq();
    
    iounmap(apollo_base_map);
    apollo_base_map = NULL;
    
    reg_GPSR1 = NULL;
    reg_GPSR2 = NULL;
    
    reg_GPCR1 = NULL;
    reg_GPCR2 = NULL;
    
    reg_GPLR1 = NULL;
    reg_GPLR2 = NULL;
    
    reg_GPDR1 = NULL;
    reg_GPDR2 = NULL;
    
    apollo_init_comm = 0;
  }

  return 0;
}

int wait_for_ack(void) {
  register int i=0;

  /* wait for nACK to go low (GPLR2_83) */

  for(i = 0; i < FAST_TIMEOUT; i++) {
      if(apollo_get_ack() == 0) {
        return (1);
      } 
  }

  /*  do_debug("going on to longer wait i %d\n", i); */

  return wait_for_ack_irq();
}

static void SEND_COMMAND(unsigned char data) { 
  if(apollo_base_map == NULL) {
    printk("apollo_base_map = NULL\n");
    return;
  }

  apollo_write_reg(reg_GPSR2, 1 << 3);
  apollo_write_reg(reg_GPSR1, data << 26);
  apollo_write_reg(reg_GPSR2, data >> 6);
  apollo_write_reg(reg_GPCR1, (0x000000ff & ~data) << 26);
  apollo_write_reg(reg_GPCR2, (0x000000ff & ~data) >> 6);
  /* now give a DS */
  apollo_write_reg(reg_GPCR2, 1 << 2);
  wait_for_ack();
  apollo_write_reg(reg_GPSR2, 1 << 2);
  /* set C/nD back to D*/
  apollo_write_reg(reg_GPCR2, 1 << 3);
  
  if (!FB_INITED()) {
    udelay(APOLLO_SEND_DELAY);
  }
  
  do_debug("send command: 0x%x\n",data);
}

void send_command(unsigned char data) {
  /* Send status command before ANY other command.      */
  SEND_COMMAND(APOLLO_GET_STATUS_CMD);
  
  /* If the command isn't status, do a dummy read,      */
  /* then send the specified command.                   */
  if (APOLLO_GET_STATUS_CMD != data) {
    apollo_read();
    SEND_COMMAND(data);
  }
}

void send_data(register unsigned char data) {
  register unsigned int gpsr1, gpsr2, gpcr1, gpcr2;
  
  gpsr1 = data << 26;
  gpsr2 = data >> 6;
  gpcr1 = (0x000000ff & ~data) << 26;
  gpcr2 = (0x000000ff & ~data) >> 6;

  /* first setup data on LDD7:0 */
  apollo_write_reg(reg_GPSR1, gpsr1);
  apollo_write_reg(reg_GPSR2, gpsr2);
  apollo_write_reg(reg_GPCR1, gpcr1);
  apollo_write_reg(reg_GPCR2, gpcr2);
  /* now give a DS */
  apollo_write_reg(reg_GPCR2, 1 << 2);
  wait_for_ack();
  apollo_write_reg(reg_GPSR2, 1 << 2);
  
  if (!FB_INITED()) {
    udelay(APOLLO_SEND_DELAY);
  }
  
  do_debug("send data: 0x%x\n",data);
}

void preack_init(register unsigned char data) {
  register unsigned int gpsr1, gpsr2, gpcr1, gpcr2;

  gpsr1 = data << 26;
  gpsr2 = data >> 6;
  gpcr1 = (0x000000ff & ~data) << 26;
  gpcr2 = (0x000000ff & ~data) >> 6;

  /* first setup data on LDD7:0 */
  apollo_write_reg(reg_GPSR1, gpsr1);
  apollo_write_reg(reg_GPSR2, gpsr2);
  apollo_write_reg(reg_GPCR1, gpcr1);
  apollo_write_reg(reg_GPCR2, gpcr2);
  /* now give a DS */
  apollo_write_reg(reg_GPCR2, 1 << 2);
}

void preack_end(void) {
  wait_for_ack();
  apollo_write_reg(reg_GPSR2, 1 << 2);
}

void send_data_preack(register unsigned char data) {
  register unsigned int gpsr1, gpsr2, gpcr1, gpcr2;
  gpsr1 = data << 26;
  gpsr2 = data >> 6;
  gpcr1 = (0x000000ff & ~data) << 26;
  gpcr2 = (0x000000ff & ~data) >> 6;
  wait_for_ack();
  apollo_write_reg(reg_GPSR2, 1 << 2);
  apollo_write_reg(reg_GPSR1, gpsr1);
  apollo_write_reg(reg_GPSR2, gpsr2);
  apollo_write_reg(reg_GPCR1, gpcr1);
  apollo_write_reg(reg_GPCR2, gpcr2);
  apollo_write_reg(reg_GPCR2, 1 << 2);
}

void data2highz(void)
{
  u32 r1, r2;

  if(apollo_base_map == NULL) return;
  /* The Data bus is mapped on GPIO 58 - 65
  *  IO 58 = bit 26 of register 1.
  *  IO 63 = bit 32 (MSB of register 1.)
  *  IO 64 = bit 0 of register 2.
  *  The GPIO can be set to input by writting 0 in the GPDRx register. */

  r1 = apollo_read_reg(reg_GPDR1);
  r2 = apollo_read_reg(reg_GPDR2);

  apollo_write_reg(reg_GPDR1, (r1 & 0x03ffffff)); /*clear upper 6 bits*/
  apollo_write_reg(reg_GPDR2, (r2 & 0xfffffffc)); /*clear lower 2 bits*/  
}

void data2out(void)
{
  u32 r1, r2;
  if(apollo_base_map == NULL) return;
  /*   The Data bus is mapped on GPIO 58 - 65
  *    IO 58 = bit 26 of register 1.
  *    IO 63 = bit 32 (MSB of register 1.)
  *    IO 64 = bit 0 of register 2.
  *    The GPIO can be set to output by writting 1 in the GPDRx register.  */

  r1 = apollo_read_reg(reg_GPDR1);
  r2 = apollo_read_reg(reg_GPDR2);

  apollo_write_reg(reg_GPDR1, (r1 | 0xfc000000)); /*set upper 6 bits*/
  apollo_write_reg(reg_GPDR2, (r2 | 0x00000003)); /*set lower 2 bits*/  
}

u32 apollo_read(void)
{
  /* define local variable*/
  u32 temp1, temp2;
  
  if(apollo_base_map == NULL) return 0;

  data2highz();
  /* Set RW high = GPIO69 = GPSR2 bit 5 */
  apollo_write_reg(reg_GPSR2, 1 << 5);
  /* Set CD low = GPIO67 = GPCR2 bit 3 */
  apollo_write_reg(reg_GPCR2, 1 << 3);

  /* now give a DS (to zero) = GPIO66 = GPCR2 bit 2*/
  apollo_write_reg(reg_GPCR2, 1 << 2);
  wait_for_ack();
  /* read data*/
  
  /*read upper 6 bits*/
  temp1 = (apollo_read_reg(reg_GPLR1) & 0xFC000000)  >> 26;
  
  /*read lower 2 bits*/
  temp2 = (apollo_read_reg(reg_GPLR2) & 0x00000003)  << 6;
  temp2 = temp2 | temp1;    
  
  /* set DS back high = GPIO66 = GPSR2 bit 2*/
  apollo_write_reg(reg_GPSR2, 1 << 2);
  /* Set RW low = GPIO69 = GPCR2 bit 5 */
  apollo_write_reg(reg_GPCR2, 1 << 5);
  data2out();
  return temp2;
}

int apollo_get_status(apollo_status_t *status) {
  unsigned char local_status;
  int status_valid = 0;
  
  send_command(APOLLO_GET_STATUS_CMD);
  local_status = (unsigned char)apollo_read();
 
  if (status) {
    memcpy(status, &local_status, sizeof(apollo_status_t));
    status_valid = 1;
  }
  
  return status_valid;
}

unsigned char apollo_get_version(void) {
  send_command(APOLLO_GET_VERSION_CMD);
  return apollo_read();
}

unsigned char apollo_get_temperature(void) {
  send_command(APOLLO_GET_TEMPERATURE_CMD);
  return apollo_read();
}

unsigned char apollo_get_init_display_speed(void) {
  unsigned long init_display_flag = get_apollo_init_display_flag();
  unsigned char init_display_speed;
  
  switch (init_display_flag) {
    case INIT_DISPLAY_FAST:
      init_display_speed = APOLLO_INIT_DISPLAY_FAST;
      break;
      
    case INIT_DISPLAY_SKIP:
      init_display_speed = APOLLO_INIT_DISPLAY_SKIP;
      break;
    
    case INIT_DISPLAY_WAKE:
      init_display_speed = APOLLO_INIT_DISPLAY_WAKE;
      break;
    
    case INIT_DISPLAY_SLOW:
    default:
      init_display_speed = APOLLO_INIT_DISPLAY_SLOW;
      break;
  }

  return init_display_speed;
}

void apollo_set_init_display_speed(unsigned char init_display_speed) {
  unsigned long init_display_flag;
  
  switch (init_display_speed) {
    case APOLLO_INIT_DISPLAY_FAST:
      init_display_flag = INIT_DISPLAY_FAST;
      break;
      
    case APOLLO_INIT_DISPLAY_SKIP:
      init_display_flag = INIT_DISPLAY_SKIP;
      break;
      
    case APOLLO_INIT_DISPLAY_WAKE:
      init_display_flag = INIT_DISPLAY_WAKE;
      break;
      
    case APOLLO_INIT_DISPLAY_SLOW:
    default:
      init_display_flag = INIT_DISPLAY_SLOW;
      break;
  }
  
  set_apollo_init_display_flag(init_display_flag);
}

u32 apollo_init_display(unsigned char init_display_speed) {
  send_command(APOLLO_INIT_DISPLAY_CMD);
  send_data(init_display_speed);
  return 0;
}

void apollo_set_fx_exclude_rect(int x1, int y1, int x2, int y2) {
  unsigned long bpp = eink_get_bpp();
  
  apollo_fx_exclude_rect  = 1;
  apollo_fx_exclude_x1    = (x1 * bpp)/8;
  apollo_fx_exclude_y1    = y1;
  apollo_fx_exclude_x2    = (x2 * bpp)/8;
  apollo_fx_exclude_y2    = y2;
}

void apollo_clear_fx_exclude_rect(void) {
  apollo_fx_exclude_rect  = 0;
  apollo_fx_exclude_x1    = 0;
  apollo_fx_exclude_y1    = 0;
  apollo_fx_exclude_x2    = 0;
  apollo_fx_exclude_y2    = 0;
}

int apollo_fx_in_exclude_rect(void) {
  int in_exclude_rect_status = 0;
  
  if (apollo_fx_exclude_rect) {
    if ((apollo_fx_exclude_x >= apollo_fx_exclude_x1) &&
        (apollo_fx_exclude_x <  apollo_fx_exclude_x2) &&
        (apollo_fx_exclude_y >= apollo_fx_exclude_y1) &&
        (apollo_fx_exclude_y <  apollo_fx_exclude_y2)) {
      in_exclude_rect_status = 1;
    }
  }

  return in_exclude_rect_status;
}

void apollo_set_display_fx(int which_fx) {
  // Default to no FX.
  apollo_display_fx = APOLLO_FX_NONE;
  
  switch (which_fx) {
    case APOLLO_FX_LIGHTEN_DARK_GRAY:
    case APOLLO_FX_LIGHTEN_LITE_GRAY:
    case APOLLO_FX_LIGHTEN_WHITE:
    case APOLLO_FX_DARKEN_LITE_GRAY:
    case APOLLO_FX_DARKEN_DARK_GRAY:
    case APOLLO_FX_DARKEN_BLACK:
    
    case APOLLO_FX_STIPPLE_LIGHTEN_DARK:
    case APOLLO_FX_STIPPLE_LIGHTEN_LITE:
    case APOLLO_FX_STIPPLE_DARKEN_LITE:
    case APOLLO_FX_STIPPLE_DARKEN_DARK:
    case APOLLO_FX_STIPPLE_DARK_GRAY:
    case APOLLO_FX_STIPPLE_LITE_GRAY:
   
    case APOLLO_FX_MASK:
      // Can't do any of the above FX unless we're at 2bpp.
      if (EINK_DEPTH_2BPP == eink_get_bpp())
        goto fx_common;
    break;

    case APOLLO_FX_INVERT:
    fx_common:
      apollo_display_fx = which_fx;
      break;
  }
}

int apollo_get_display_fx(void) {
  return apollo_display_fx;
}

void apollo_set_mask_fx(int which_fx) {
  // Default to no FX.
  apollo_mask_fx = APOLLO_FX_NONE;
  
  switch (which_fx) {
    case APOLLO_FX_LIGHTEN_DARK_GRAY:
    case APOLLO_FX_LIGHTEN_LITE_GRAY:
    case APOLLO_FX_LIGHTEN_WHITE:
    case APOLLO_FX_DARKEN_LITE_GRAY:
    case APOLLO_FX_DARKEN_DARK_GRAY:
    case APOLLO_FX_DARKEN_BLACK:

    case APOLLO_FX_STIPPLE_LIGHTEN_DARK:
    case APOLLO_FX_STIPPLE_LIGHTEN_LITE:
    case APOLLO_FX_STIPPLE_DARKEN_LITE:
    case APOLLO_FX_STIPPLE_DARKEN_DARK:
    case APOLLO_FX_STIPPLE_DARK_GRAY:
    case APOLLO_FX_STIPPLE_LITE_GRAY:
      // Can't do any of the above FX unless we're at 2bpp.
      if (EINK_DEPTH_2BPP == eink_get_bpp())
        goto fx_common;
    break;

    case APOLLO_FX_INVERT:
    fx_common:
      apollo_mask_fx = which_fx;
      break;
  }
}

int apollo_get_mask_fx(void) {
  return apollo_mask_fx;
}

static unsigned char apollo_shade(unsigned char data) {
  return (apollo_shade_table[(0xF0 & data) >> 4] << 4) | apollo_shade_table[(0x0F & data)];
}

static unsigned char apollo_stipple(unsigned char data) {
  int index = (apollo_stipple_rows & 1) ? stipple_odd_row_mask : stipple_even_row_mask;
  u8  mask  = apollo_stipple_table[index++], mask_type = apollo_stipple_table[index];

  if (0 == (++apollo_stipple_bytes % apollo_stipple_rowbytes)) {
    apollo_stipple_rows++;
  }

  return STIPPLE_TABLE_MASK(mask_type, mask, data);
}

static unsigned char apollo_mask(unsigned char data, int i) {
  register unsigned char mask, byte;

  byte = apollo_mask_framebuffer[(apollo_mask_fb_rowbytes * apollo_mask_y) + apollo_mask_x++];
  byte = apollo_apply_fx(byte, 0, apollo_mask_fx);

  mask = apollo_mask_buffer[apollo_mask_bufsize + i];
  
  if (0 == ((apollo_mask_x - apollo_mask_x_start) % apollo_mask_rowbytes)) {
    apollo_mask_x = apollo_mask_x_start;
    apollo_mask_y++;
  }

  return (data & mask) | (byte & ~mask);
}

static void apollo_begin_fx(unsigned char *buffer, int bufsize, int rowbytes, int x_start, int y_start)
{
  int which_fx = apollo_display_fx;
  
  reapply_fx:
  switch (which_fx) {
    case APOLLO_FX_LIGHTEN_DARK_GRAY:
      apollo_shade_table      = lighten_2bpp_nybble_table_dark_gray;
      break;
    case APOLLO_FX_LIGHTEN_LITE_GRAY:
      apollo_shade_table      = lighten_2bpp_nybble_table_lite_gray;
      break;
    case APOLLO_FX_DARKEN_LITE_GRAY:
      apollo_shade_table      = darken_2bpp_nybble_table_lite_gray;
      break;
    case APOLLO_FX_DARKEN_DARK_GRAY:
      apollo_shade_table      = darken_2bpp_nybble_table_dark_gray;
      break;
    
    case APOLLO_FX_STIPPLE_LIGHTEN_DARK:
      apollo_stipple_table    = stipple_lighten_dark_gray_2bpp_table;
      goto stipple_common;
    case APOLLO_FX_STIPPLE_LIGHTEN_LITE:
      apollo_stipple_table    = stipple_lighten_lite_gray_2bpp_table;
      goto stipple_common;
    case APOLLO_FX_STIPPLE_DARKEN_LITE:
      apollo_stipple_table    = stipple_darken_lite_gray_2bpp_table;
      goto stipple_common;
    case APOLLO_FX_STIPPLE_DARKEN_DARK:
      apollo_stipple_table    = stipple_darken_dark_gray_2bpp_table;
      goto stipple_common;
    
    case APOLLO_FX_STIPPLE_DARK_GRAY:
      apollo_stipple_table    = stipple_dark_gray_2bpp_table;
      apollo_shade_table      = noblack_2bpp_nybble_table;
      goto stipple_common;
    case APOLLO_FX_STIPPLE_LITE_GRAY:
      apollo_stipple_table    = stipple_lite_gray_2bpp_table;
      apollo_shade_table      = lighten_2bpp_nybble_table_dark_gray;
      /* goto stipple_common; */

    stipple_common:
      apollo_stipple_bytes    = 0;
      apollo_stipple_rows     = 0;
      apollo_stipple_rowbytes = rowbytes;
      break;
    
    case APOLLO_FX_MASK:
      apollo_mask_framebuffer = eink_get_framebuffer();
      apollo_mask_fb_rowbytes = APOLLO_ROWBYTES_2BPP;
      apollo_mask_buffer      = buffer;
      apollo_mask_bufsize     = bufsize;
      apollo_mask_rowbytes    = rowbytes;
      apollo_mask_x_start     = x_start;
      apollo_mask_x           = x_start;
      apollo_mask_y           = y_start;
      which_fx                = apollo_mask_fx;
      goto reapply_fx;
  }

  if (apollo_fx_exclude_rect) {
    apollo_fx_exclude_x = apollo_fx_exclude_y = 0;
  }
}

static void apollo_end_fx(void) {
  int which_fx = apollo_display_fx;
  
  reapply_fx:
  switch (which_fx) {
    case APOLLO_FX_LIGHTEN_DARK_GRAY:
    case APOLLO_FX_LIGHTEN_LITE_GRAY:
    case APOLLO_FX_DARKEN_LITE_GRAY:
    case APOLLO_FX_DARKEN_DARK_GRAY:
      apollo_shade_table      = NULL;
      break;

    case APOLLO_FX_STIPPLE_LIGHTEN_DARK:
    case APOLLO_FX_STIPPLE_LIGHTEN_LITE:
    case APOLLO_FX_STIPPLE_DARKEN_LITE:
    case APOLLO_FX_STIPPLE_DARKEN_DARK:
      goto stipple_common;
    
    case APOLLO_FX_STIPPLE_DARK_GRAY:
    case APOLLO_FX_STIPPLE_LITE_GRAY:
      apollo_shade_table      = NULL;
      /* goto stipple_common; */

    stipple_common:
      apollo_stipple_bytes    = 0;
      apollo_stipple_rows     = 0;
      apollo_stipple_rowbytes = 0;
      apollo_stipple_table    = NULL;
      break;
    
    case APOLLO_FX_MASK:
      apollo_mask_framebuffer = NULL;
      apollo_mask_fb_rowbytes = 0;;
      apollo_mask_buffer      = NULL;
      apollo_mask_bufsize     = 0;
      apollo_mask_rowbytes    = 0;
      apollo_mask_x_start     = 0;
      apollo_mask_x           = 0;
      apollo_mask_y           = 0;
      which_fx                = apollo_mask_fx;
      goto reapply_fx;
  }

  if (apollo_fx_exclude_rect) {
    apollo_fx_exclude_x = apollo_fx_exclude_y = 0;
  }
}

static unsigned char apollo_apply_fx(unsigned char data, int i, int which_fx) {
  unsigned char fx_data = data;
  int exclude = 0;
  
  if (apollo_fx_exclude_rect) {
    if (apollo_fx_in_exclude_rect()) {
      exclude = 1;
    }

    if (0 == (++apollo_fx_exclude_x % APOLLO_ROWBYTES_2BPP)) {
      apollo_fx_exclude_x = 0;
      apollo_fx_exclude_y++;
    }
  }

  switch (which_fx) {
    case APOLLO_FX_LIGHTEN_DARK_GRAY:
    case APOLLO_FX_LIGHTEN_LITE_GRAY:
    case APOLLO_FX_DARKEN_LITE_GRAY:
    case APOLLO_FX_DARKEN_DARK_GRAY:
      fx_data = apollo_shade(data);
      break;
    
    case APOLLO_FX_LIGHTEN_WHITE:
      fx_data = APOLLO_2BPP_WHITE;
      break;
    
    case APOLLO_FX_DARKEN_BLACK:
      fx_data = APOLLO_2BPP_BLACK;
      break;
  
    case APOLLO_FX_INVERT:
      fx_data = ~data;
      break;
      
    case APOLLO_FX_STIPPLE_LIGHTEN_DARK:
    case APOLLO_FX_STIPPLE_LIGHTEN_LITE:
    case APOLLO_FX_STIPPLE_DARKEN_LITE:
    case APOLLO_FX_STIPPLE_DARKEN_DARK:
      fx_data = apollo_stipple(data);
      break;
    case APOLLO_FX_STIPPLE_DARK_GRAY:
    case APOLLO_FX_STIPPLE_LITE_GRAY:
      fx_data = apollo_shade(apollo_stipple(data));
      break;
      
    case APOLLO_FX_MASK:
      fx_data = apollo_mask(data, i);
      break;
  }

  return exclude ? data : fx_data;
}

static u32 apollo_load_data(unsigned char cmd, unsigned char *buffer, int bufsize,
  u16 x1, u16 y1, u16 x2, u16 y2) {
  unsigned long bpp = eink_get_bpp();
  int i, rowbytes = EINK_DEPTH_2BPP == bpp ? APOLLO_ROWBYTES_2BPP : APOLLO_ROWBYTES_1BPP;
  unsigned char data, white, black;
  u32 data_cleared = 0;

  data = white = black = buffer[0];
  send_command(cmd);

  if (APOLLO_LOAD_PARTIAL_IMAGE_CMD == cmd) {
      rowbytes = (((x2 - x1) + 1) * bpp)/8;
      
      send_data((x1 >> 8) & 0xff);
      send_data(x1 & 0xff);
      
      send_data((y1 >> 8) & 0xff);
      send_data(y1 & 0xff);
      
      send_data((x2 >> 8) & 0xff);
      send_data(x2 & 0xff);
      
      send_data((y2 >> 8) & 0xff);
      send_data(y2 & 0xff);
  }
  
  apollo_begin_fx(buffer, bufsize, rowbytes, (x1 * bpp)/8, y1);
  
#ifdef WRITE_FX_BACK_TO_BUFFER
  buffer[0] = apollo_apply_fx(data, 0, apollo_display_fx);
  preack_init(buffer[0]);
#else
  preack_init(apollo_apply_fx(data, 0, apollo_display_fx));
#endif

  for(i = 1; i < bufsize; i++) {
    data = buffer[i]; 

#ifdef WRITE_FX_BACK_TO_BUFFER
    buffer[i] = apollo_apply_fx(data, i, apollo_display_fx);
    send_data_preack(buffer[i]);
#else
    send_data_preack(apollo_apply_fx(data, i, apollo_display_fx));
#endif

    switch (bpp) {
      case EINK_DEPTH_1BPP:
        white &= data;
        black |= data;
      break;
      
      case EINK_DEPTH_2BPP:
        white |= data;
        black &= data;
      break;
    }
  }

  preack_end();
  
  send_command(APOLLO_STOP_IMGLD_CMD);
  
  apollo_end_fx();

  switch (bpp) {
    case EINK_DEPTH_1BPP:
      if ((APOLLO_1BPP_WHITE == white) || (APOLLO_1BPP_BLACK == black))
        data_cleared = 1;
    break;
    
    case EINK_DEPTH_2BPP:
      if ((APOLLO_2BPP_WHITE == white) || (APOLLO_2BPP_BLACK == black))
        data_cleared = 1;
    break;
  }

  return data_cleared;
}

u32 apollo_load_image_data(unsigned char *buffer, int bufsize) {
  return apollo_load_data(APOLLO_LOAD_IMAGE_CMD, buffer, bufsize, 0, 0, 0, 0);
}

u32 apollo_load_partial_data(unsigned char *buffer, int bufsize, 
  u16 x1, u16 y1, u16 x2, u16 y2) {
  return apollo_load_data(APOLLO_LOAD_PARTIAL_IMAGE_CMD, buffer, bufsize, x1, y1, --x2, --y2);
}

u32 apollo_display(void) {
  send_command(APOLLO_DISP_IMAGE_CMD);
  return 0;
}

u32 apollo_display_partial(void) {
  send_command(APOLLO_DISP_PARTIAL_IMAGE_CMD);
  return 0;
}

u32 apollo_restore(void) {
    send_command(APOLLO_RESTORE_IMAGE_CMD);
    return 0;
}

u32 apollo_erase_screen(unsigned char color) {
  send_command(APOLLO_ERASE_SCRN_CMD);
  send_data(color);
  return 0;
}

void apollo_clear_screen(unsigned long bpp, unsigned char data, int full) {
	unsigned char fore_color = APOLLO_ERASE_SCREEN_BLACK,
	              back_color = APOLLO_ERASE_SCREEN_WHITE;
	
	switch (bpp) {
	  case EINK_DEPTH_1BPP:
	    if (data == APOLLO_1BPP_BLACK) {
	      fore_color = APOLLO_ERASE_SCREEN_WHITE;
	      back_color = APOLLO_ERASE_SCREEN_BLACK;
	    }
	    break;
	  
	  case EINK_DEPTH_2BPP:
	    if (data == APOLLO_2BPP_BLACK) {
	      fore_color = APOLLO_ERASE_SCREEN_WHITE;
	      back_color = APOLLO_ERASE_SCREEN_BLACK;
	    }
	    break;
	}

  if (full) {
	  apollo_erase_screen(fore_color);
	}
	
	apollo_erase_screen(back_color);
}

void apollo_temp_sensor_init(void) {
  if (needs_to_enable_temp_sensor()) {
    pxa_gpio_mode(GPIO_FIONA_TEMP_SENSOR_MD);
    pxa_gpio_mode(GPIO_FIONA_APOLLO_COMMON_MD);
    pxa_gpio_mode(GPIO_FIONA_APOLLO_POWER_MD);
  }
}

void apollo_power_up(void) {
  if (needs_to_enable_temp_sensor()) {
    GPCR(GPIO_FIONA_TEMP_SENSOR) = GPIO_bit(GPIO_FIONA_TEMP_SENSOR);
  }
    
  GPSR(GPIO_FIONA_APOLLO_COMMON) = GPIO_bit(GPIO_FIONA_APOLLO_COMMON);
  GPSR(GPIO_FIONA_APOLLO_POWER)	 = GPIO_bit(GPIO_FIONA_APOLLO_POWER);

  apollo_temp_sensor_init();
  
  GPSR(68) = GPIO_bit(68);
  GPCR(69) = GPIO_bit(69);
  GPCR(70) = GPIO_bit(70);
  
  GPDR(83) &= ~GPIO_bit(83);
  GPSR(83) = GPIO_bit(83);

  GAFR(58) &= ~GAFR_bit(58);
  GAFR(59) &= ~GAFR_bit(59);
  GAFR(60) &= ~GAFR_bit(60);
  GAFR(61) &= ~GAFR_bit(61);
  GAFR(62) &= ~GAFR_bit(62);
  GAFR(63) &= ~GAFR_bit(63);
  GAFR(64) &= ~GAFR_bit(64);
  GAFR(65) &= ~GAFR_bit(65);
  GAFR(66) &= ~GAFR_bit(66);
  GAFR(67) &= ~GAFR_bit(67);
  GAFR(68) &= ~GAFR_bit(68);
  GAFR(69) &= ~GAFR_bit(69);
  GAFR(70) &= ~GAFR_bit(70);

  GPDR(58) |= GPIO_bit(58);
  GPDR(59) |= GPIO_bit(59);
  GPDR(60) |= GPIO_bit(60);
  GPDR(61) |= GPIO_bit(61);
  GPDR(62) |= GPIO_bit(62);
  GPDR(63) |= GPIO_bit(63);
  GPDR(64) |= GPIO_bit(64);
  GPDR(65) |= GPIO_bit(65);
  GPDR(66) |= GPIO_bit(66);
  GPDR(67) |= GPIO_bit(67);
  GPDR(68) |= GPIO_bit(68);
  GPDR(69) |= GPIO_bit(69);
  GPDR(70) |= GPIO_bit(70);
  
  apollo_comm_init();
}

void apollo_power_down(int full) {
  if (full) {
    apollo_comm_shutdown();
  }
  
  if (needs_to_enable_temp_sensor()) {
    GPSR(GPIO_FIONA_TEMP_SENSOR) = GPIO_bit(GPIO_FIONA_TEMP_SENSOR);
  }

  GPCR(GPIO_FIONA_APOLLO_COMMON) = GPIO_bit(GPIO_FIONA_APOLLO_COMMON);
  GPCR(GPIO_FIONA_APOLLO_POWER)  = GPIO_bit(GPIO_FIONA_APOLLO_POWER);
}

void apollo_init(unsigned long bpp) {
  unsigned char init_display_speed = apollo_get_init_display_speed();
  
  // Waking temporarily from kernel?  Yes, skip everything.
  //
  if (APOLLO_INIT_DISPLAY_SKIP != init_display_speed) {
    
    // Waking from bootloader?  Yes, just wake.
    //
    if (APOLLO_INIT_DISPLAY_WAKE == init_display_speed) {
      send_command(APOLLO_NORMAL_MODE_CMD);
    
    } else {

      // Otherwise, go through normal init sequence.
      //
      apollo_write_register(APOLLO_OFF_DELAY_REGISTER, APOLLO_OFF_DELAY_DEFAULT);
      apollo_write_register(APOLLO_T1_DELAY_REGISTER,  APOLLO_T1_DELAY_DEFAULT);
      
      apollo_init_display(init_display_speed);
      
      send_command(APOLLO_CANCEL_AUTO_REFRESH_CMD);
    
      send_command(APOLLO_SET_ORIENTATION_CMD);
      send_data(system_rev ? APOLLO_ORIENTATION_270 : APOLLO_ORIENTATION_90);
    
      switch (bpp) {
        case EINK_DEPTH_1BPP:
          send_command(APOLLO_SET_DEPTH_CMD);
          send_data(APOLLO_DEPTH_1BPP);
          send_command(INVERT_VIDEO() ? APOLLO_NEGATIVE_PICTURE_CMD : APOLLO_POSITIVE_PICTURE_CMD);
          break;
    
        case EINK_DEPTH_2BPP:
          send_command(APOLLO_SET_DEPTH_CMD);
          send_data(APOLLO_DEPTH_2BPP);
          send_command(INVERT_VIDEO() ? APOLLO_POSITIVE_PICTURE_CMD : APOLLO_NEGATIVE_PICTURE_CMD);
          break;
      }
      
      // Say that we're init'ed.
      //
      set_fb_init_flag(BOOT_GLOBALS_FB_INIT_FLAG);
    }
  }
}

int apollo_write_register(unsigned char regAddress, unsigned char data) {
  send_command(APOLLO_WRITE_REGISTER);
  send_data(regAddress);
  send_data(data);
  return 0;
}

int apollo_read_register(unsigned char regAddress, unsigned char *data) {
  send_command(APOLLO_READ_REGISTER);
  send_data(regAddress);
  *data = apollo_read();
  return 0;
}

int apollo_read_from_flash_byte(unsigned long addr, unsigned char *data) {
  send_command(APOLLO_FLASH_READ);

  send_data((unsigned char)((addr & 0x00FF0000) >> 16));
  send_data((unsigned char)((addr & 0x0000FF00) >> 8));
  send_data((unsigned char)((addr & 0x000000FF)));

  *data = apollo_read();
  return 1;
}

int apollo_read_from_flash_short(unsigned long addr, unsigned short *data) {
  unsigned char byte;
  int i;
  
  for (i = 0, *data = 0; i < sizeof(unsigned short); i++) {
    apollo_read_from_flash_byte((addr + i), &byte);
    *data |= (byte << (i * 8));
  }
  
  return 1;
}

int apollo_read_from_flash_long(unsigned long addr, unsigned long *data) {
  unsigned char byte;
  int i;

  for (i = 0, *data = 0; i < sizeof(unsigned long); i++) {
    apollo_read_from_flash_byte((addr + i), &byte);
    *data |= (byte << (i * 8));
  }
  
  return 1;
}

void apollo_get_temperature_info(apollo_temperature_t *temperature) {
  if (temperature) {
    apollo_info_t info; apollo_get_info(&info);
    
    temperature->min = info.waveform_temp_min;
    temperature->max = info.waveform_temp_max;
    temperature->inc = info.waveform_temp_inc;
    temperature->cur = apollo_get_temperature();
    
    temperature->valid = !LEGACY_WAVEFORM_DATA(info.waveform_temp_min);
  }
}

void apollo_override_fpl(apollo_fpl_t *fpl) {
  panel_id_t *panel_id = get_panel_id();

  if ((panel_id && panel_id->override) && fpl) {
    fpl->platform            = panel_id->platform;
    fpl->lot                 = panel_id->lot;
    fpl->adhesive_run_number = panel_id->adhesive_run_number;
    
    fpl->overridden          = 1;
  }
}
