/*
 * e-ink display framebuffer driver
 * 
 * Copyright (C)2005-2007 Lab126, Inc.  All rights reserved.
 */

#ifndef _APOLLO_H
#define _APOLLO_H

#ifndef FOR_UBOOT_BUILD //------------- Linux Build

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/fb.h>
#include <linux/version.h>
#include <linux/sched.h>
#include <linux/syscalls.h>

#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/segment.h>
#include <asm/hardware.h>
#include <asm/io.h>

#include <asm/arch/fiona.h>
#include <asm/arch/fpow.h>
#include <asm/delay.h>
#include <asm/arch/pxa-regs.h>

#include <asm/arch/boot_globals.h>

extern wait_queue_head_t apollo_ack_wq;
extern void apollo_init_irq(void);
extern void apollo_shutdown_irq(void);
extern int apollo_init_kernel(int full);
extern void apollo_done_kernel(int full);

extern int g_fiona_power_debug;
extern int eink_debug;

#define einkdebug(x...)	if (g_fiona_power_debug || eink_debug) printk(x)

#else   // ---------------------------- UBoot Build

#include <common.h>
#include <asm/arch/pxa-regs.h>

#include "../board/fiona/ub_boot_globals.h"

#define init_waitqueue_head(q)		do {} while(0)
#define apollo_init_irq()			do {} while(0)
#define apollo_shutdown_irq()		do {} while(0)

#define writel(v,r)					(*(r) = (v))
#define readl(r)					(*(r))
#define ioremap_nocache(a,l)		((a))
#define iounmap(a)					do {} while(0)
#define printk						printf

#define HZ                          CFG_HZ
#define system_rev					1

extern void pxa_gpio_mode(int gpio_mode);
extern int do_reboot(void);

extern bool g_ack_timed_out;

#endif  // ----------------------------

#include "apollo_waveform.h"		// EINK_WAVEFORM_XXX defines and apollo_XXX_t structs

typedef u32 * volatile reg_addr_t;

#undef WRITE_FX_BACK_TO_BUFFER
#undef ACTIVATE_DO_DEBUG
#ifdef ACTIVATE_DO_DEBUG
#define do_debug(x...) printk(x)
#else
#define do_debug(x...)
#endif

#define	GAFR_bit(x) (3 << (((x) & 0x0f) << 1))

#define MAX_TIMEOUT  (HZ * 4)
#define FAST_TIMEOUT 1000
#define MAX_TIMEOUTS 5

#define BASE_ADDR 0x40E00000
#define MAP_SIZE 4096
#define MAP_MASK (MAP_SIZE - 1)

#define GPSR1_ADDR 0x40E0001C
#define GPSR2_ADDR 0x40E00020
#define GPCR1_ADDR 0x40E00028
#define GPCR2_ADDR 0x40E0002C
#define GPLR1_ADDR 0x40E00004
#define GPLR2_ADDR 0x40E00008

#define GPDR1_ADDR 0x40E00010
#define GPDR2_ADDR 0x40E00014

// #define apollo_write_reg(reg, val) {printk("APOLLO WRITE: Addr: 0x%x  ::  Value: 0x%x\n",reg,val);writel((val), (reg));}
#define apollo_write_reg(reg, val) writel((val), (reg))
#define apollo_read_reg(reg) readl((reg))

#define APOLLO_WRITE_REGISTER			0x10
#define APOLLO_READ_REGISTER			0x11

#define APOLLO_GET_TEMPERATURE_CMD		0x21

#define APOLLO_LOAD_IMAGE_CMD			0xA0
#define APOLLO_STOP_IMGLD_CMD			0xA1
#define APOLLO_DISP_IMAGE_CMD			0xA2
#define APOLLO_ERASE_SCRN_CMD			0xA3
#define APOLLO_INIT_DISPLAY_CMD			0xA4
#define APOLLO_RESTORE_IMAGE_CMD		0xA5

#define APOLLO_GET_STATUS_CMD			0xAA

#define APOLLO_LOAD_PARTIAL_IMAGE_CMD	0xB0
#define APOLLO_DISP_PARTIAL_IMAGE_CMD	0xB1

#define APOLLO_GET_VERSION_CMD			0xE0
#define APOLLO_RESET_CMD				0xEE

#define APOLLO_NORMAL_MODE_CMD			0xF0
#define APOLLO_SLEEP_MODE_CMD			0xF1
#define APOLLO_STANDBY_MODE_CMD			0xF2

#define APOLLO_SET_DEPTH_CMD			0xF3
#define APOLLO_SET_ORIENTATION_CMD		0xF5
#define APOLLO_POSITIVE_PICTURE_CMD		0xF7
#define APOLLO_NEGATIVE_PICTURE_CMD		0xF8
#define APOLLO_CANCEL_AUTO_REFRESH_CMD	0xFA
#define APOLLO_MANUAL_REFRESH_CMD		0xFC

#define APOLLO_FLASH_WRITE				0x01
#define APOLLO_FLASH_READ				0x02

#define APOLLO_FLASH_MAX_RETRIES		5

#define APOLLO_DEPTH_1BPP				0x00
#define APOLLO_DEPTH_2BPP				0x02

#define APOLLO_1BPP_WHITE				0xFF
#define APOLLO_1BPP_BLACK				0x00
#define APOLLO_2BPP_WHITE				0x00
#define APOLLO_2BPP_BLACK				0xFF

#define APOLLO_ORIENTATION_0			0x00
#define APOLLO_ORIENTATION_90			0x01
#define APOLLO_ORIENTATION_180			0x02
#define APOLLO_ORIENTATION_270			0x03

// For use with apollo_set_init_display_speed() & apollo_get_init_display_speed().
// These are local to either the kernel or bootloader.
//
#define APOLLO_INIT_DISPLAY_FAST		0x00
#define APOLLO_INIT_DISPLAY_SLOW		0x01
#define APOLLO_INIT_DISPLAY_SKIP		0x02		// flag, only used by kernel
#define APOLLO_INIT_DISPLAY_WAKE		0x03		// flag, only used bootloader

// For use with set_apollo_init_display_flag() && get_apollo_init_display_flag()
// These are global and are used between the kernel & bootloader.
//
#define INIT_DISPLAY_FAST				0x74736166
#define INIT_DISPLAY_SLOW				0x776F6C73
#define INIT_DISPLAY_SKIP				0x70696B63	// flag, only used by kernel
#define INIT_DISPLAY_WAKE				0x656B6177	// flag, only used bootloader

#define APOLLO_ERASE_SCREEN_BLACK		0x00
#define APOLLO_ERASE_SCREEN_WHITE		0x01
#define APOLLO_ERASE_SCREEN_DARK_GRAY	0x02
#define APOLLO_ERASE_SCREEN_LITE_GRAY	0x03

#define APOLLO_STANDBY_CMD_DELAY		600		// uSec delay after Standby cmd is issued
#define APOLLO_SEND_DELAY				100		// uSec delay after sending cmds or data

#define APOLLO_T1_DELAY_REGISTER		0x20
#define APOLLO_T1_DELAY_DEFAULT			0x40

#define APOLLO_OFF_DELAY_REGISTER		0x22
#define APOLLO_OFF_DELAY_DEFAULT		0
#define APOLLO_OFF_DELAY_THRESH			4
#define APOLLO_OFF_DELAY_MAX			255

#define APOLLO_STANDBY_OFF_POWER_MODE	0		// standby + off
#define APOLLO_NORMAL_POWER_MODE		1
#define APOLLO_SLEEP_POWER_MODE			2
#define APOLLO_STANDBY_POWER_MODE		3		// just standby

#define APOLLO_UNKNOWN_POWER_MODE		42

#define IS_APOLLO_STANDBY_MODE(m)		((APOLLO_STANDBY_OFF_POWER_MODE == m)	||	\
										 (APOLLO_STANDBY_POWER_MODE == m))

#define APOLLO_POWER_DOWN_SLEEP			0
#define APOLLO_POWER_DOWN_FULL			1

#define APOLLO_FX_LIGHTEN_DARK_GRAY		0
#define APOLLO_FX_LIGHTEN_LITE_GRAY		1
#define APOLLO_FX_LIGHTEN_WHITE			2

#define APOLLO_FX_DARKEN_LITE_GRAY		3
#define APOLLO_FX_DARKEN_DARK_GRAY		4
#define APOLLO_FX_DARKEN_BLACK			5

#define APOLLO_FX_INVERT				6

#define APOLLO_FX_STIPPLE_LIGHTEN_DARK	7
#define APOLLO_FX_STIPPLE_LIGHTEN_LITE	8

#define APOLLO_FX_STIPPLE_DARKEN_LITE	9
#define APOLLO_FX_STIPPLE_DARKEN_DARK	10

#define APOLLO_FX_MASK					11

#define APOLLO_FX_STIPPLE_DARK_GRAY		12
#define APOLLO_FX_STIPPLE_LITE_GRAY		13

#define APOLLO_FX_NONE					(-1)

#define EINK_DEPTH_1BPP					1
#define EINK_DEPTH_2BPP					2

#define APOLLO_X_RES					600
#define APOLLO_Y_RES					800

#define APOLLO_ROWBYTES_1BPP			((APOLLO_X_RES * EINK_DEPTH_1BPP)/8)
#define APOLLO_ROWBYTES_2BPP			((APOLLO_X_RES * EINK_DEPTH_2BPP)/8)

#define	INIT_DISPLAY_FLAG_UNKNOWN(f)	(!((INIT_DISPLAY_FAST == f)	|| \
										   (INIT_DISPLAY_SLOW == f)))

#define snuz_apollo()					apollo_power_state(APOLLO_SLEEP_POWER_MODE)
#define ReadFromFlash(a,d)				apollo_read_from_flash_byte(a,d)

struct apollo_temperature_t
{
	unsigned char	min,
					max,
					inc,
					cur,
					valid;					// if true, min.max.inc aren't 0xFF
};
typedef struct apollo_temperature_t apollo_temperature_t;

struct apollo_status_t
{
	unsigned char	operation_mode : 1;
	unsigned char	screen		   : 2;
	unsigned char	auto_refresh   : 1;
	unsigned char	bpp			   : 1;
	unsigned char	reserved	   : 3;
};
typedef struct apollo_status_t apollo_status_t;

extern unsigned long eink_get_bpp(void);
extern u8 *eink_get_framebuffer(void);
extern void eink_clear_screen(void);

extern int needs_to_enable_temp_sensor(void);

extern void apollo_set_init_display_speed(unsigned char init_display_speed);
extern unsigned char apollo_get_init_display_speed(void);

extern unsigned char apollo_get_version(void);
extern unsigned char apollo_get_temperature(void);

extern void apollo_init(unsigned long bpp);
extern void apollo_wait(void);
extern u32  apollo_get_ack(void);
extern u32  apollo_read(void);
extern int 	apollo_get_status(apollo_status_t *status);
extern u32  apollo_init_display(unsigned char init_display_speed);
extern u32  apollo_erase_screen(unsigned char color);
extern void apollo_clear_screen(unsigned long bpp, unsigned char data, int full);
extern u32  apollo_load_image_data(unsigned char *buffer, int bufsize);
extern int  apollo_wake_on_activity(void);

extern u32 apollo_load_partial_data(unsigned char *buffer, int bufsize, 
			     u16 x1, u16 y1, u16 x2, u16 y2);
extern u32 apollo_display(void);
extern u32 apollo_display_partial(void);
extern u32 apollo_restore(void);

extern char *apollo_get_power_state_str(int state);
extern int apollo_set_power_state(int target_mode);
extern int apollo_get_power_state(void);

extern void apollo_set_display_fx(int which_fx);
extern int apollo_get_display_fx(void);
extern void apollo_set_mask_fx(int which_fx);
extern int apollo_get_mask_fx(void);

extern void apollo_set_fx_exclude_rect(int x1, int y1, int x2, int y2);
extern void apollo_clear_fx_exclude_rect(void);

extern void data2highz(void);
extern void data2out(void);
extern void send_data(unsigned char data);
extern void send_command(unsigned char command);
extern void apollo_power_up(void);
extern void apollo_power_down(int full);

extern int apollo_comm_init(void);
extern int apollo_comm_shutdown(void);

int apollo_write_register(unsigned char regAddress, unsigned char data);
int apollo_read_register(unsigned char regAddress, unsigned char *data);

int apollo_read_from_flash_byte(unsigned long addr, unsigned char *data);
int apollo_read_from_flash_short(unsigned long addr, unsigned short *data);
int apollo_read_from_flash_long(unsigned long addr, unsigned long *data);

extern void apollo_get_temperature_info(apollo_temperature_t *temperature);
extern void apollo_override_fpl(apollo_fpl_t *fpl);

extern int ReadFromRam(unsigned long Address, unsigned long *Data);
extern int WriteToRam(unsigned long Address, unsigned long Data);

extern int ProgramFlash(unsigned long start_addr, 
		unsigned char *buffer, unsigned long blen);

extern int ProgramRam(unsigned long start_addr, 
		unsigned char *buffer, unsigned long blen);

extern int wait_for_ack_irq(void);

#endif // _APOLLO_H
