#ifndef __FIONA_EVENTS_H__
#define __FIONA_EVENTS_H__

#include <linux/input.h>
#include <linux/keyboard.h>     // For KEY_WAS_RELEASED definition
#include <linux/fpow_events.h>  // For FPOW_EVENT definition

// Event Constants That Are Passed Up To Java
#define EVENT_BATTERY_LOW               KEY_SYSRQ
#define EVENT_BATTERY_CRITICAL          KEY_F1
#define EVENT_POWER_SWITCH              KEY_F2
#define EVENT_WAN_SWITCH                KEY_F3
#define EVENT_AC                        KEY_F4
#define EVENT_USB                       KEY_F5
#define EVENT_HEADPHONES                KEY_F6
#define EVENT_WAN_RING                  KEY_F7
#define EVENT_KEY_WAKE                  KEY_F8
#define EVENT_AUDIO                     KEY_F9
#define EVENT_MMC                       KEY_F10
#define EVENT_PASSWORD                  KEY_F11
#define EVENT_TIMEOUT                   KEY_F12
#define EVENT_SWEEPER                   KEY_RO
#define EVENT_ALT_TIMEOUTS              KEY_102ND


// Handy conversions for event constants (including pressed/released status)
#define EVENT_POWER_SWITCH_ON           EVENT_POWER_SWITCH
#define EVENT_POWER_SWITCH_OFF          (EVENT_POWER_SWITCH | KEY_WAS_RELEASED)
#define EVENT_WAN_SWITCH_ON             EVENT_WAN_SWITCH
#define EVENT_WAN_SWITCH_OFF            (EVENT_WAN_SWITCH | KEY_WAS_RELEASED)
#define EVENT_AC_INSERTED               EVENT_AC
#define EVENT_AC_REMOVED                (EVENT_AC | KEY_WAS_RELEASED)
#define EVENT_USB_PLUG_INSERTED         EVENT_USB
#define EVENT_USB_PLUG_REMOVED          (EVENT_USB | KEY_WAS_RELEASED)
#define EVENT_HEADPHONES_INSERTED       EVENT_HEADPHONES
#define EVENT_HEADPHONES_REMOVED        (EVENT_HEADPHONES | KEY_WAS_RELEASED)
#define EVENT_AUDIO_ON                  EVENT_AUDIO
#define EVENT_AUDIO_OFF                 (EVENT_AUDIO | KEY_WAS_RELEASED)
#define EVENT_VALID_PASSWORD            EVENT_PASSWORD
#define EVENT_INVALID_PASSWORD          (EVENT_VALID_PASSWORD | KEY_WAS_RELEASED)
#define EVENT_TIMEOUT_1                 EVENT_TIMEOUT
#define EVENT_TIMEOUT_2                 (EVENT_TIMEOUT | KEY_WAS_RELEASED)
#define EVENT_TIMEOUT_3                 EVENT_ALT_TIMEOUTS
#define EVENT_TIMEOUT_USB_WAKE          (EVENT_ALT_TIMEOUTS | KEY_WAS_RELEASED)
#define EVENT_WATERMARK_CHECK           (EVENT_BATTERY_CRITICAL | KEY_WAS_RELEASED)
#define EVENT_BATTERY_NO_WAN            (EVENT_BATTERY_LOW | KEY_WAS_RELEASED)
#define EVENT_PRIME_PUMP                (EVENT_MMC | KEY_WAS_RELEASED)
#define EVENT_TPH_COMPLETE              (EVENT_WAN_RING | KEY_WAS_RELEASED)
#define EVENT_USB_WAKE                  (EVENT_KEY_WAKE | KEY_WAS_RELEASED)

#define EVENT_WAN_SWITCH_SWEEPER        EVENT_SWEEPER
#define EVENT_TWO_FINGER_SWEEPER        (EVENT_SWEEPER | KEY_WAS_RELEASED)

#endif // __FIONA_EVENTS_H__

