/*
 * pxa2xx-ssp.c  --  ALSA Soc Audio Layer
 *
 * Copyright 2005 Wolfson Microelectronics PLC.
 * Author: Liam Girdwood
 *         liam.girdwood@wolfsonmicro.com or linux@wolfsonmicro.com
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  Revision history
 *    12th Aug 2005   Initial version.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/device.h>

#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/initval.h>
#include <sound/soc.h>

#include <asm/hardware.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/audio.h>
#include <asm/arch/ssp.h>

#include "pxa2xx-pcm.h"

/*
 * SSP GPIO's
 */
#define GPIO26_SSP1RX_MD	(26 | GPIO_ALT_FN_1_IN)
#define GPIO25_SSP1TX_MD	(25 | GPIO_ALT_FN_2_OUT)
#define GPIO23_SSP1CLKS_MD	(23 | GPIO_ALT_FN_2_IN)
#define GPIO24_SSP1FRMS_MD	(24 | GPIO_ALT_FN_2_IN)
#define GPIO23_SSP1CLKM_MD	(23 | GPIO_ALT_FN_2_OUT)
#define GPIO24_SSP1FRMM_MD	(24 | GPIO_ALT_FN_2_OUT) 
 
#define GPIO11_SSP2RX_MD	(11 | GPIO_ALT_FN_2_IN)
#define GPIO13_SSP2TX_MD	(13 | GPIO_ALT_FN_1_OUT)
#define GPIO22_SSP2CLKS_MD	(22 | GPIO_ALT_FN_3_IN)
#define GPIO88_SSP2FRMS_MD	(88 | GPIO_ALT_FN_3_IN)
#define GPIO22_SSP2CLKM_MD	(22 | GPIO_ALT_FN_3_OUT)
#define GPIO88_SSP2FRMM_MD	(88 | GPIO_ALT_FN_3_OUT)

#define GPIO82_SSP3RX_MD	(82 | GPIO_ALT_FN_1_IN)
#define GPIO81_SSP3TX_MD	(81 | GPIO_ALT_FN_1_OUT)
#define GPIO84_SSP3CLKS_MD	(84 | GPIO_ALT_FN_1_IN)
#define GPIO83_SSP3FRMS_MD	(83 | GPIO_ALT_FN_1_IN)
#define GPIO84_SSP3CLKM_MD	(84 | GPIO_ALT_FN_1_OUT)
#define GPIO83_SSP3FRMM_MD	(83 | GPIO_ALT_FN_1_OUT)

#define PXA_SSP_HWFMT \
	(SND_SOC_HWFMT_DSP_B  |SND_SOC_HWFMT_CBM_CFM | SND_SOC_HWFMT_CBM_CFS | \
	SND_SOC_HWFMT_CBS_CFM | SND_SOC_HWFMT_CBS_CFS | SND_SOC_HWFMT_NB_NF | \
	SND_SOC_HWFMT_NB_IF)

/*
 * priv1 is scr divider when in master mode
 * -- liam, need some method of "doesn't matter" for fs, maybe 0  
 */
static struct snd_soc_hw_mode pxa2xx_ssp[] = {
	{PXA_SSP_HWFMT, 	SND_SOC_HWBITS(16),		8000,	256,	1626},
	{PXA_SSP_HWFMT, 	SND_SOC_HWBITS(16),		8000,	384,	1626}, 
	{PXA_SSP_HWFMT, 	SND_SOC_HWBITS(16),		11025,	256,	1180},
	{PXA_SSP_HWFMT, 	SND_SOC_HWBITS(16),		16000,	256,	813},
	{PXA_SSP_HWFMT, 	SND_SOC_HWBITS(16),		22050,	256,	590},
	{PXA_SSP_HWFMT, 	SND_SOC_HWBITS(16),		44100,	256,	295},
	{PXA_SSP_HWFMT, 	SND_SOC_HWBITS(16),		48000,	256,	271},
};

static struct ssp_dev ssp[3];

static pxa2xx_pcm_dma_params_t pxa2xx_ssp1_pcm_mono_out = {
	.name			= "SSP1 PCM Mono out",
	.dev_addr		= __PREG(SSDR_P1),
	.drcmr			= &DRCMRTXSSDR,
	.dcmd			= DCMD_INCSRCADDR | DCMD_FLOWTRG |
				  DCMD_BURST16 | DCMD_WIDTH2,
};

static pxa2xx_pcm_dma_params_t pxa2xx_ssp1_pcm_mono_in = {
	.name			= "SSP1 PCM Mono in",
	.dev_addr		= __PREG(SSDR_P1),
	.drcmr			= &DRCMRRXSSDR,
	.dcmd			= DCMD_INCTRGADDR | DCMD_FLOWSRC |
				  DCMD_BURST16 | DCMD_WIDTH2,
};

static pxa2xx_pcm_dma_params_t pxa2xx_ssp1_pcm_stereo_out = {
	.name			= "SSP1 PCM Stereo out",
	.dev_addr		= __PREG(SSDR_P1),
	.drcmr			= &DRCMRTXSSDR,
	.dcmd			= DCMD_INCSRCADDR | DCMD_FLOWTRG |
				  DCMD_BURST32 | DCMD_WIDTH4,
};

static pxa2xx_pcm_dma_params_t pxa2xx_ssp1_pcm_stereo_in = {
	.name			= "SSP1 PCM Stereo in",
	.dev_addr		= __PREG(SSDR_P1),
	.drcmr			= &DRCMRRXSSDR,
	.dcmd			= DCMD_INCTRGADDR | DCMD_FLOWSRC |
				  DCMD_BURST32 | DCMD_WIDTH4,
};

static pxa2xx_pcm_dma_params_t pxa2xx_ssp2_pcm_mono_out = {
	.name			= "SSP2 PCM Mono out",
	.dev_addr		= __PREG(SSDR_P2),
	.drcmr			= &DRCMRTXSS2DR,
	.dcmd			= DCMD_INCSRCADDR | DCMD_FLOWTRG |
				  DCMD_BURST16 | DCMD_WIDTH2,
};

static pxa2xx_pcm_dma_params_t pxa2xx_ssp2_pcm_mono_in = {
	.name			= "SSP2 PCM Mono in",
	.dev_addr		= __PREG(SSDR_P2),
	.drcmr			= &DRCMRRXSS2DR,
	.dcmd			= DCMD_INCTRGADDR | DCMD_FLOWSRC |
				  DCMD_BURST16 | DCMD_WIDTH2,
};

static pxa2xx_pcm_dma_params_t pxa2xx_ssp2_pcm_stereo_out = {
	.name			= "SSP2 PCM Stereo out",
	.dev_addr		= __PREG(SSDR_P2),
	.drcmr			= &DRCMRTXSS2DR,
	.dcmd			= DCMD_INCSRCADDR | DCMD_FLOWTRG |
				  DCMD_BURST32 | DCMD_WIDTH4,
};

static pxa2xx_pcm_dma_params_t pxa2xx_ssp2_pcm_stereo_in = {
	.name			= "SSP2 PCM Stereo in",
	.dev_addr		= __PREG(SSDR_P2),
	.drcmr			= &DRCMRRXSS2DR,
	.dcmd			= DCMD_INCTRGADDR | DCMD_FLOWSRC |
				  DCMD_BURST32 | DCMD_WIDTH4,
};

static pxa2xx_pcm_dma_params_t pxa2xx_ssp3_pcm_mono_out = {
	.name			= "SSP3 PCM Mono out",
	.dev_addr		= __PREG(SSDR_P3),
	.drcmr			= &DRCMRTXSS3DR,
	.dcmd			= DCMD_INCSRCADDR | DCMD_FLOWTRG |
				  DCMD_BURST16 | DCMD_WIDTH2,
};

static pxa2xx_pcm_dma_params_t pxa2xx_ssp3_pcm_mono_in = {
	.name			= "SSP3 PCM Mono in",
	.dev_addr		= __PREG(SSDR_P3),
	.drcmr			= &DRCMRRXSS3DR,
	.dcmd			= DCMD_INCTRGADDR | DCMD_FLOWSRC |
				  DCMD_BURST16 | DCMD_WIDTH2,
};

static pxa2xx_pcm_dma_params_t pxa2xx_ssp3_pcm_stereo_out = {
	.name			= "SSP3 PCM Stereo out",
	.dev_addr		= __PREG(SSDR_P3),
	.drcmr			= &DRCMRTXSS3DR,
	.dcmd			= DCMD_INCSRCADDR | DCMD_FLOWTRG |
				  DCMD_BURST32 | DCMD_WIDTH4,
};

static pxa2xx_pcm_dma_params_t pxa2xx_ssp3_pcm_stereo_in = {
	.name			= "SSP3 PCM Stereo in",
	.dev_addr		= __PREG(SSDR_P3),
	.drcmr			= &DRCMRRXSS3DR,
	.dcmd			= DCMD_INCTRGADDR | DCMD_FLOWSRC |
				  DCMD_BURST32 | DCMD_WIDTH4,
};

static pxa2xx_pcm_dma_params_t *ssp_dma_params[3][4] = {
	{&pxa2xx_ssp1_pcm_mono_out, &pxa2xx_ssp1_pcm_mono_in,
	&pxa2xx_ssp1_pcm_stereo_out,&pxa2xx_ssp1_pcm_stereo_in,},
	{&pxa2xx_ssp2_pcm_mono_out, &pxa2xx_ssp2_pcm_mono_in,
	&pxa2xx_ssp2_pcm_stereo_out, &pxa2xx_ssp2_pcm_stereo_in,},
	{&pxa2xx_ssp3_pcm_mono_out, &pxa2xx_ssp3_pcm_mono_in,
	&pxa2xx_ssp3_pcm_stereo_out,&pxa2xx_ssp3_pcm_stereo_in,},
};

static pxa2xx_gpio_t ssp_gpios[3][2] = {
	{{ /* SSP1 Slave */	
		.rx = GPIO26_SSP1RX_MD,
		.tx = GPIO25_SSP1TX_MD,
		.clk = GPIO23_SSP1CLKS_MD,
		.frm = GPIO24_SSP1FRMS_MD,
	},
	{ /* SSP1 Master */	
		.rx = GPIO26_SSP1RX_MD,
		.tx = GPIO25_SSP1TX_MD,
		.clk = GPIO23_SSP1CLKM_MD,
		.frm = GPIO24_SSP1FRMM_MD,
	}},
	{{ /* SSP2 Slave */	
		.rx = GPIO11_SSP2RX_MD,
		.tx = GPIO13_SSP2TX_MD,
		.clk = GPIO22_SSP2CLKS_MD,
		.frm = GPIO88_SSP2FRMS_MD,
	},
	{ /* SSP2 Master */	
		.rx = GPIO11_SSP2RX_MD,
		.tx = GPIO13_SSP2TX_MD,
		.clk = GPIO22_SSP2CLKM_MD,
		.frm = GPIO88_SSP2FRMM_MD,
	}},
	{{ /* SSP3 Slave */	
		.rx = GPIO82_SSP3RX_MD,
		.tx = GPIO81_SSP3TX_MD,
		.clk = GPIO84_SSP3CLKS_MD,
		.frm = GPIO83_SSP3FRMS_MD,
	},
	{ /* SSP3 Master */	
		.rx = GPIO82_SSP3RX_MD,
		.tx = GPIO81_SSP3TX_MD,
		.clk = GPIO84_SSP3CLKM_MD,
		.frm = GPIO83_SSP3FRMM_MD,
	}},
};

static int pxa2xx_ssp_startup(snd_pcm_substream_t *substream)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	struct snd_soc_pcm_interface *pcm_i = pcm_c->pcm_i;
	int ret = 0;
	
	if((ret = ssp_init (&ssp[pcm_i->id], pcm_i->id + 1, SSP_NO_IRQ)) < 0)
		return ret;
	
	ssp_disable(&ssp[pcm_i->id]);
	return ret;
}

static void pxa2xx_ssp_shutdown(snd_pcm_substream_t *substream)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	struct snd_soc_pcm_interface *pcm_i = pcm_c->pcm_i;

	ssp_disable(&ssp[pcm_i->id]);
	ssp_exit(&ssp[pcm_i->id]);
}

#ifdef CONFIG_PM
static int pxa2xx_ssp_suspend(struct snd_soc_pcm_interface *interface, pm_message_t state)
{
	int ret = 0;

	//if (interface->card && level == SUSPEND_DISABLE)
		//ret = pxa2xx_ac97_do_suspend(card, SNDRV_CTL_POWER_D3cold);

	return ret;
}

static int pxa2xx_ssp_resume(struct snd_soc_pcm_interface *interface)
{
	int ret = 0;

	//if (interface->card && level == RESUME_ENABLE)
		//ret = pxa2xx_ac97_do_resume(card, SNDRV_CTL_POWER_D0);

	return ret;
}

#else
#define pxa2xx_ssp_suspend	NULL
#define pxa2xx_ssp_resume	NULL
#endif

static int pxa2xx_ssp_hw_params(snd_pcm_substream_t *substream,
				snd_pcm_hw_params_t *params)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	struct snd_soc_pcm_interface *pcm_i = pcm_c->pcm_i;
	int master = 0, dma = 0;
	u32 ssp_mode = 0, ssp_setup = 0, psp_mode = 0;
	int bits = generic_ffs(pcm_i->hw_runtime.hbits);
	
	if(pcm_i->hw_runtime.hformat & SND_SOC_HWFMT_CBS_CFS)
		master = 1;
	
	pxa_gpio_mode(ssp_gpios[pcm_i->id][master].rx);
	pxa_gpio_mode(ssp_gpios[pcm_i->id][master].tx);
	pxa_gpio_mode(ssp_gpios[pcm_i->id][master].frm);
	pxa_gpio_mode(ssp_gpios[pcm_i->id][master].clk);
	
	if (substream->stream != SNDRV_PCM_STREAM_PLAYBACK)
		dma = 1;
	if (params_channels(params) == 2)
		dma += 2;
	pcm_i->platform_data = ssp_dma_params[pcm_i->id][dma];
	
	ssp_mode = SSCR0_PSP | SSCR0_DataSize(bits);
	ssp_setup = SSCR1_RxTresh(14) | SSCR1_TxTresh(1) | SSCR1_TRAIL |
				SSCR1_TSRE | SSCR1_RSRE | SSCR1_TIE | SSCR1_RIE;
	
	switch(pcm_i->hw_runtime.hformat & SND_SOC_CLOCK_MASK) {
		case SND_SOC_HWFMT_CBM_CFM:
			ssp_setup |= (SSCR1_SCLKDIR | SSCR1_SFRMDIR);
			break;
		case SND_SOC_HWFMT_CBM_CFS:
			ssp_setup |= SSCR1_SCLKDIR;
			break;
		case SND_SOC_HWFMT_CBS_CFM:
			ssp_setup |= SSCR1_SFRMDIR;
			break;
	}
	
	switch(pcm_i->hw_runtime.hformat & SND_SOC_INV_MASK){
		case SND_SOC_HWFMT_NB_NF:
			psp_mode |= SSPSP_SFRMP;
			break;
	}
	if (pcm_i->hw_runtime.hformat & SND_SOC_HWFMT_DSP_A)
		psp_mode |= SSPSP_SCMODE(2);
	if (pcm_i->hw_runtime.hformat & SND_SOC_HWFMT_DSP_B)
		psp_mode |= SSPSP_SCMODE(3);

	ssp_disable(&ssp[pcm_i->id]);
	ssp_config(&ssp[pcm_i->id], ssp_mode, ssp_setup, psp_mode,
					SSCR0_SerClkDiv(pcm_i->hw_runtime.priv1));
	ssp_enable(&ssp[pcm_i->id]);

	return 0;
}

static struct snd_soc_pcm_interface pxa_ssp_interfaces[] = {
	{	.name = "pxa2xx-ssp1",
		.id = 0,
		.type = SND_SOC_SSP,
		.suspend = pxa2xx_ssp_suspend,
		.resume = pxa2xx_ssp_resume,
		.playback = {
			.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 |
			SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100,
			.formats = SNDRV_PCM_FMTBIT_S16_LE,
			.channels_min = 1,
			.channels_max = 1,},
		.capture = {
			.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 |
			SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100,
			.formats = SNDRV_PCM_FMTBIT_S16_LE,
			.channels_min = 1,
			.channels_max = 2,},
		.ops = {
			.startup = pxa2xx_ssp_startup,
			.shutdown = pxa2xx_ssp_shutdown,
			.hw_params = pxa2xx_ssp_hw_params,},
		.hw = {
			.hmodes = &pxa2xx_ssp[0],
			.num_hmodes = ARRAY_SIZE(pxa2xx_ssp),},
	},
	{	.name = "pxa2xx-ssp2",
		.id = 1,
		.type = SND_SOC_SSP,
		.suspend = pxa2xx_ssp_suspend,
		.resume = pxa2xx_ssp_resume,
		.playback = {
			.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 |
			SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100,
			.formats = SNDRV_PCM_FMTBIT_S16_LE,
			.channels_min = 1,
			.channels_max = 1,},
		.capture = {
			.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 |
			SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100,
			.formats = SNDRV_PCM_FMTBIT_S16_LE,
			.channels_min = 1,
			.channels_max = 1,},
		.ops = {
			.startup = pxa2xx_ssp_startup,
			.shutdown = pxa2xx_ssp_shutdown,
			.hw_params = pxa2xx_ssp_hw_params,},
		.hw = {
			.hmodes = &pxa2xx_ssp[0],
			.num_hmodes = ARRAY_SIZE(pxa2xx_ssp),},
	},
	{	.name = "pxa2xx-ssp3",
		.id = 2,
		.type = SND_SOC_SSP,
		.suspend = pxa2xx_ssp_suspend,
		.resume = pxa2xx_ssp_resume,
		.playback = {
			.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 |
			SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100,
			.formats = SNDRV_PCM_FMTBIT_S16_LE,
			.channels_min = 1,
			.channels_max = 1,},
		.capture = {
			.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 |
			SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100,
			.formats = SNDRV_PCM_FMTBIT_S16_LE,
			.channels_min = 1,
			.channels_max = 1,},
		.ops = {
			.startup = pxa2xx_ssp_startup,
			.shutdown = pxa2xx_ssp_shutdown,
			.hw_params = pxa2xx_ssp_hw_params,},
		.hw = {
			.hmodes = &pxa2xx_ssp[0],
			.num_hmodes = ARRAY_SIZE(pxa2xx_ssp),},
	},
};

static int __init pxa2xx_soc_ssp_init(void)
{
	snd_soc_register_interface(&pxa_ssp_interfaces[0]);
	snd_soc_register_interface(&pxa_ssp_interfaces[1]);
	snd_soc_register_interface(&pxa_ssp_interfaces[2]);
	return 0;
}

static void __exit pxa2xx_soc_ssp_exit(void)
{
	snd_soc_unregister_interface(&pxa_ssp_interfaces[2]);
	snd_soc_unregister_interface(&pxa_ssp_interfaces[1]);
	snd_soc_unregister_interface(&pxa_ssp_interfaces[0]);
}

EXPORT_SYMBOL_GPL(pxa_ssp_interfaces);

module_init(pxa2xx_soc_ssp_init);
module_exit(pxa2xx_soc_ssp_exit);

/* Module information */ 
MODULE_AUTHOR("Liam Girdwood, liam.girdwood@wolfsonmicro.com, www.wolfsonmicro.com");
MODULE_DESCRIPTION("pxa2xx SSP/PCM SoC Interface");
MODULE_LICENSE("GPL");
